﻿using FukuracamDX;
using System.Collections.Generic;

namespace FatopTheWorld{
	class Common{
		public static readonly FCAMSize FONT_SIZE = new FCAMSize(16);
		const int FONT_NUMBER_COLUMNS = 16;
		private static readonly string[] MODE_TEXT_FORMAT = new string[]{"{0}\n{1, 6}{2, 2}", "{0}\n BATTLE ", "FREE\nMODE"};
		public const string FILEPATH_FONT = @"Common\font.png";
		public const string FILEPATH_SOUND_SELECT = @"SE\select.wav";
		public const string FILEPATH_SOUND_DECIDE = @"SE\decide.wav";
		public const string FILEPATH_SOUND_CANCEL = @"SE\cancel.wav";
		public const string FILEPATH_CHARACTER_FORMAT = @"Character\{0:D2}-{1:D1}.png";
		/// <summary>最大ステージ（≠ステージ数。※ステージは、オープニングやエンディングも含む）。MemoryクラスのDifficult参照。</summary>
#if TRIAL
		public static readonly sbyte[] MAX_STAGE = {4, 4, 0, 0};
#else
		public static readonly sbyte[] MAX_STAGE = {6, 9, 12, 2};
#endif
		public static readonly sbyte[] STAGE_SHIFT = {0, 0, 0, 10};
		public static readonly sbyte[] CHARA_SHIFT = {1, 1, 1, 9};
		public const int SECRETSTAGE_FROM = 10;
		public const int SECRETCHARA_FROM = 8;
		public const int MAX_CHARA = 11;
		//	とあるステージまでいった時に倒したキャラクター数
		//	※必ず 1 以上を指定。そうしないと対戦モードでキャラクターが選択出来ない。
		//											                STAGE op   1   2   3   4   5   6   7   l  ed  s1  s2  sl
		/// <summary>[difficult, stages]</summary>
#if TRIAL
		public static readonly sbyte[,] CLEARED_CHARACTERS_AT_STAGE = {
																		  {1,  1,  2,  3,  3},
																		  {1,  1,  2,  3,  3},
																	  };
#else
		public static readonly sbyte[,] CLEARED_CHARACTERS_AT_STAGE = {
																		  {1,  1,  2,  3,  4,  5,  5,  5,  5,  5,  5,  5,  5},
																		  {1,  1,  2,  3,  4,  5,  6,  7,  8,  8,  8,  8,  8},
																		  {1,  1,  2,  3,  4,  5,  6,  7,  8,  8,  9, 10, 11},
																		  {9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9, 10, 11},
																	  };
#endif
		/// <summary>これは参照されたものなので Common では解放しない。</summary>
		public static FCAMGraphics Graphics;
		/// <summary>これは参照されたものなので Common では解放しない。</summary>
		public static FCAMContentsManager ContentsManager;
		public static FCAMTexture Texture;
		public static FCAMInputState KeyLeft = null, KeyUp = null, KeyRight = null, KeyDown = null, KeyA = null, KeyB = null, KeyL = null, KeyR = null, KeyDebug1 = null, KeyDebug2 = null, KeyDebug3 = null;
		public static FCAMInputState KeyLeft2 = null, KeyUp2 = null, KeyRight2 = null, KeyDown2 = null, KeyA2 = null, KeyB2 = null, KeyL2 = null, KeyR2 = null;
		public static FCAMColor WindowBorderColor, WindowBackLTColor, WindowBackRTColor, WindowBackLBColor, WindowBackRBColor;
		public static int WindowBorderWeight = 3;
		private static sbyte __Stage = 0;
		private static int BeforeSelect;
		private static IFCAMAudioCue SECue = null;
		private static string __BGMFilePath; public static string BGMFilePath{
			get{
				return __BGMFilePath == null ? "" : __BGMFilePath;
			}
			private set{
				__BGMFilePath = value;
			}
		}
		public static IFCAMAudioCue BGMCue = null;
		/// <summary>ステージ数。</summary>
		public static sbyte Stage{get{return(__Stage);} set{if((__Stage = value) > Game.MemoryData.ClearedStages){Game.MemoryData.ClearedStages = value;}}}
		/// <summary>ファイル上でのステージ数。例えばMADNESS!はシークレットステージのみのため、ステージ数とファイルのステージが異なるため。</summary>
		public static sbyte StageOnFile{get{return((sbyte)(Stage + STAGE_SHIFT[Game.MemoryData.Difficult]));}}
		//	現在の難易度における最大ステージ（＝ステージ数 - 1）
		public static sbyte MaxStage{get{return(MAX_STAGE[Game.MemoryData.Difficult]);}}
		public static int Mode = 0;
		public static FCAMFont FontCommon, FontStory;
		public static IFCAMAudioCue CueSelect, CueDecide, CueCancel;
		private static Dictionary<string, long> BGMPositionDictionary;
		//
		public static Scenes.Battle Battle;
		static Common(){
			WindowBorderColor = FCAMColor.DarkBlue;
			WindowBackLTColor = FCAMColor.HalfLightBlue;
			WindowBackLBColor = WindowBackRTColor = FCAMColor.HalfBlue;
			WindowBackRBColor = FCAMColor.HalfDarkBlue;
			BGMPositionDictionary = new Dictionary<string,long>();

		}
		#pragma	warning	disable	0465
		public static void Finalize(){
			FCAMSystem.DisposeClass(out CueCancel);
			FCAMSystem.DisposeClass(out CueDecide);
			FCAMSystem.DisposeClass(out CueSelect);
			FCAMSystem.DisposeClass(out FontCommon);
			FCAMSystem.DisposeClass(out FontStory);
			FCAMSystem.DisposeClass(out SECue);
			FCAMSystem.DisposeClass(out BGMCue);
			FCAMSystem.DisposeClass(out Texture);
		}
		/// <summary>
		/// 選択音を鳴らす準備をする。
		/// </summary>
		public static void ReadyPlayCueSelect(int select){
			BeforeSelect = select;
		}
		/// <summary>
		/// Ready との差で選択音を再生する。
		/// </summary>
		/// <param name="not">not = true の時、再生するための条件を反転させる。</param>
		public static void PlayCueSelect(int select, bool not = false){
			if(BeforeSelect != select){
				if(!not){
					CueSelect.Play();
				}
			}else{
				if(not){
					CueSelect.Play();
				}
			}
		}
		/// <summary>
		/// Sound Effect を再生。
		/// </summary>
		public static void PlaySE(string filePath){
			FCAMSystem.DisposeClass(out SECue);
			SECue = ContentsManager.MakeCue(filePath);
			SECue.Play();
		}
		/// <summary>
		/// BGMを再生。null or 空文字列を指定すると無音
		/// </summary>
		public static void ChangeBGM(string filePath, bool posContinue = false, int end = -1, int loop = 0){
			if(filePath == BGMFilePath){
				return;
			}
			//	途中再生用に、再生前に再生していたBGMの位置情報を記録。
			if(!BGMPositionDictionary.ContainsKey(BGMFilePath)){
				BGMPositionDictionary.Add(BGMFilePath, 0);
			}
			BGMPositionDictionary[BGMFilePath] = FCAMAudioCue.IsCueUsable(BGMCue) ? BGMCue.Position : 0;
			FCAMSystem.DisposeClass(out BGMCue);
			BGMFilePath = filePath;
			if(!string.IsNullOrEmpty(filePath)){
				try{
					BGMCue = ContentsManager.MakeCue(BGMFilePath, loop, end);
					if(posContinue && BGMPositionDictionary.ContainsKey(filePath)){
						BGMCue.Position = BGMPositionDictionary[filePath];
					}
					BGMCue.Loop();
				}catch{
					ChangeBGM(BGM.Nothing);
				}
			}
		}
		public static void ChangeBGM(BGM bgm, bool posContinue = false, int end = -1, int loop = 0){
			ChangeBGM(bgm == BGM.Nothing ? null : Audio.GetPath(bgm), posContinue, end, loop);
		}
		public static void InitializeKeySettings(FCAMInput input){
			Common.KeyLeft = input.Keyboard[FCAMKey.LeftArrow];
			Common.KeyUp = input.Keyboard[FCAMKey.UpArrow];
			Common.KeyRight = input.Keyboard[FCAMKey.RightArrow];
			Common.KeyDown = input.Keyboard[FCAMKey.DownArrow];
			Common.KeyA = input.Keyboard[FCAMKey.Z];
			Common.KeyB = input.Keyboard[FCAMKey.X];
			Common.KeyL = input.Keyboard[FCAMKey.A];
#if ATTACK_MAGIC
			Common.KeyR = input.Keyboard[FCAMKey.S];
#else
			Common.KeyR = input.Keyboard[FCAMKey.C];
#endif
			Common.KeyLeft2 = input.Keyboard[FCAMKey.NumPad4];
			Common.KeyUp2 = input.Keyboard[FCAMKey.NumPad8];
			Common.KeyRight2 = input.Keyboard[FCAMKey.NumPad6];
			Common.KeyDown2 = input.Keyboard[FCAMKey.NumPad2];
			Common.KeyA2 = input.Keyboard[FCAMKey.NumPad5];
			Common.KeyL2 = input.Keyboard[FCAMKey.NumPad7];
#if ATTACK_MAGIC
			Common.KeyB2 = input.Keyboard[FCAMKey.NumPad0];
			Common.KeyR2 = input.Keyboard[FCAMKey.NumPad9];
#else
			Common.KeyB2 = input.Keyboard[FCAMKey.NumPad1];
			Common.KeyR2 = input.Keyboard[FCAMKey.NumPad3];
#endif
#if DEBUG
			if(FCAMDebug.IsDebug)
			{
				Common.KeyDebug1 = input.Keyboard[FCAMKey.D1];
				Common.KeyDebug2 = input.Keyboard[FCAMKey.D2];
				Common.KeyDebug3 = input.Keyboard[FCAMKey.D3];
			}
			else
#endif
			{
				Common.KeyDebug1 = Common.KeyDebug2 = Common.KeyDebug3 = FCAMInputState.Empty;
			}
		}
		public static void Initialize(FCAMGraphics g, FCAMInput input, FCAMContentsManager manager){
			ContentsManager = manager;
			FontCommon = manager.MakeFont(16, "MS Gothic");
			FontStory = manager.MakeFont(24, "MS Gothic");
			CueSelect = manager.MakeCue(FILEPATH_SOUND_SELECT);
			CueDecide = manager.MakeCue(FILEPATH_SOUND_DECIDE);
			CueCancel = manager.MakeCue(FILEPATH_SOUND_CANCEL);
			Common.Graphics = g;
			Common.Texture = manager.MakeTexture(Common.FILEPATH_FONT);
			InitializeKeySettings(input);
			Battle = new Scenes.Battle(true);
		}
		/// <summary>
		/// 指定 FCAMInputState がどのプレイヤーのキーかを ID で返す。どれにも当てはまらない時は -1 を返す。
		/// </summary>
		public static sbyte CommonKeyPlayerID(FCAMInputState state){
			if(
				state == Common.KeyLeft ||
				state == Common.KeyUp ||
				state == Common.KeyRight ||
				state == Common.KeyDown ||
				state == Common.KeyA ||
				state == Common.KeyB ||
#if ATTACK_MAGIC
				state == Common.KeyL ||
#endif
				state == Common.KeyR
			){
				return(0);
			}else if(
				state == Common.KeyLeft2 ||
				state == Common.KeyUp2 ||
				state == Common.KeyRight2 ||
				state == Common.KeyDown2 ||
				state == Common.KeyA2 ||
				state == Common.KeyB2 ||
#if ATTACK_MAGIC
				state == Common.KeyL2 ||
#endif
				state == Common.KeyR2
			){
				return(1);
			}
			return(-1);
		}
		public static FCAMInputState CommonKeyLeft(){
			return(FCAMInputState.Choose(Common.KeyLeft, Common.KeyLeft2));
		}
		public static FCAMInputState CommonKeyUp(){
			return(FCAMInputState.Choose(Common.KeyUp, Common.KeyUp2));
		}
		public static FCAMInputState CommonKeyRight(){
			return(FCAMInputState.Choose(Common.KeyRight, Common.KeyRight2));
		}
		public static FCAMInputState CommonKeyDown(){
			return(FCAMInputState.Choose(Common.KeyDown, Common.KeyDown2));
		}
		public static FCAMInputState CommonKeyA(){
			return(FCAMInputState.Choose(Common.KeyA, Common.KeyA2));
		}
		public static FCAMInputState CommonKeyB(){
			return(FCAMInputState.Choose(Common.KeyB, Common.KeyB2));
		}
#if ATTACK_MAGIC
		public static FCAMInputState CommonKeyL(){
			return(FCAMInputState.Choose(Common.KeyL, Common.KeyL2));
		}
#endif
		public static FCAMInputState CommonKeyR(){
			return(FCAMInputState.Choose(Common.KeyR, Common.KeyR2));
		}
		/*
		public static FCAMInputState CommonKeyLeft(FCAMInputStateGetType getType = FCAMInputStateGetType.Once){
			return(FCAMInputState.Choose(getType, Common.KeyLeft, Common.KeyLeft2));
		}
		public static FCAMInputState CommonKeyUp(FCAMInputStateGetType getType = FCAMInputStateGetType.Once){
			return(FCAMInputState.Choose(getType, Common.KeyUp, Common.KeyUp2));
		}
		public static FCAMInputState CommonKeyRight(FCAMInputStateGetType getType = FCAMInputStateGetType.Once){
			return(FCAMInputState.Choose(getType, Common.KeyRight, Common.KeyRight2));
		}
		public static FCAMInputState CommonKeyDown(FCAMInputStateGetType getType = FCAMInputStateGetType.Once){
			return(FCAMInputState.Choose(getType, Common.KeyDown, Common.KeyDown2));
		}
		public static FCAMInputState CommonKeyA(FCAMInputStateGetType getType = FCAMInputStateGetType.Once){
			return(FCAMInputState.Choose(getType, Common.KeyA, Common.KeyA2));
		}
		public static FCAMInputState CommonKeyB(FCAMInputStateGetType getType = FCAMInputStateGetType.Once){
			return(FCAMInputState.Choose(getType, Common.KeyB, Common.KeyB2));
		}
		public static FCAMInputState CommonKeyL(FCAMInputStateGetType getType = FCAMInputStateGetType.Once){
			return(FCAMInputState.Choose(getType, Common.KeyL, Common.KeyL2));
		}
		public static FCAMInputState CommonKeyR(FCAMInputStateGetType getType = FCAMInputStateGetType.Once){
			return(FCAMInputState.Choose(getType, Common.KeyR, Common.KeyR2));
		}
		 */
		public static string GetModeText(int? mode = null){
			return FCAMStrings.Format(MODE_TEXT_FORMAT[mode ?? Mode],
					Game.MemoryData.GetDifficultName(),
					StageOnFile >= Common.SECRETSTAGE_FROM ? "SECRET" : "STAGE ",
					Memory.GetClearedCharactersAtStage(StageOnFile) - (StageOnFile >= Common.SECRETSTAGE_FROM ? Common.SECRETCHARA_FROM : 0)
				   );
		}
		public static void DrawWindow(FCAMRectangle? rectangle = null){
			FCAMRectangle rectangleWindow = rectangle.GetValueOrDefault(new FCAMRectangle(false, 0, 0, Graphics.Width, Graphics.Height));
			rectangleWindow.Inflate(WindowBorderWeight, WindowBorderWeight);
			Graphics.FillRectangle(rectangleWindow.X, rectangleWindow.Y, rectangleWindow.Width, rectangleWindow.Height, WindowBorderColor);
			rectangleWindow.Inflate(-WindowBorderWeight, -WindowBorderWeight);
			Graphics.FillGradatedRectangle(rectangleWindow.X, rectangleWindow.Y, rectangleWindow.Width, rectangleWindow.Height, WindowBackLTColor, WindowBackRTColor, WindowBackLBColor, WindowBackRBColor);
		}
		/// <summary>
		/// 一文字ずつ描画（固定幅・高さ）
		/// </summary>
		public static void DrawString(float x, float y, string text, bool center, FCAMColor? blender = null){
			string[] lines = FCAMStrings.SplitLine(text);
			float start = x;
			int maxColumns = 0;
			foreach(string sz in lines)
				if(sz.Length > maxColumns)
					maxColumns = sz.Length;
			foreach(string sz in lines){
				x = start + (maxColumns - sz.Length) * FONT_SIZE.Width / 2;
				foreach(char c in sz){
					Graphics.Draw(Texture, x, y, FONT_SIZE.Width, FONT_SIZE.Height,
						(c % FONT_NUMBER_COLUMNS) * FONT_SIZE.Width,
						(c / FONT_NUMBER_COLUMNS) * FONT_SIZE.Height, blender.GetValueOrDefault(FCAMColor.White));
					x += FONT_SIZE.Width;
				}
				y += FONT_SIZE.Height;
			}
		}
		public static void DrawString(float x, float y, string text, FCAMColor? blender = null){
			DrawString(x, y, text, true, blender);
		}
		public static void DrawString(FCAMPoint pos, string text, FCAMColor? blender = null){
			DrawString(pos.X, pos.Y, text, blender);
		}
		/// <summary>
		/// ヘルプメッセージ表示
		/// </summary>
		public static void DrawHelp(string text){
			FCAMRectangle rectangle = Graphics.Font.Measure(text);
			Graphics.DrawString(FCAMDrawStringFormat.Hemming, Graphics.Width - rectangle.Width, Graphics.Height - rectangle.Height, text);
		}
		/// <summary>
		/// 
		/// </summary>
		/// <returns></returns>
		public static string GetCharacterPath(int characterID, int index){
			return FCAMStrings.Format(FILEPATH_CHARACTER_FORMAT, characterID, index);
		}
	}
}
