﻿using FukuracamDX;

namespace FatopTheWorld{
	enum Languages{
		Jpn, NonJpn
	}
	class Game: FCAMGame{
		public static Memory MemoryData;
		public static Game Instance;
		//
		static readonly string[] DEBUG_MENU_ITEMS = new string[]{
			"BACK",
			"GOTO: LOGO",
			"GOTO: TITLE",
			"GOTO: BATTLE(TEST)",
			"GOTO: BATTLE(VS)",
			"GOTO: STORY",
			"GOTO: GALLERY",
			"",
			"",
			"",
			"",
		};
		const string COMMAND_BATTLE = "Battle";
		const string COMMAND_TITLE = "Title";
		const string COMMAND_VS = "VS";
		const string COMMAND_STORY = "Story";
		const string COMMAND_CPUBATTLE = "CPUBattle";
		const string COMMAND_CPU = "CPU";
		public static bool DrawDebug{get; private set;}
		public Languages Language{get; set;}
		//
		private bool debugMenu;
		private int menu;
		[System.STAThread] static void Main(){
			MemoryData = new Memory();
#if LOG
			FCAMDebug.OutputLog = true;
#endif
			(Instance = new Game()).Run(640, 480);
			FCAMSystem.DisposeClass(out Instance);
		}
		protected override void Load(FCAMGraphics g, FCAMInput input, FCAMContentsManager manager){
#if LOG
			FCAMDebug.TraceLine("Game.Load(): start");
#endif
#if DEBUG
			if(FCAMDebug.IsDebug){
				//this.Window.TopMost = true;
			}
#endif
			Common.Initialize(g, input, manager);
			//	Common.Initialize でキー情報を初期してから読み込む
			FCAMFile.Read(Memory.FILEPATH_MEMORY, MemoryData);
			debugMenu = false;
			DrawDebug = true;
			Language = Languages.NonJpn;
			//	デバッグ用コマンドラインによる処理の分岐
#if DEBUG
			string command = FCAMSystem.GetCommandLineArgs().Length > 1 ? FCAMSystem.GetCommandLineArgs()[1] : "";
			switch(command){
				case COMMAND_BATTLE:{
					Common.Battle.Mode = Scenes.Battle.BattleMode.Test;
					PushScene(Common.Battle);
					break;
				}
				case COMMAND_TITLE:{
					PushScene<Scenes.Title>();
					break;
				}
				case COMMAND_VS:{
					Common.Battle.Mode = Scenes.Battle.BattleMode.VS;
					PushScene(Common.Battle);
					break;
				}
				case COMMAND_STORY:{
					PushScene<Scenes.Story>();
					break;
				}
				case COMMAND_CPUBATTLE:{
					Common.Battle.Mode = Scenes.Battle.BattleMode.CPUBattleTest;
					PushScene(Common.Battle);
					break;
				}
				case COMMAND_CPU:{
					Common.Battle.Mode = Scenes.Battle.BattleMode.CPUTest;
					PushScene(Common.Battle);
					break;
				}
				default:{
					PushScene<Scenes.Title>();
					PushScene<Scenes.Logo>();
					PushScene<Scenes.ChooseLanguage>();
					break;
				}
			}
#else
			PushScene<Scenes.Title>();
			PushScene<Scenes.Logo>();
			PushScene<Scenes.ChooseLanguage>();
#endif
#if LOG
			FCAMDebug.TraceLine("Game.Load(): end");
#endif
		}
		protected override void Unload(){
#if LOG
			FCAMDebug.TraceLine("Game.Unload(): start");
#endif
			FCAMSystem.DisposeClass(out Common.Texture);
			Common.Finalize();
#if LOG
			FCAMDebug.TraceLine("Game.Unload(): end");
#endif
		}
		protected override void Calc(FCAMInput input, FCAMContentsManager manager){
#if DEBUG
			if(FCAMDebug.IsDebug){
				Turbo = input.Keyboard[FCAMKey.D0].Press;
				if(Common.KeyDebug3){
					DrawDebug ^= true;
				}
				if(!debugMenu){
					if(Common.KeyDebug1)
						debugMenu = true;
					return;
				}
			}
#endif
			TemporaryRegulation = debugMenu;
			if(Common.KeyLeft)
				menu--;
			if(Common.KeyRight)
				menu++;
			if(Common.KeyUp)
				menu += 10;
			if(Common.KeyDown)
				menu -= 10;
			menu = FCAMMath.Within(0, menu, DEBUG_MENU_ITEMS.Length - 1);
			if(Common.KeyDebug1){
				switch(menu){
					//	BACK
					case 0:{
						debugMenu = false;
						break;
					}
					//	LOGO
					case 1:{
						ClearScenes();
						PushScene<Scenes.Logo>();
						debugMenu = false;
						break;
					}
					//	TITLE
					case 2:{
						ClearScenes();
						PushScene<Scenes.Title>();
						debugMenu = false;
						break;
					}
					//	BATTLE(TEST)
					case 3:{
						ClearScenes();
						Common.Battle.Mode = Scenes.Battle.BattleMode.Test;
						PushScene(Common.Battle);
						debugMenu = false;
						break;
					}
					//	BATTLE(VS)
					case 4:{
						ClearScenes();
						Common.Battle.Mode = Scenes.Battle.BattleMode.VS;
						PushScene(Common.Battle);
						debugMenu = false;
						break;
					}
					//	BATTLE(STORY)
					case 5:{
						ClearScenes();
						PushScene<Scenes.Story>();
						debugMenu = false;
						break;
					}
					//	GALLERY
					case 6:{
						ClearScenes();
						PushScene<Scenes.Gallery>();
						debugMenu = false;
						break;
					}
				}
			}
		}
		protected override void Draw(FCAMGraphics g){
			string sz = "";
			if(debugMenu){
				g.Clear(FCAMColor.Black);
				sz += FCAMStrings.FormatLine("Debug Menu = {0}: {1}", menu, DEBUG_MENU_ITEMS[menu]);
			}
#if DEBUG
			if(FCAMDebug.IsDebug){
				if(DrawDebug){
					sz += "Now = " + TemporaryRegulation + ", " + FCAMDateTime.Now.ToString() + FCAMSystem.NewLine;
					sz += FCAMStrings.Format("Skips: {0}, Turbo: {1}, FPMS: {2}{3}",DrawSkips, Turbo, ProcFPMS, FCAMSystem.NewLine);
				}
			}
#endif
			g.DrawString(0, 0, sz);
		}
	}
}

