﻿using FukuracamDX;
using System.Collections.Generic;

namespace FatopTheWorld.Scenes{
	partial class Battle: FCAMScene{
		/// <summary>
		/// ゲームを開始する際に必要なプレイヤー情報
		/// </summary>
		class CharacterInfo{
			static CharacterInfo(){
				//	キャラクターのベース情報を元に難易度個数分のCPU情報を含めて作成
				//	もしCPU情報が登録済みなら、それを元に作成
				//	なおCPUが不要なら null の場合もあるが、その例外処理は行なわない。
				//	（あくまでCPU情報がnullを含めて登録済みでない時のみ自動生成）
				Infos = new CharacterInfo[CharacterBase.Length, Memory.NumberDifficults];
				for(int chara = 0; chara < CharacterBase.Length; chara++){
					for(int i = 0; i < Memory.NumberDifficults; i++){
						CharacterInfo info = CharacterBase[chara];
						CPUBrain cpu;
						if(i > 0 && i >= info.BaseCPUs.Length){
							int operateInterval = FCAMMath.Max(0, info.BaseCPUs[0].OperateInterval - i * 10);
							sbyte maxRotations = (sbyte)FCAMMath.Min(info.BaseCPUs[0].MaxRotations + i, JEWEL_SET_SIZE.Rows - 1);
							sbyte depth = (sbyte)(info.BaseCPUs[0].Depth + i / 2);
							CPUDropType dropType = info.BaseCPUs[0].DropType;	//	(CPUDropType)(FCAMMath.Min<sbyte>((sbyte)CPUDropType.MAX, (sbyte)((sbyte)info.BaseCPUs[0].DropType + i / 2)));
							short precision = (short)(FCAMMath.Max(1, info.BaseCPUs[0].Precision - i * 2));
							short rough = (short)(FCAMMath.Max(0, info.BaseCPUs[0].Precision - i * 2));
							cpu = new CPUBrain(operateInterval,
											   maxRotations,
											   depth,
											   info.BaseCPUs[0].MultiplierSteps,
											   info.BaseCPUs[0].RecvDamagePoints,
											   info.BaseCPUs[0].SendDamagePoints,
											   info.BaseCPUs[0].NoSendDamagePoints,
											   dropType,
											   precision,
											   rough,
											   info.BaseCPUs[0].DamageJewelPoints,
											   info.BaseCPUs[0].HardDamageJewelPoints,
											   info.BaseCPUs[0].DeletePoints,
											   info.BaseCPUs[0].HVPairPoints,
											   info.BaseCPUs[0].DiagonallyPairPoints,
											   info.BaseCPUs[0].EnergyLevelScore,
											   info.BaseCPUs[0].OpponentCPUScoreDivide,
											   info.BaseCPUs[0].ScoreBoard);
						}else{
							cpu = info.BaseCPUs[i];
						}
						Infos[chara, i] = new CharacterInfo(info.Name,
#if ATTACK_MAGIC
															info.AttackMagic,
#endif
															info.SupportMagic,
															cpu);
					}
				}
			}
			public static readonly CharacterInfo[,] Infos;
			public static readonly CharacterInfo[] CharacterBase = new CharacterInfo[]{
				new CharacterInfo("<YOU>",
#if ATTACK_MAGIC
					new AttackMagicOfProtagonist(),
#endif
					new SupportMagicOfProtagonist(),
					new CPUBrain(0, 2, 2, 3, 1, 1, 0, CPUDropType.Instant, 10, 10, -30, -200, 20, 40, 60, 100, 10,
						/*new int[, ]{
							{ 0000,  0000,  0000, -9999,  0000,  0000,  0000},
							{ 0000,  0000,  0000, -9999,  0000,  0000,  0000},
							{ 0000,  0000,  0000, -9999,  0000,  0000,  0000},
							{ 0000,  0000,  0000, -5000,  0000,  0000,  0000},
							{ 0000,  0000,  0000, -5000,  0000,  0000,  0000},
							{ 0000,  0000,  0000, -5000,  0000,  0000,  0000},
							{ 0000,  0000,  0000,  0000,  0000,  0000,  0000},
							{ 0000,  0000,  0000,  0000,  0000,  0000,  0000},
							{ 0000,  0000,  0000,  0000,  0000,  0000,  0000},
							{ 0000,  0000,  0000,  0000,  0000,  0000,  0000},
							{ 0000,  0000,  0000,  0000,  0000,  0000,  0000},
							{ 0000,  0000,  0000,  0000,  0000,  0000,  0000},
							{ 0000,  0000,  0000,  0000,  0000,  0000,  0000},
							{ 0000,  0000,  0000,  0000,  0000,  0000,  0000},
							{ 0000,  0000,  0000,  0000,  0000,  0000,  0000},
						}*/
						new CPUScoreBoard(FIELD_TABLESIZE, -500, -9999, -650, -5000, -10, -30, 10)
					)
				),
				//	MaxRotation が負の値なのは、Difficult により加算されることを想定し、Difficult が高くても回転しないよう
				new CharacterInfo("BURNAT",
#if ATTACK_MAGIC
					new AttackMagicOfBurnat(),
#endif
					new SupportMagicOfBurnat(),
					new CPUBrain(
						operateInterval: 40,
						maxRotations: -99,
						depth: 1,
						multiplierSteps: 3,
						recvDamagePoints: 0,
						sendDamagePoints: 0,
						noSendDamagePoints: 30,
						dropType: CPUDropType.None,
						precision: 10,
						rough: 10,
						damageJewelPoints: -30,
						hardDamageJewelPoints: -150,
						deletePoints: 20,
						hvPairPoints: 40,
						diagonallyPairPoints: 60,
						energyLevelScore: 100,
						opponentCPUScoreDivide: 10,
						scoreBoard: new CPUScoreBoard(FIELD_TABLESIZE, -1000, -9999, -1000, -3000, 0, 0, 0)
					)
				),
				new CharacterInfo("FLUIP",
#if ATTACK_MAGIC
					new AttackMagicOfFluip(),
#endif
					new SupportMagicOfFluip(),
					new CPUBrain(
						operateInterval: 36,
						maxRotations: -1,
						depth: 1,
						multiplierSteps: 7,
						recvDamagePoints: -1,
						sendDamagePoints: 1,
						noSendDamagePoints: 15,
						dropType: CPUDropType.None,
						precision: 20,
						rough: 40,
						damageJewelPoints: -100,
						hardDamageJewelPoints: -500,
						deletePoints: 100,
						hvPairPoints: 10,
						diagonallyPairPoints: 50,
						energyLevelScore: 100,
						opponentCPUScoreDivide: 10,
						scoreBoard: new CPUScoreBoard(FIELD_TABLESIZE, 140, -9999, 140, -1000, 140, -100, 140)
					)
				),
				new CharacterInfo("SCAPPION",
#if ATTACK_MAGIC
					new AttackMagicOfScappion(),
#endif
					new SupportMagicOfScappion(),
					new CPUBrain(
						operateInterval: 32,
						maxRotations: 0,
						depth: 2,
						multiplierSteps: 4,
						recvDamagePoints: -1,
						sendDamagePoints: 1,
						noSendDamagePoints: 1,
						dropType: CPUDropType.Normal,
						precision: 5,
						rough: 20,
						damageJewelPoints: -50,
						hardDamageJewelPoints: -300,
						deletePoints: 50,
						hvPairPoints: 10,
						diagonallyPairPoints: 20,
						energyLevelScore: 100,
						opponentCPUScoreDivide: 10,
						cpuScore: new int[, ]{
							{-0000, -0500, -1000,CPUScoreBoard.DEAD_AREA_POINT , -0500, -0250,  0099},
							{-0000, -0500, -1000,CPUScoreBoard.DEAD_AREA_POINT, -0500, -0250,  0099},
							{-0000, -0500, -1000,CPUScoreBoard.DEAD_AREA_POINT, -0500, -0250,  0099},
							{-0000, -0250, -0500, -5000, -0250,  0000,  0099},
							{-0000, -0250, -0500, -5000, -0250,  0000,  0099},
							{-0000, -0250, -0500, -5000, -0250,  0000,  0099},
							{-0000, -0200, -0100, -1000, -0100,  0000,  0099},
							{-0000, -0200, -0100, -1000, -0100,  0000,  0099},
							{-0000, -0200, -0100, -1000, -0100,  0000,  0099},
							{-0000, -0200, -0100, -0500, -0100,  0000,  0099},
							{-0000, -0200, -0100, -0500, -0100,  0000,  0099},
							{-0000, -0200, -0100, -0500, -0100,  0000,  0099},
							{-0000, -0200, -0100, -0100, -0100,  0000,  0099},
							{-0000, -0200, -0100, -0100, -0100,  0000,  0099},
							{-0000, -0200, -0100, -0100, -0100,  0000,  0099},
						}
					)
				),
				new CharacterInfo("SMOW",
#if ATTACK_MAGIC
					new AttackMagicOfSmow(),
#endif
					new SupportMagicOfSmow(),
					new CPUBrain(
						operateInterval: 30,
						maxRotations: 0,
						depth: 1,
						multiplierSteps: 7,
						recvDamagePoints: -10,
						sendDamagePoints: 5,
						noSendDamagePoints: -5,
						dropType: CPUDropType.Normal,
						precision: 30,
						rough: 50,
						damageJewelPoints: -50,
						hardDamageJewelPoints: -100,
						deletePoints: 200,
						hvPairPoints: 50,
						diagonallyPairPoints: 50,
						energyLevelScore: 100,
						opponentCPUScoreDivide: 10,
						scoreBoard: new CPUScoreBoard(FIELD_TABLESIZE, 140, -9999, 140, -1500, 140, -100, 140)
					)
				),
				new CharacterInfo("HYDEAD",
#if ATTACK_MAGIC
					new AttackMagicOfHydead(),
#endif
					new SupportMagicOfHydead(), 
					new CPUBrain(
						operateInterval: 28,
						maxRotations: 1,
						depth: 1,
						multiplierSteps: 2,
						recvDamagePoints: -5,
						sendDamagePoints: 10,
						noSendDamagePoints: -0,
						dropType: CPUDropType.Normal,
						precision: 20,
						rough: 30,
						damageJewelPoints: -30,
						hardDamageJewelPoints: -50,
						deletePoints: 100,
						hvPairPoints: 30,
						diagonallyPairPoints: 40,
						energyLevelScore: 100,
						opponentCPUScoreDivide: 20,
						scoreBoard: new CPUScoreBoard(FIELD_TABLESIZE, -1000, -9999, 9999, -1000, 0, 0, 9999)
					)
				),
				new CharacterInfo("REMILSTEID",
#if ATTACK_MAGIC
					new AttackMagicOfRemilsteid(),
#endif
					new SupportMagicOfRemilsteid(),
					new CPUBrain(
						operateInterval: 26,
						maxRotations: 1,
						depth: 1,
						multiplierSteps: 5,
						recvDamagePoints: -30,
						sendDamagePoints: 30,
						noSendDamagePoints: -15,
						dropType: CPUDropType.Normal,
						precision: 30,
						rough: 40,
						damageJewelPoints: -50,
						hardDamageJewelPoints: -250,
						deletePoints: 70,
						hvPairPoints: 20,
						diagonallyPairPoints: 30,
						energyLevelScore: 100,
						opponentCPUScoreDivide: 20,
						scoreBoard: new CPUScoreBoard(FIELD_TABLESIZE, -3000, -9999, -3000, -1000, 0, 0, 0)
					)
				),
				new CharacterInfo("SHADOWS",
#if ATTACK_MAGIC
					new AttackMagicOfProtagonist(),
#endif
					new SupportMagicOfProtagonist(),
					new CPUBrain(
						operateInterval: 24,
						maxRotations: 1,
						depth: 1,
						multiplierSteps: 8,
						recvDamagePoints: -60,
						sendDamagePoints: 60,
						noSendDamagePoints: -20,
						dropType: CPUDropType.Normal,
						precision: 25,
						rough: 30,
						damageJewelPoints: -100,
						hardDamageJewelPoints: -250,
						deletePoints: 150,
						hvPairPoints: 50,
						diagonallyPairPoints: 100,
						energyLevelScore: 200,
						opponentCPUScoreDivide: 15,
						scoreBoard: new CPUScoreBoard(FIELD_TABLESIZE, 9999, -9999, 9999, -1000, 9999, 0, 9999)
					)
				),
				new CharacterInfo("REMILSTEID(RAGE)",
#if ATTACK_MAGIC
					new AttackMagicOfRemilsteid(),
#endif
					new SupportMagicOfRemilsteid(),
					new CPUBrain(
						operateInterval: 22,
						maxRotations: 2,
						depth: 2,
						multiplierSteps: 4,
						recvDamagePoints: -50,
						sendDamagePoints: 50,
						noSendDamagePoints: -30,
						dropType: CPUDropType.Normal,
						precision: 20,
						rough: 20,
						damageJewelPoints: -100,
						hardDamageJewelPoints: -300,
						deletePoints: 80,
						hvPairPoints: 30,
						diagonallyPairPoints: 50,
						energyLevelScore: 100,
						opponentCPUScoreDivide: 10,
						cpuScore: new int[, ]{
							{ 2000,  1000,  0000,CPUScoreBoard.DEAD_AREA_POINT,  0000,  1000,  2000},
							{ 0700,  0100,  0100,CPUScoreBoard.DEAD_AREA_POINT,  0100,  0100,  0700},
							{ 0800,  0200,  0150,CPUScoreBoard.DEAD_AREA_POINT,  0150,  0200,  0800},
							{ 0900,  0350,  0200,                         0000,  0200,  0300,  0900},
							{ 1000,  0450,  0250,                         0000,  0250,  0400,  1000},
							{ 1100,  0550,  0300,                         0000,  0300,  0500,  1100},
							{ 1200,  0600,  0350,                         0000,  0350,  0600,  1200},
							{ 1300,  0700,  0400,                         0000,  0400,  0700,  1300},
							{ 1400,  0800,  0450,                         0000,  0450,  0800,  1400},
							{ 1500,  0900,  0500,                         0000,  0500,  0900,  1500},
							{ 1600,  1000,  0600,                         0000,  0600,  1000,  1600},
							{ 1700,  1200,  0700,                         0000,  0700,  1200,  1700},
							{ 1800,  1400,  0800,                         0000,  0800,  1400,  1800},
							{ 1900,  1600,  0900,                         0000,  0900,  1600,  1900},
							{ 2000,  1800,  1000,                         0000,  1000,  1800,  2000},
						}
					)
				),
				//	ここら辺の奴は、もし自動CPU設定の場合 depth が自動で +1 される
				//	⇒HARD以上でしか存在しない場合、大概のものは +1 される？
				//	逆に言えば、自動でCPU設定しない場合は自動計算との比較を少なからず脳内でざっくばらんに行なっておくこと。
				new CharacterInfo("SECRET 1",
#if ATTACK_MAGIC
					new AttackMagicOfBurnat(),
#endif
					new SupportMagicOfBurnat(),
					null,
					null,
					new CPUBrain(
						operateInterval: 2,
						maxRotations: 2,
						depth: 3,
						multiplierSteps: 10,
						recvDamagePoints: -0,
						sendDamagePoints: 0,
						noSendDamagePoints: -50,
						dropType: CPUDropType.Normal,
						precision: 40,
						rough: 70,
						damageJewelPoints: -50,
						hardDamageJewelPoints: -100,
						deletePoints: 150,
						hvPairPoints: 20,
						diagonallyPairPoints: 40,
						energyLevelScore: 100,
						opponentCPUScoreDivide: 20,
						scoreBoard: new CPUScoreBoard(FIELD_TABLESIZE, -1000, -9999, -1000, -3000, 0, 0, 0)
					),
					new CPUBrain(
						operateInterval: 4,
						maxRotations: 2,
						depth: 4,
						multiplierSteps: 10,
						recvDamagePoints: -0,
						sendDamagePoints: 0,
						noSendDamagePoints: -50,
						dropType: CPUDropType.Instant,
						precision: 40,
						rough: 70,
						damageJewelPoints: -50,
						hardDamageJewelPoints: -100,
						deletePoints: 150,
						hvPairPoints: 20,
						diagonallyPairPoints: 40,
						energyLevelScore: 100,
						opponentCPUScoreDivide: 20,
						scoreBoard: new CPUScoreBoard(FIELD_TABLESIZE, -1000, -9999, -1000, -3000, 0, 0, 0)
					)
				),
				new CharacterInfo("SECRET 2",
#if ATTACK_MAGIC
					new AttackMagicOfSmow(),
#endif
					new SupportMagicOfSmow(),
					null,
					null,
					new CPUBrain(
						operateInterval: 2,
						maxRotations: 2,
						depth: 4,
						multiplierSteps: 5,
						recvDamagePoints: -100,
						sendDamagePoints: 100,
						noSendDamagePoints: -100,
						dropType: CPUDropType.Normal,
						precision: 30,
						rough: 50,
						damageJewelPoints: -150,
						hardDamageJewelPoints: -200,
						deletePoints: 100,
						hvPairPoints: 50,
						diagonallyPairPoints: 70,
						energyLevelScore: 100,
						opponentCPUScoreDivide: 10,
						scoreBoard: new CPUScoreBoard(FIELD_TABLESIZE, 140, -9999, 140, -1500, 140, -100, 140)
					),
					new CPUBrain(
						operateInterval: 2,
						maxRotations: 2,
						depth: 5,
						multiplierSteps: 5,
						recvDamagePoints: -100,
						sendDamagePoints: 100,
						noSendDamagePoints: -100,
						dropType: CPUDropType.Instant,
						precision: 30,
						rough: 50,
						damageJewelPoints: -150,
						hardDamageJewelPoints: -200,
						deletePoints: 100,
						hvPairPoints: 50,
						diagonallyPairPoints: 70,
						energyLevelScore: 100,
						opponentCPUScoreDivide: 10,
						scoreBoard: new CPUScoreBoard(FIELD_TABLESIZE, 140, -9999, 140, -1500, 140, -100, 140)
					)
				),
				new CharacterInfo("INIF",
#if ATTACK_MAGIC
					new AttackMagicOfRemilsteid(),
#endif
					new SupportMagicOfRemilsteid(),
					null,
					null,
					new CPUBrain(
						operateInterval: 1,
						maxRotations: 2,
						depth: 5,
						multiplierSteps: 4,
						recvDamagePoints: -30,
						sendDamagePoints: 30,
						noSendDamagePoints: -50,
						dropType: CPUDropType.Normal,
						precision: 20,
						rough: 30,
						damageJewelPoints: -100,
						hardDamageJewelPoints: -300,
						deletePoints: 80,
						hvPairPoints: 30,
						diagonallyPairPoints: 50,
						energyLevelScore: 100,
						opponentCPUScoreDivide: 5,
						cpuScore: new int[, ]{
							{ 2000,  1000,  0000,CPUScoreBoard.DEAD_AREA_POINT,  0000,  1000,  2000},
							{ 0700,  0100,  0100,CPUScoreBoard.DEAD_AREA_POINT,  0100,  0100,  0700},
							{ 0800,  0200,  0150,CPUScoreBoard.DEAD_AREA_POINT,  0150,  0200,  0800},
							{ 0900,  0350,  0200,                         0000,  0200,  0300,  0900},
							{ 1000,  0450,  0250,                         0000,  0250,  0400,  1000},
							{ 1100,  0550,  0300,                         0000,  0300,  0500,  1100},
							{ 1200,  0600,  0350,                         0000,  0350,  0600,  1200},
							{ 1300,  0700,  0400,                         0000,  0400,  0700,  1300},
							{ 1400,  0800,  0450,                         0000,  0450,  0800,  1400},
							{ 1500,  0900,  0500,                         0000,  0500,  0900,  1500},
							{ 1600,  1000,  0600,                         0000,  0600,  1000,  1600},
							{ 1700,  1200,  0700,                         0000,  0700,  1200,  1700},
							{ 1800,  1400,  0800,                         0000,  0800,  1400,  1800},
							{ 1900,  1600,  0900,                         0000,  0900,  1600,  1900},
							{ 2000,  1800,  1000,                         0000,  1000,  1800,  2000},
						}
					),
					new CPUBrain(
						operateInterval: 0,
						maxRotations: 2,
						depth: 10,
						multiplierSteps: 4,
						recvDamagePoints: -30,
						sendDamagePoints: 30,
						noSendDamagePoints: -50,
						dropType: CPUDropType.Instant,
						precision: 20,
						rough: 30,
						damageJewelPoints: -100,
						hardDamageJewelPoints: -300,
						deletePoints: 80,
						hvPairPoints: 30,
						diagonallyPairPoints: 50,
						energyLevelScore: 100,
						opponentCPUScoreDivide: 5,
						cpuScore: new int[, ]{
							{ 2000,  1000,  0000,CPUScoreBoard.DEAD_AREA_POINT,  0000,  1000,  2000},
							{ 0700,  0100,  0100,CPUScoreBoard.DEAD_AREA_POINT,  0100,  0100,  0700},
							{ 0800,  0200,  0150,CPUScoreBoard.DEAD_AREA_POINT,  0150,  0200,  0800},
							{ 0900,  0350,  0200,                         0000,  0200,  0300,  0900},
							{ 1000,  0450,  0250,                         0000,  0250,  0400,  1000},
							{ 1100,  0550,  0300,                         0000,  0300,  0500,  1100},
							{ 1200,  0600,  0350,                         0000,  0350,  0600,  1200},
							{ 1300,  0700,  0400,                         0000,  0400,  0700,  1300},
							{ 1400,  0800,  0450,                         0000,  0450,  0800,  1400},
							{ 1500,  0900,  0500,                         0000,  0500,  0900,  1500},
							{ 1600,  1000,  0600,                         0000,  0600,  1000,  1600},
							{ 1700,  1200,  0700,                         0000,  0700,  1200,  1700},
							{ 1800,  1400,  0800,                         0000,  0800,  1400,  1800},
							{ 1900,  1600,  0900,                         0000,  0900,  1600,  1900},
							{ 2000,  1800,  1000,                         0000,  1000,  1800,  2000},
						}
					)
				),
			};
			private CPUBrain[] BaseCPUs;
			public string Name;
#if ATTACK_MAGIC
			public MagicProcessor AttackMagic
#endif
			public MagicProcessor SupportMagic;
			public CPUBrain CPU{get{return(BaseCPUs[0]);}}
			/// <summary>
			/// キャラクター情報を作成。
			/// private であるのは、計算用に baseCPUs 引数があるため。
			/// これはキャラクターベース情報用のやつ（のちに静的コンストラクタで計算を行なうのに使う。）
			/// </summary>
			/// <param name="name"></param>
			/// <param name="attackMagic"></param>
			/// <param name="supportMagic"></param>
			/// <param name="cpus">必ず1個は指定。指定しない場合は自動生成。なお自動生成不要の場合は null を指定。</param>
			private CharacterInfo(string name,
#if ATTACK_MAGIC
								  MagicProcessor attackMagic,
#endif
								  MagicProcessor supportMagic,
								  params CPUBrain[] baseCPUs){
				Name = name;
				BaseCPUs = baseCPUs;
#if ATTACK_MAGIC
				AttackMagic = attackMagic;
#endif
				SupportMagic = supportMagic;
			}
		}
	}
}
