﻿using FukuracamDX;
using System.Collections.Generic;

namespace FatopTheWorld.Scenes{
	partial class Battle: FCAMScene{
#if ATTACK_MAGIC
		/// <summary>
		/// 攻撃：主人公
		/// 敵の動きを15秒間止める
		/// </summary>
		class AttackMagicOfProtagonist: MagicProcessor{
			private const float SECONDS = 15.0f;
			private const short INTERVAL = 15;
			static private readonly FCAMRectangle POSIZE_EFFECT = new FCAMRectangle(false, 0, 0, FIELD_SIZE.Width, FIELD_SIZE.Height);
			private int NumberFrames{get{return((int)(Game.Instance.FPS * SECONDS));}}
			public override long ProcAndGetResult(ref PlayerField casterField, ref FCAMList<JewelSet> casterJewelSetList, Player caster, int? calcColumn, params Player[] targets){
				if(targets.Length == 0){
					//	敵が居ない時は魔法の意味がないので、使わせない。
					return(int.MinValue);
				}
				long sum = 0;
				foreach(Player target in targets){
					List<Player> targetTargetList = new List<Player>(targets);
					targetTargetList.Add(caster);
					targetTargetList.Remove(target);
					//	敵が魔法を唱えた際の得点を計算
					sum += target.CPU.GetAim(false, target, targetTargetList.ToArray()).CPUScore / target.CPU.OpponentCPUScoreDivide / targets.Length;
				}
				return(sum);
			}
			protected override void Load(){
				State = MagicState.BeforeAftereffect;
			}
			protected override void Calc(){
				if(Frame >= NumberFrames){
					State = MagicState.BeforeEnd;
				}
			}
			protected override void Draw(FCAMGraphics g){
				int alpha;
				FCAMColor blender;
				alpha = Frame % (INTERVAL * 2);
				if(alpha >= INTERVAL){
					alpha = INTERVAL * 2 - alpha;
				}
				blender = new FCAMColor(FCAMMath.Within(FCAMColor.MinValue, FCAMColor.MaxValue * alpha / INTERVAL, FCAMColor.MaxValue), FCAMColor.MaxValue);
				foreach(Player player in Targets){
					g.Draw(Texture, player.PosVisibleField.X, player.PosVisibleField.Y, POSIZE_EFFECT.Width, POSIZE_EFFECT.Height, POSIZE_EFFECT.X, POSIZE_EFFECT.Y, blender);
				}
			}
		}
#endif
		/// <summary>
		/// 補助：主人公
		/// マジックジュエルを出現させる。
		/// </summary>
		class SupportMagicOfProtagonist: MagicProcessor{
			//	間違った状態でのスコア。
			private const int MISS_SCORE = CPUScoreBoard.DEAD_AREA_POINT;
			public override MagicProcessor Begin(FCAMTexture texture, Player caster, params Player[] targets){
				//	敵には関係のないサポート魔法のため、ターゲットリストは排除する。
				return(base.Begin(texture, caster));
			}
			public override long ProcAndGetResult(ref PlayerField casterField, ref FCAMList<JewelSet> casterJewelSetList, Player caster, int? calcColumn, params Player[] targets){
				for(int row = 0; row < JEWEL_SET_SIZE.Rows; row++){
					for(int column = 0; column < JEWEL_SET_SIZE.Columns; column++){
						casterJewelSetList[0].Jewels[column, row].Kind = MAGIC_JEWEL;
					}
				}
				if(calcColumn.HasValue){
					for(int row = 0; row < caster.Field.Size.Rows; row++){
						sbyte kind = caster.Field.Data[calcColumn.Value, row].Kind;
						if(kind == 0){
							continue;
						}
						if(Battle.IsJewelNormal(kind)){
							//	下が通常ジュエルなら、普通に考える。
							break;
						}
						if(Battle.IsJewelObstract(kind)){
							//	もし落下先がお邪魔ジュエルなら、マジックジュエルの効果が発揮されないためスコアを下げる。
							return(MISS_SCORE * caster.NowJewelSet.Size.Rows);
						}
					}
					//	もしピンチ状態なら、是が非でも使う。
					if(caster.Field.IsPinch(caster.NowJewelSet)){
						return(caster.CPU.ForSupportMagicPoint);
					}
				}
				//	通常の考えなら、魔法なので極力使用したくない
				if(caster.EnergyLevel >= 2){
					return(-1 * caster.CPU.ForSupportMagicPoint);
				}else{
					return(-3 * caster.CPU.ForSupportMagicPoint);
				}
			}
			protected override void Load(){
				ProcAndGetResult(ref Caster.Field, ref Caster.JewelSetList, Caster, null, Targets);
				FCAMRectangle posize = Caster.GetNowJewelSetPosize();
				Caster.Parent.FadeoutDiagramList.Add(new FadeoutDiagram(new FCAMRectangle(false, posize.X, posize.Y, posize.Width, posize.Height), FCAMColor.White, DiagramKind.Rectangle));
				//	マジックジュエル配置後にチェック処理を行なわせるため
				State = MagicState.BeforeAftereffect;
			}
			protected override void Calc(){
				for(int row = 0; row < JEWEL_SET_SIZE.Rows; row++){
					for(int column = 0; column < JEWEL_SET_SIZE.Columns; column++){
						if(Caster.NowJewelSet.Jewels[column, row].Kind == MAGIC_JEWEL)
							return;
					}
				}
				State = MagicState.BeforeEnd;
			}
		}
#if ATTACK_MAGIC
		/// <summary>
		/// 攻撃：バーナット
		/// 敵のフィールド上のジュエルの一部を燃やして超お邪魔ジュエル化させる
		/// </summary>
		class AttackMagicOfBurnat: MagicProcessor{
			private class DeleteInfo{
				public FCAMTablePoint Point;
				public int Frame;
				public DeleteInfo(int column, int row){
					Point = new FCAMTablePoint(column, row);
					Frame = 0;
				}
			}
			private const short CONVERT_RATES = 20;
			private const float SECONDS = 3.0f;
			private int NumberFrames{get{return((int)(Game.Instance.FPS * SECONDS));}}
			private List<DeleteInfo>[] DeleteInfoLists;
			/// <inheritdoc/>
			public override long ProcAndGetResult(ref PlayerField casterField, ref FCAMList<JewelSet> casterJewelSetList, Player caster, int? calcColumn, params Player[] targets){
				if(targets.Length == 0){
					//	敵が居ない時は魔法の意味がないので、使わせない。
					return(int.MinValue);
				}
				long sum = 0;
				foreach(Player target in targets){
					//	フィールドの1/3は最低でも埋まっていた方が魔法の効果があるように思う
					sum += ((target.Field.GetNumberNormalJewels() - target.Field.GetNumberObstractJewels() * 2) - FIELD_TABLESIZE.Columns * FIELD_TABLESIZE.Rows * 3 ) * target.CPU.ForAttackMagicPoint / targets.Length;
				}
				return(sum);
			}
			protected override void Load(){
				DeleteInfoLists = new List<DeleteInfo>[Targets.Length];
				for(int i = 0; i < Targets.Length; i++){
					Player player = Targets[i];
					DeleteInfoLists[i] = new List<DeleteInfo>();
					for(int row = 0; row < player.Field.Size.Rows; row++){
						for(int column = 0; column < player.Field.Size.Columns; column++){
							if(IsJewelNormal(player.Field.Data[column, row].Kind)){
								if(FCAMRandom.Common.NextInt(100) < CONVERT_RATES){
									DeleteInfoLists[i].Add(new DeleteInfo(column, row));
								}
							}
						}
					}
				}
			}
			protected override void Unload(){
				for(int i = 0; i < Targets.Length; i++){
					Player player = Targets[i];
					foreach(DeleteInfo info in DeleteInfoLists[i]){
						player.Field.Data[info.Point.Column, info.Point.Row].Put(HARDOBSTACLE_JEWEL);
					}
				}
				DeleteInfoLists = null;
			}
			protected override void Calc(){
				if(Frame >= NumberFrames){
					State = MagicState.BeforeEnd;
					return;
				}
				int alpha = FCAMMath.Max(0, FCAMColor.MaxValue - FCAMColor.MaxValue * Frame / (NumberFrames / 2));
				FCAMColor blender = new FCAMColor(FCAMColor.MaxValue, FCAMColor.MaxValue, alpha, alpha);
				for(int i = 0; i < Targets.Length; i++){
					Player player = Targets[i];
					foreach(DeleteInfo info in DeleteInfoLists[i]){
						player.Field.Data[info.Point.Column, info.Point.Row].Blender = blender;
						info.Frame++;
					}
				}
			}
			protected override void Draw(FCAMGraphics g){
				for(int i = 0; i < Targets.Length; i++){
					Player player = Targets[i];
					foreach(DeleteInfo info in DeleteInfoLists[i]){
						JewelData jewel = player.Field.Data[info.Point.Column, info.Point.Row];
						FCAMPoint pos = player.GetJewelPos(info.Point.Column, info.Point.Row);
						FCAMRectangle rectangle = new FCAMRectangle(pos, JEWEL_SIZE);
						rectangle.Inflate(info.Frame / 4);
						g.FillRotatedRectangleAtOrigin(new FCAMVector(0, 0, info.Frame * 32), null, rectangle.Center.X, rectangle.Center.Y, rectangle.Width, rectangle.Height, FCAMColor.HalfRed);
					}
				}
			}
		}
#endif
		/// <summary>
		/// 補助：バーナット
		/// 主人公同じ
		/// </summary>
		class SupportMagicOfBurnat: SupportMagicOfProtagonist{
			public override MagicProcessor Begin(FCAMTexture texture, Player caster, params Player[] targets){
				//	敵には関係のないサポート魔法のため、ターゲットリストは排除する。
				return(base.Begin(texture, caster));
			}
		}
#if ATTACK_MAGIC
		/// <summary>
		/// 攻撃：フルィプ
		/// フィールドの下3段を水で洗い流す
		/// </summary>
		class AttackMagicOfFluip: MagicProcessor{
			private const sbyte NUMBER_POP_LINES = 3;
			public override MagicProcessor Begin(FCAMTexture texture, Player caster, params Player[] targets){
				//	敵には関係の魔法のため、ターゲットリストは排除する。
				return(base.Begin(texture, caster));
			}
			public override long ProcAndGetResult(ref PlayerField casterField, ref FCAMList<JewelSet> casterJewelSetList, Player caster, int? calcColumn, params Player[] targets){
				for(int row = Caster.Field.Size.Rows - 1; row >= Caster.Field.Size.Rows - NUMBER_POP_LINES; row--){
					for(int column = 0; column < Caster.Field.Size.Columns; column++){
						casterField.Data[column, row].Delete();
					}
				}
				return(0);
			}
		}
#endif
		/// <summary>
		/// 補助：フルィプ
		/// 主人公同じ
		/// </summary>
		class SupportMagicOfFluip: SupportMagicOfProtagonist{
			public override MagicProcessor Begin(FCAMTexture texture, Player caster, params Player[] targets){
				//	敵には関係のないサポート魔法のため、ターゲットリストは排除する。
				return(base.Begin(texture, caster));
			}
		}
#if ATTACK_MAGIC
		/// <summary>
		/// 攻撃：スキャッピオン
		/// 敵のフィールドをおかしくする？
		/// </summary>
	#warning このクラスは未完成です！
		class AttackMagicOfScappion: MagicProcessor{
			public override long ProcAndGetResult(ref PlayerField casterField, ref FCAMList<JewelSet> casterJewelSetList, Player caster, int? calcColumn, params Player[] targets){
				return int.MinValue;
				long sum = 0;
				if(Targets.Length > 0){
					foreach(Player target in targets){
						List<Player> targetTargetList = new List<Player>(targets);
						targetTargetList.Add(caster);
						targetTargetList.Remove(target);
						sum += target.CPU.GetAim(false, target, targetTargetList.ToArray()).CPUScore / target.CPU.OpponentCPUScoreDivide;
					}
					sum /= Targets.Length;
				}
				return(sum);
			}
		}
#endif
		/// <summary>
		/// 補助：スキャッピオン
		/// 主人公同じ
		/// </summary>
		class SupportMagicOfScappion: SupportMagicOfProtagonist{
			public override MagicProcessor Begin(FCAMTexture texture, Player caster, params Player[] targets){
				//	敵には関係のないサポート魔法のため、ターゲットリストは排除する。
				return(base.Begin(texture, caster));
			}
		}
#if ATTACK_MAGIC
		/// <summary>
		/// 攻撃：スモー
		/// ？
		/// </summary>
	#warning このクラスは未完成です！
		class AttackMagicOfSmow: MagicProcessor{
			public override long ProcAndGetResult(ref PlayerField casterField, ref FCAMList<JewelSet> casterJewelSetList, Player caster, int? calcColumn, params Player[] targets){
				return(int.MinValue);
			}
		}
#endif
		/// <summary>
		/// 補助：スモー
		/// 主人公同じ
		/// </summary>
		class SupportMagicOfSmow: SupportMagicOfProtagonist{
			public override MagicProcessor Begin(FCAMTexture texture, Player caster, params Player[] targets){
				//	敵には関係のないサポート魔法のため、ターゲットリストは排除する。
				return(base.Begin(texture, caster));
			}
		}
#if ATTACK_MAGIC
		/// <summary>
		/// 攻撃：ハイデッド
		/// ？
		/// </summary>
	#warning このクラスは未完成です！
		class AttackMagicOfHydead: MagicProcessor{
			public override long ProcAndGetResult(ref PlayerField casterField, ref FCAMList<JewelSet> casterJewelSetList, Player caster, int? calcColumn, params Player[] targets){
				return(int.MinValue);
			}
		}
#endif
		/// <summary>
		/// 補助：ハイデッド
		/// 主人公同じ
		/// </summary>
		class SupportMagicOfHydead: SupportMagicOfProtagonist{
			public override MagicProcessor Begin(FCAMTexture texture, Player caster, params Player[] targets){
				//	敵には関係のないサポート魔法のため、ターゲットリストは排除する。
				return(base.Begin(texture, caster));
			}
		}
#if ATTACK_MAGIC
		/// <summary>
		/// 攻撃：レミルスティッド
		/// ？
		/// </summary>
	#warning このクラスは未完成です！
		class AttackMagicOfRemilsteid: MagicProcessor{
			public override long ProcAndGetResult(ref PlayerField casterField, ref FCAMList<JewelSet> casterJewelSetList, Player caster, int? calcColumn, params Player[] targets){
				return(int.MinValue);
			}
		}
#endif
		/// <summary>
		/// 補助：レミルスティッド
		/// 主人公同じ
		/// </summary>
		class SupportMagicOfRemilsteid: SupportMagicOfProtagonist{
			public override MagicProcessor Begin(FCAMTexture texture, Player caster, params Player[] targets){
				//	敵には関係のないサポート魔法のため、ターゲットリストは排除する。
				return(base.Begin(texture, caster));
			}
		}
		/// <summary>
		/// 
		/// </summary>
		enum MagicKind{
			Nothing,
			/// <summary>旧攻撃魔法。現キャラクター固有魔法</summary>
			Attack,
			Support,
			/// <summary>魔法使用後、この値にして目標地点を再計算するため用</summary>
			Used,
		}
		/// <summary>
		/// 
		/// </summary>
		enum MagicState{
			///	<summary>魔法による待機中</summary>
			Wait,
			///	<summary>待機後の余波的なもの（敵の動きを止めるとか）</summary>
			BeforeAftereffect,
			Aftereffect,
			///	<summary>この状態になると魔法が終了</summary>
			BeforeEnd,
			End,
		}
		/// <summary>
		/// 
		/// </summary>
		abstract class MagicProcessor{
			/// <summary>どんな状態の魔法からでも復帰出来るように</summary>
			protected PlayerState BeforePlayerState;
			protected MagicState State;
			protected Player Caster;
			protected Player[] Targets;
			protected FCAMTexture Texture;
			public int Frame;
			public virtual MagicProcessor Begin(FCAMTexture texture, Player caster, params Player[] targets){
				MagicProcessor magic = (MagicProcessor)GetType().GetConstructor(new System.Type[]{}).Invoke(new object[]{});
				magic.Caster = caster;
				magic.Targets = targets;
				magic.BeforePlayerState = caster.State;
				magic.Texture = texture;
				caster.State = PlayerState.CastMagic;
				State = MagicState.Wait;
				Frame = 0;
				foreach(Player player in targets){
					player.NumberReceivingMagics++;
				}
				magic.Load();
				return(magic);
			}
			public bool IsEffecting(){
				return(State != MagicState.End);
			}
			public virtual sbyte GetNeedLevel(){
				return(1);
			}
			/// <summary>
			/// true を返すと処理終了
			/// </summary>
			public void DoCalc(){
				Calc();
				Frame++;
				switch(State){
					case MagicState.BeforeAftereffect:{
						Caster.State = BeforePlayerState;
						State = MagicState.Aftereffect;
						break;
					}
					case MagicState.BeforeEnd:{
						Unload();
						Caster.State = BeforePlayerState;
						foreach(Player player in Targets){
							player.NumberReceivingMagics--;
						}
						State = MagicState.End;
						break;
					}
				}
			}
			public void DoDraw(FCAMGraphics g){
				Draw(g);
			}
			/// <summary>
			/// CPU 計算用で結果をここから推測させるため、
			/// 必ず SetResult をオーバーライドすること。
			/// またスコア計算用に、戻り値を設定すること。
			/// (※注意 Player のデータを変更するとそれが反映されるので
			///	書き込みは絶対にしないこと。∵これはあくまでもスコア計算用なので。)
			///	その代わり引数に、詠唱者の情報(Class系)がある(呼び出し元でコピーしたものなので、それを使って計算)
			/// </summary>
			/// <param name="calcColumn">現在計算中のカラム位置。なおProcも兼ねたメソッドのため、計算に使用しない場合は null を指定。</param>
			public abstract long ProcAndGetResult(ref PlayerField casterField, ref FCAMList<JewelSet> casterJewelSetList, Player caster, int? calcColumn, params Player[] targets);
			protected virtual void Load(){}
			protected virtual void Unload(){}
			protected virtual void Calc(){}
			protected virtual void Draw(FCAMGraphics g){}
		}
	}
}
