﻿using FukuracamDX;
using System.Collections.Generic;

namespace FatopTheWorld.Scenes{
	/// <summary>
	/// 
	/// </summary>
	partial class Battle: FCAMScene{
		delegate T Func<T, U>(U arg1, U arg2);
		delegate void DoToJewelDelegate(ref JewelData jewel);
		enum DiagramKind{
			Rectangle,
			FillRectangle,
		}
		/// <summary>
		/// バトルのモード
		/// </summary>
		public enum BattleMode{
			Story, VS, Free,
			//	フリーモードで 1P がベース（すぐに Free に変更される）
			Free1P,
			//	フリーモードで 2P がベース（すぐに Free に変更される）
			Free2P,
			//	1P Player, 2P CPU でのテスト
			Test,
			//	1P 2P 共に CPU でのテスト
			CPUBattleTest,
			//	1P CPU、2P 空 でのテスト
			CPUTest,
		}
		/// <summary>
		/// プレイヤーの種類
		/// </summary>
		public enum PlayerKind{
			MIN = 0,
			/// <summary>操作</summary>
			Human = MIN,
			/// <summary>CPU</summary>
			CPU,
			/// <summary>非プレイヤー（誰もいない）</summary>
			NonParticipant,
			/// <summary></summary>
			MAX = NonParticipant,
		}
		/// <summary>
		/// ジュエルの種類
		/// </summary>
		public enum JewelKind{
			Normal, Damage, HardDamage
		}
		/// <summary>
		/// プレイヤーの状態。
		/// </summary>
		enum PlayerState{
			Normal, Wait, Chain, Check, MoveLeft, MoveRight, Win, Lose, Draw, CastMagic,
			//	魔法を受ける状態は特殊
			//	複数から受ける可能性もあるため、インクリメント方式の値を Player に持たせ
			//	それが >0 の時にこの値とする。
			ReceiveMagic
		}
		/// <summary>
		///	バトルの状態。
		/// </summary>
		enum BattleState{
			Menu, BattleStart, BattleNow, BattleBeforeEnd, BattleEnd, Pause, SceneEnd
		}
		/// <summary>
		///	絶対に構造体でなくてはならない！　∵PlayerField にて Clone してるから。
		/// </summary>
		struct JewelData{
			/// <summary>
			/// 位置のずれ
			/// </summary>
			public FCAMPoint PosGap;
			public sbyte Kind;
			public sbyte Frame;
			public FCAMColor Blender;
			/// <summary>
			/// 落下速度。始めは定数でも良かったが、ダメージを与えられた時のお邪魔の落下速度と通常ジュエルの落下速度が異なるため、
			/// いざお邪魔と通常ジュエルが接地したあとからの落下では、お互いの落下速度の違いで動作がおかしくなったため。
			/// 使い方としてはデフォルトのまんまで行ない、お邪魔に関しては接地後、落下速度を通常のジュエルと同じにする。
			/// </summary>
			private int __FallSpeed;
			public int FallSpeed{
				get{
					return __FallSpeed <= 0 ? JEWEL_SIZE.Height : __FallSpeed;
				}
				set{
					__FallSpeed = value;
				}
			}
			public JewelData(sbyte kind){
				Kind = kind;
				Frame = 0;
				PosGap = new FCAMPoint();
				Blender = FCAMColor.White;
				__FallSpeed = GetJewelFallSpeed(kind);
			}
			/// <summary>
			/// PosGap も反映させた位置を取得
			/// </summary>
			public FCAMPoint GetRealPoint(int column, int row){
				return(new FCAMPoint(column * JEWEL_SIZE.Width + PosGap.X, row * JEWEL_SIZE.Height + PosGap.Y));
			}
			public void Put(sbyte kind, FCAMColor color, int fallSpeed){
				Kind = kind;
				Blender = color;
				__FallSpeed = fallSpeed;
			}
			public void Put(sbyte kind, FCAMColor color){
				Put(kind, color, GetJewelFallSpeed(kind));
				Kind = kind;
				Blender = color;
			}
			public void Put(sbyte kind){
				Put(kind, FCAMColor.White);
			}
			public void Initialize(sbyte kind = 0){
				Kind = kind;
				Frame = 0;
				PosGap = new FCAMPoint();
				Blender = FCAMColor.White;
				__FallSpeed = GetJewelFallSpeed(kind);
			}
			/// <summary>
			/// 消去されたら true を返す（強固ジュエルは消去せず、お邪魔ジュエルに削られるだけ）。
			/// </summary>
			public bool Delete(){
				if(Kind == HARDOBSTACLE_JEWEL){
					Kind = OBSTACLE_JEWEL;
					return(false);
				}
				PosGap = new FCAMPoint();
				Kind = 0;
				Frame = 0;
				return(true);
			}
			public override string ToString(){
				return(FCAMStrings.Format("JewelData: Kind={0}", Kind));
			}
		}
		/// <summary>
		///	
		/// </summary>
		class PlayerKeyState{
			public FCAMInputState Left, Up, Right, Down, Rotate, Pause, Support;
#if ATTACK_MAGIC
			public FCAMInputState Attack;
#endif
			/// <summary>
			/// 
			/// </summary>
			/// <param name="keys">index = 0 から順に Left, Up, Right, Down, Rotate, Pause, Attack, Support</param>
			public PlayerKeyState(FCAMInputState[] keys){
				Left = keys[0];
				Up = keys[1];
				Right = keys[2];
				Down = keys[3];
				Rotate = keys[4];
				Pause = keys[5];
#if ATTACK_MAGIC
				Attack = keys[6];
#endif
				Support = keys[7];
			}
		}
		/// <summary>
		///	連鎖のテキスト
		/// </summary>
		class ChainText{
			public short Chain;
			public FCAMPoint CenterPos;
			public short Frame;
			public ChainText(short chain, FCAMPoint centerPos, short numberFrameSteps = 1){
				Chain = chain;
				CenterPos = centerPos;
				Frame = 0;
			}
		}
		/// <summary>
		///	
		/// </summary>
		class JewelSet: IFCAMClonable<JewelSet>{
			public JewelData[, ] Jewels;
			public FCAMTableSize Size;
			public JewelSet(FCAMTableSize size){
				Size = size;
				Jewels = new JewelData[size.Columns, size.Rows];
			}
			public JewelSet Clone(){
				JewelSet clone = new JewelSet(Size);
				clone.Jewels = (JewelData[, ])Jewels.Clone();
				return(clone);
			}
			/// <summary>
			/// 回転（シフト）を行なう。
			/// </summary>
			public void Rotate(int num = 1){
				JewelData[, ] jewelDataArray = (JewelData[, ])Jewels.Clone();
				for(int i = Jewels.Length - 1; i >= 0; i--){
					Jewels[0, (i * num + 1) % Jewels.Length] = jewelDataArray[0, i * num];
				}
			}
			/// <summary>
			/// 全て空にする。
			/// </summary>
			public void Clear(){
				for(int row = 0; row < Size.Rows; row++){
					for(int column = 0; column < Size.Columns; column++){
						Jewels[column, row].Kind = 0;
					}
				}
			}
			/// <summary>
			/// 全てが空かどうか。
			/// </summary>
			public bool IsEmpty(){
				for(int row = 0; row < Size.Rows; row++){
					for(int column = 0; column < Size.Columns; column++){
						if(Jewels[column, row].Kind != 0){
							return false;
						}
					}
				}
				return true;
			}
		}
		/// <summary>
		///	ジュエルを消去した時の破片
		/// </summary>
		class JewelFragment{
			public FCAMPoint CenterPos;
			public sbyte Jewel;
			public sbyte Shape;
			public short Rotation;
			public short RotationSpeed;
			public FCAMPoint Velocity;
			public JewelFragment(sbyte jewel, sbyte shape, short rotation, short rotationSpeed, FCAMPoint centerPos, FCAMPoint velocity){
				Jewel = jewel;
				Shape = shape;
				Rotation = rotation;
				RotationSpeed = rotationSpeed;
				CenterPos = centerPos;
				Velocity = velocity;
			}
		}
		/// <summary>
		///	大きくなって薄らぐ図形のアニメーション
		/// </summary>
		class FadeoutDiagram{
			public FCAMRectangle Posize;
			public DiagramKind Kind;
			public FCAMColor Color;
			public sbyte FadeoutSteps;
			public sbyte InfalteSteps;
			public int Alpha;
			public FadeoutDiagram(FCAMRectangle posize, FCAMColor color, DiagramKind kind, sbyte fadeoutSteps = 16, sbyte inflateSteps = 1, int? alpha = null){
				Posize = posize;
				Kind = kind;
				Color = color;
				FadeoutSteps = fadeoutSteps;
				InfalteSteps = inflateSteps;
				Alpha = alpha ?? FCAMColor.MaxValue;
			}
			/// <summary>
			/// アニメーションを終えたら true を返す。
			/// </summary>
			public bool Calc(){
				Alpha -= FadeoutSteps;
				if(Alpha <= 0){
					Alpha = 0;
					return(true);
				}
				Posize.Inflate(InfalteSteps);
				return(false);
			}
			public void Draw(FCAMGraphics g){
				Color.A = Alpha;
				switch(Kind){
					case DiagramKind.Rectangle:{
						g.DrawRectangle(Posize.X, Posize.Y, Posize.Width, Posize.Height, Color);
						break;
					}
					case DiagramKind.FillRectangle:{
						g.FillRectangle(Posize.X, Posize.Y, Posize.Width, Posize.Height, Color);
						break;
					}
				}
			}
		}
		/// <summary>
		///	
		/// </summary>
		class Board<T>{
			public T [, ] Data;
			public FCAMTableSize Size{get; private set;}
			public Board(FCAMTableSize size){
				Data = new T[size.Columns, size.Rows];
				Size = size;
			}
			public void Clear(){
				Data.Initialize();
			}
		}
		class DeleteField: Board<bool>{
			/// <summary>消去には至らないが水平・垂直に２つ以上繋がったジュエルの数。</summary>
			public int HVPairJewels;
			/// <summary>消去には至らないが斜めに２つ以上繋がったジュエルの数。</summary>
			public int DiagonallyPairJewels;
			/// <summary>消去することで得られるエナジー</summary>
			public int Energy;
			/// <summary>消去することで与えられるダメージ</summary>
			public int Damage;
			/// <summary>消去（砕石）可能なジュエル数</summary>
			public int DeleteOrScrapeJewels;
			/// <summary>消去されるお邪魔ジュエル数</summary>
			public int DamageJewels;
			/// <summary>削られる強固ジュエル数</summary>
			public int HardDamageJewels;
			/// <summary>消去可能なジュエル数</summary>
			public int DeleteJewels{get{return(DeleteOrScrapeJewels - HardDamageJewels);}}
			/// <summary></summary>
			public DeleteField(FCAMTableSize size): base(size){}
			/// <summary>
			/// 消去フィールド内に消去可能なジュエルが存在するか。
			/// </summary>
			public bool ExistDeleteJewel(){
				for(int row = 0; row < Size.Rows; row++){
					for(int column = 0; column < Size.Columns; column++){
						if(Data[column, row])
							return(true);
					}
				}
				return(false);
			}
		}
		class PlayerField: Board<JewelData>{
			public PlayerField(FCAMTableSize size): base(size){}
			public PlayerField Clone(){
				PlayerField clone = new PlayerField(Size);
				clone.Data = (JewelData[, ])Data.Clone();
				return(clone);
			}
			/// <summary>
			/// 現在のフィールドの状態がピンチかどうか（あと一個真ん中に落とすとゲームオーバー）
			/// </summary>
			/// <param name="jewelSet">ジュエルの長さを計算するため用。</param>
			public bool IsPinch(JewelSet jewelSet){
				return Data[Size.Columns / 2, jewelSet.Size.Rows + FIELD_BLIND_HEIGHT - 1].Kind != 0;
			}
			public void PutJewelSet(JewelSet jewelSet, int putColumn, int putRow){
				for(int row = 0; row < jewelSet.Size.Rows; row++){
					for(int column = 0; column < jewelSet.Size.Columns; column++){
						Data[putColumn + column, putRow + row].Initialize(jewelSet.Jewels[column, row].Kind);
					}
				}
			}
			public int GetNumberNormalJewels(){
				int sum = 0;
				for(int row = Size.Rows - 1; row >= 0; row--){
					for(int column = Size.Columns - 1; column >= 0; column--){
						if(IsJewelNormal(Data[column, row].Kind)){
							sum++;
						}
					}
				}
				return(sum);
			}
			public int GetNumberObstractJewels(){
				int sum = 0;
				for(int row = Size.Rows - 1; row >= 0; row--){
					for(int column = Size.Columns - 1; column >= 0; column--){
						if(IsJewelObstract(Data[column, row].Kind)){
							sum++;
						}
					}
				}
				return(sum);
			}
			public new void Clear(){
				for(int row = Size.Rows - 1; row >= 0; row--){
					for(int column = Size.Columns - 1; column >= 0; column--){
						Data[column, row] = new JewelData();
					}
				}
			}
			/// <summary>
			/// 指定位置の Jewel を取得。フィールド外は -1 として取得する。
			/// </summary>
			public int GetJewel(int column, int row){
				if(column < 0 || row < 0 || column >= Size.Columns || row >= Size.Rows)
					return(-1);
				return(Data[column, row].Kind);
			}
		}
		/// <summary>
		///	必ず初期設定しておくべきもの
		///	DeleteLength
		/// </summary>
		class Player{
			/// <summary>プレイヤーごとにジュエルが異ならないよう、静的にする。</summary>
			public static int NextJewelRandomSeed{get; private set;}
			public static int DamageRandomSeed{get; private set;}
			public static readonly FCAMSize LED_SIZE = new FCAMSize(8, 2);
			public const int MAX_NAME_LENGTH = 16;
			public const int NUMBER_LEDS = 4;
			public static readonly int DIAGRAM_ANIMATION_INTERVAL_AT_EFFECTING_MAGIC = Game.Instance.FPS;
			public readonly FCAMPoint	PosName, PosField, PosPortrait, PosScore, PosJewels, PosEnergy, PosDamage, PosEnergyLevel,
										PosVisibleField, PosNextJewelSetOfTopOnField;
			public readonly FCAMPoint[]	PosNextJewelSets, PosLEDArray;
			public PlayerField Field;
			public bool KeyLeft, KeyUp, KeyRight, KeyDown, KeyRotate, KeyPause, KeyAttack, KeySupport;
			public bool RightAlign;
			public string ScoreTextFormat;	//	{0, -8}
			public string JewelsTextFormat;
			public MagicProcessor Magic;
			/// <summary>幾つの魔法を受けているか</summary>
			public sbyte NumberReceivingMagics;
			public int JewelFrame;
			public int Score;
			public int Jewels;
			public int Damage;
			public string Name;
			public FCAMRandom NextJewelRandom{get; private set;}
			public FCAMRandom DamageRandom{get; private set;}
			public CPUBrain CPU;
			/// <summary>
			/// 計算する範囲。例えば一番左に置こうとした際、一番左が積み上がっていたら置けないので、
			/// その時に一番左を除外した範囲で計算する必要があるため、その時に使用する。
			/// ジュエルを置いたりして操作ジュエルが元の位置に戻ると、MoveRange は初期状態となる。
			/// </summary>
			public FCAMTableRectangle MoveRange;			//	プロパティだと外部から設定出来ないため、通常の変数とする。
			public FCAMList<JewelSet> JewelSetList;
			public JewelSet NowJewelSet{
				get{
					return(JewelSetList[0]);
				}
			}
			public FCAMTablePoint JewelSetPos;
			public Battle Parent{get; private set;}
			public int PutWait{get; set;}
			public short Chain;
			public short Level{get{return(FCAMMath.Min((short)(Jewels / LEVEL_UP_INTERVAL + 1), MAX_LEVEL));}}
			/// <summary>指定の energy から EnergyLevel を取得する。</summary>
			public static sbyte GetEnergyLevel(int energy){
				return((sbyte)(energy / ENERGY_LEVEL_INTERVAL));
			}
			/// <summary>指定の Energy を EnergyLevel に合わせて加減算</summary>
			public static int GetEnergyShiftedByLevel(int energy, sbyte shiftLevel){
				return(FCAMMath.Within(0, (energy % ENERGY_LEVEL_INTERVAL) + (GetEnergyLevel(energy) - shiftLevel) * ENERGY_LEVEL_INTERVAL, MAX_ENERGY));
			}
			public sbyte EnergyLevel{
				get{
					return(GetEnergyLevel(Energy));
				}
				set{
					Energy = (sbyte)FCAMMath.Max(0, (Energy % ENERGY_LEVEL_INTERVAL) + value * ENERGY_LEVEL_INTERVAL);
				}
			}
			/// <summary>
			/// 中に浮いたジュエルを落下させる時のウェイト。
			/// </summary>
			public int FallWait{get; private set;}
			/// <summary> ジュエルセットのずれ位置 </summary>
			public FCAMPoint JewelSetPosGap;
			public int DeleteLength;
			private PlayerState __State;
			public PlayerState State{
				get{
					return(NumberReceivingMagics > 0 ? PlayerState.ReceiveMagic : __State);
				}
				set{
					__State = value;
				}
			}
			/// <summary>CPUの計算した移動先</summary>
			public CPUBrainCalcResult CPUAim;
			/// <summary>指定の Energy を既定範囲に収めて正しい Energy を取得する。</summary>
			public static int GetTrueEnergy(int energy){
				return(FCAMMath.Within(0, energy, MAX_ENERGY));
			}
			/// <summary>魔法を発動するための Energy。</summary>
			private int __Energy;
			public int Energy{
				get{
					return(__Energy);
				}
				set{
					__Energy = GetTrueEnergy(value);
				}
			}
			public PlayerKeyState Keys;
			/// <summary>どの画面のプレイヤーか。0:左、1:右。</summary>
			public sbyte PlayerId{get; private set;}
			public sbyte Select;
			public bool IsCPU{get{return(Kind == PlayerKind.CPU);}}
			/// <summary>プレイヤーが参加（誰かが操作している）か。</summary>
			public bool IsParticipant{get{return(Kind != PlayerKind.NonParticipant);}}
			public sbyte Figure;
			public PlayerKind Kind;
			public sbyte CharacterID{get; private set;}
			public FCAMTexture[] Portraits{get; private set;}
			private int MagicFrame;
			/// <summary>ゲーム開始前に使用される変数</summary>
			public sbyte StartHeight;
			public JewelKind StartJewel;
			/// <summary>デバッグ用変数</summary>
			private const int SCOREBOARD_FONT_SIZE = 10;
			private const string SCOREBOARD_FONT_NAME = "ＭＳ ゴシック";
			private FCAMFont ScoreboardFont;
			/// <summary>
			/// 
			/// </summary>
			public static void ResetNextJewelRandomSeed(){
				NextJewelRandomSeed = FCAMSystem.Ticks;
				DamageRandomSeed = (int)FCAMSystem.MicroTicks;
			}
			/// <summary>
			/// 
			/// </summary>
			public Player(FCAMContentsManager manager, Battle parent, sbyte playerId, string name, PlayerKeyState keys){
				///	readonly の初期化は必ず new で行なう。set メソッドを呼び出しても readonly のためか 0 のままである。
				///	
				Portraits = new FCAMTexture[NUMBER_CHARACTER_FIGURES];
				JewelSetList = new FCAMList<JewelSet>();
				Parent = parent;
				Jewels = 0;
				JewelFrame = 0;
				Figure = 0;
				StartHeight = 0;
				StartJewel = JewelKind.Normal;
				Kind = PlayerKind.Human;
				for(int i = 0; i < NUMBER_NEXT_JEWEL_SETS + 1; i++){
					JewelSetList.Add(new JewelSet(JEWEL_SET_SIZE));
				}
				Field = new PlayerField(FIELD_TABLESIZE);
				PosNextJewelSets = new FCAMPoint[NUMBER_NEXT_JEWEL_SETS];
				PosLEDArray = new FCAMPoint[NUMBER_LEDS];
				Keys = keys;
				PlayerId = playerId;
				switch(playerId){
					case 0:
						PosName = new FCAMPoint(32, 8);
						PosField = new FCAMPoint(16, 48);
						PosNextJewelSets[0] = new FCAMPoint(256, 112);
						PosNextJewelSets[1] = new FCAMPoint(296, 160);
						PosEnergy = new FCAMPoint(272, 224);
						PosEnergyLevel = new FCAMPoint(254, 224);
						PosScore = new FCAMPoint(248, 352);
						PosJewels = new FCAMPoint(256, 416);
						PosDamage = new FCAMPoint(16, 36);
						RightAlign = true;
						PosLEDArray[0] = new FCAMPoint(32, 4);
						PosLEDArray[1] = new FCAMPoint(48, 4);
						PosLEDArray[2] = new FCAMPoint(64, 4);
						PosLEDArray[3] = new FCAMPoint(80, 4);
						PosNextJewelSetOfTopOnField = new FCAMPoint(PosField.X + (Field.Size.Columns / 2) * JEWEL_SIZE.Width
																,PosField.Y - JEWEL_SET_SIZE.Rows * JEWEL_SIZE.Height);
						break;
					case 1:
						PosName = new FCAMPoint(352, 8);
						PosField = new FCAMPoint(400, 48);
						PosNextJewelSets[0] = new FCAMPoint(352, 112);
						PosNextJewelSets[1] = new FCAMPoint(328, 160);
						PosEnergy = new FCAMPoint(272, 256);
						PosEnergyLevel = new FCAMPoint(370, 256);
						PosScore = new FCAMPoint(264, 384);
						PosJewels = new FCAMPoint(320, 448);
						PosDamage = new FCAMPoint(400, 36);
						PosLEDArray[0] = new FCAMPoint(552, 4);
						PosLEDArray[1] = new FCAMPoint(568, 4);
						PosLEDArray[2] = new FCAMPoint(584, 4);
						PosLEDArray[3] = new FCAMPoint(600, 4);
						RightAlign = false;
						PosNextJewelSetOfTopOnField = new FCAMPoint(PosField.X + (Field.Size.Columns / 2) * JEWEL_SIZE.Width
																,PosField.Y - JEWEL_SET_SIZE.Rows * JEWEL_SIZE.Height);
						break;
					default:
						return;
				}
				PosField.Y -= FIELD_BLIND_HEIGHT * JEWEL_SIZE.Height;
				PosVisibleField = new FCAMPoint(PosField.X, PosField.Y + FIELD_BLIND_HEIGHT * JEWEL_SIZE.Height);
				PosPortrait = new FCAMPoint(PosVisibleField.X, PosVisibleField.Y);
				//以下の設定は、SetCharacterID で行なわれる
				//CharacterID = 1;
				//Name = name;
				//if(RightAlign){
				//    while(Name.Length < MAX_NAME_LENGTH){
				//        Name = " " + Name;
				//    }
				//}
				ScoreTextFormat = "{0, " + (RightAlign ? "-" : "") + "8}";
				JewelsTextFormat = "{0, " + (RightAlign ? "-" : "") + "4}";
				SetCharacterID(manager, Game.MemoryData.CharaShift);
				Reset();
				ScoreboardFont = manager.MakeFont(SCOREBOARD_FONT_SIZE, SCOREBOARD_FONT_NAME);
			}
			public bool IsMagicEffecting(){
				return(Magic != null && Magic.IsEffecting());
			}
			/// <summary>
			/// プレイヤーの処理を行なう際の初期処理(キー)
			/// </summary>
			public void ResetKeys(){
				KeyLeft = KeyUp = KeyRight = KeyDown = KeyRotate = KeyPause = KeyAttack = KeySupport = false;
			}
			/// <summary>
			/// プレイヤーを初期状態(バトル開始前)にする
			/// </summary>
			public void Reset(){
				NextJewelRandom = new FCAMRandom(NextJewelRandomSeed);
				DamageRandom = new FCAMRandom(DamageRandomSeed);
				State = PlayerState.Normal;
				Energy = 0;
				Damage = 0;
				Select = 0;
				Score = 0;
				Jewels = 0;
				NumberReceivingMagics = 0;
				MagicFrame = 0;
			}
			public void Dispose(){
				FCAMSystem.DisposeClass(out ScoreboardFont);
				foreach(FCAMTexture texture in Portraits){
					texture.Dispose();
				}
			}
#if ATTACK_MAGIC
			public void CastAttackMagic(Player caster, params Player[] targets){
				Magic = CharacterInfo.Infos[CharacterID, Game.MemoryData.Difficult].AttackMagic.Begin(Parent.TextureMagic, caster, targets);
				MagicFrame = 0;
			}
#endif
			public void CastSupportMagic(Player caster, params Player[] targets){
				Magic = CharacterInfo.Infos[CharacterID, Game.MemoryData.Difficult].SupportMagic.Begin(Parent.TextureMagic, caster, targets);
				MagicFrame = 0;
			}
			/// <summary>
			/// プレイヤー情報をリロードする（ポートレートのテクスチャなど）。
			/// </summary>
			public void SetCharacterID(FCAMContentsManager manager, sbyte characterID){
				CharacterInfo info = CharacterInfo.Infos[characterID, Game.MemoryData.Difficult];
				CharacterID = characterID;
				Name = info.Name;
				if(RightAlign){
				    while(Name.Length < MAX_NAME_LENGTH){
				        Name = " " + Name;
				    }
				}
				CPU = info.CPU;
				for(int i = 0; i < NUMBER_CHARACTER_FIGURES; i++){
					FCAMSystem.DisposeClass(out Portraits[i]);
					Portraits[i] = manager.MakeTexture(Common.GetCharacterPath(characterID, i));
				}
			}
			/// <summary>
			/// フィールド上の全てのジュエルに対して処理を施す。
			/// </summary>
			public void DoToAllJewelsInField(DoToJewelDelegate func){
				for(int row = FIELD_TABLESIZE.Rows - 1; row >= 0; row--){
					for(int column = FIELD_TABLESIZE.Columns - 1; column >= 0 ; column--){
						func(ref Field.Data[column, row]);
					}
				}
			}
			/// <summary>
			/// 
			/// </summary>
			public void ClearField(){
				Field.Clear();
				do{
					for(int row = FIELD_TABLESIZE.Rows - 1; row > FIELD_TABLESIZE.Rows - StartHeight - 1; row--){
						for(int column = 0; column < FIELD_TABLESIZE.Columns; column++){
							sbyte jewel;
							switch(StartJewel){
								case JewelKind.HardDamage:{jewel = HARDOBSTACLE_JEWEL; break;}
								case JewelKind.Damage:{jewel = OBSTACLE_JEWEL; break;}
								case JewelKind.Normal:{
									jewel = (sbyte)NextJewelRandom.NextInt(1, NUMBER_KIND_NORMAL_JEWELS);
									break;
								}
								default:{
									Error.ID_00005.Throw();
									return;
								}
							}
							Field.Data[column, row].Initialize(jewel);
						}
					}
				//	初期状態で削除するジュエルが無いようループさせる。
				}while(GetDeleteField(Field, DeleteLength, 0).ExistDeleteJewel());
			}
			public void PutJewelSet(){
				Field.PutJewelSet(NowJewelSet, JewelSetPos.Column, JewelSetPos.Row);
			}
			/// <summary>
			/// 設置タイム PutWait を 0 にリセットする。
			/// </summary>
			public void ResetPutWait(){
				PutWait = 0;
			}
			/// <summary>
			/// CPU 目標を再計算。事前に MoveRange が正しく設定されていること。
			/// </summary>
			public void RecalcCPUAim(Battle battle){
				CPUAim = CPU.GetAim(battle, this, !IsMagicEffecting());
			}
			/// <summary>
			/// MoveRange を初期化する。
			/// RecalcCPUAim() を呼び出す前に MoveRange を設定する必要があるため、
			/// もし MoveRange を設定しない場合はこれを利用する。
			/// </summary>
			public void ResetMoveRange(){
				MoveRange.Set(new FCAMTablePoint(), Field.Size);
			}
			/// <summary>
			/// 次の JewelSet を設定し、CPU なら更に目標地点範囲と操作待ち時間をリセットする。
			/// </summary>
			public void SetNextJewelSet(){
				for(int i = 0; i < NUMBER_NEXT_JEWEL_SETS; i++){
					JewelSetList[i] = JewelSetList[i + 1];
				}
				JewelSetList[NUMBER_NEXT_JEWEL_SETS] = new JewelSet(JEWEL_SET_SIZE);
				for(int row = 0; row < NowJewelSet.Size.Rows; row++){
					for(int column = 0; column < NowJewelSet.Size.Columns; column++){
						JewelSetList[NUMBER_NEXT_JEWEL_SETS].Jewels[column, row] = new JewelData((sbyte)NextJewelRandom.NextInt(1, NUMBER_KIND_NORMAL_JEWELS));
					}
				}
				JewelSetPos.Set(JEWEL_SET_START_POS_COLUMN, 0);
				JewelSetPosGap.Y = -JEWEL_SIZE.Height;
				if(!IsCPU){
					return;
				}
				CPU.ResetOperateWait();
				ResetMoveRange();
			}
			/// <summary>
			/// 現在の JewelSet を空に設定する。
			/// ジュエルセットを配置したあとの連鎖やチェック状態で、現在のジュエルセットを表示しないためなど用。
			/// </summary>
			public void SetNowJewelSetEmpty(){
				NowJewelSet.Clear();
			}
			/// <summary>
			/// JewelSet の指定方向に移動が行なえる時は true を返す。
			/// </summary>
			public bool CanMove(int moveColumn, int moveRow){
				return(Field.GetJewel(JewelSetPos.Column + moveColumn, JewelSetPos.Row + NowJewelSet.Size.Rows - 1 + moveRow) == 0);
			}
			/// <summary>
			/// JewelSet の列方向移動を行なう。
			/// </summary>
			///	<returns>移動が行われたら TRUE</returns>
			public bool MoveJewelSetColumn(int column){
				int before = JewelSetPos.Column;
				if(column > 0){
					for(; JewelSetPos.Column + 1 < FIELD_TABLESIZE.Columns && column > 0; column--){
						//	row + JewelSet.Length がフィールド外にならないよう判定を入れるべきだが、
						//	そもそもジュエルセットの行位置がフィールド外になるのは事前のミスのため、ここでは不要である
						for(int row = 0; row < NowJewelSet.Size.Rows; row++){
							if(Field.Data[JewelSetPos.Column + 1, JewelSetPos.Row + row].Kind != 0){
								column = 0;
								break;
							}
						}
						if(column > 0)
							JewelSetPos.Column++;
					}
				}else if(column < 0){
					for(; JewelSetPos.Column > 0 && column < 0; column++){
						//	row + JewelSet.Length がフィールド外にならないよう判定を入れるべきだが、
						//	そもそもジュエルセットの行位置がフィールド外になるのは事前のミスのため、ここでは不要である
						for(int row = 0; row < NowJewelSet.Size.Rows; row++){
							if(Field.Data[JewelSetPos.Column - 1, JewelSetPos.Row + row].Kind != 0){
								column = 0;
								break;
							}
						}
						if(column < 0)
							JewelSetPos.Column--;
					}
				}
				JewelSetPosGap.X = 0;
				return(before != JewelSetPos.Column);
			}
			/// <summary>
			/// JewelSet の行方向移動（落下）を行なう。
			/// </summary>
			///	<returns>移動が行われたら TRUE</returns>
			public bool MoveJewelSetRow(int row){
				int before = JewelSetPos.Row;
				if(row > 0){
					for(; JewelSetPos.Row + NowJewelSet.Size.Rows < FIELD_TABLESIZE.Rows && row > 0 && Field.Data[JewelSetPos.Column, JewelSetPos.Row + NowJewelSet.Size.Rows].Kind == 0; JewelSetPos.Row++, row--){}
				}else if(row < 0){
					for(; JewelSetPos.Row > 0 && row < 0 && Field.Data[JewelSetPos.Column, JewelSetPos.Row - 1].Kind == 0; JewelSetPos.Row--, row++){}
				}else{
					return(JewelSetPos.Row + NowJewelSet.Size.Rows >= FIELD_TABLESIZE.Rows || Field.Data[JewelSetPos.Column, JewelSetPos.Row + NowJewelSet.Size.Rows].Kind != 0);
				}
				if(before == JewelSetPos.Row){
					return(false);
				}
				return(true);
			}
			/// <summary>
			/// 
			/// </summary>
			public FCAMPoint GetJewelPos(int column, int row){
				return(new FCAMPoint(
					column * JEWEL_SIZE.Width + PosField.X,
					row * JEWEL_SIZE.Height + PosField.Y
				));
			}
			/// <summary>
			/// VS モードでキャラクター選択中、名前まで暗くなるので、
			/// 二重処理にはなってしまうが再度この関数を呼んで名前を明確にする。
			/// </summary>
			public void DrawName(){
				Common.DrawString(PosName, Name);
			}
			/// <summary>
			/// 現在のジュエルセットのポサイズを取得
			/// </summary>
			public FCAMRectangle GetNowJewelSetPosize(){
#if SHOW_ALL_CONTROL_JEWELS
				return new FCAMRectangle(false
										,JewelSetPos.Column * JEWEL_SIZE.Width + PosField.X + JewelSetPosGap.X
										,JewelSetPos.Row * JEWEL_SIZE.Height + PosField.Y + JewelSetPosGap.Y
										,JEWEL_SIZE.Width
										,NowJewelSet.Size.Rows * JEWEL_SIZE.Height);
#else				
				int y = JEWEL_SIZE.Height * (JewelSetPos.Row - FIELD_BLIND_HEIGHT) + JewelSetPosGap.Y;
				int height = JEWEL_SIZE.Height * NowJewelSet.Size.Rows;
				if(y < 0){
					height += y;
					y = 0;
				}
				return(new FCAMRectangle(false, JewelSetPos.Column * JEWEL_SIZE.Width + PosField.X + JewelSetPosGap.X, y + JEWEL_SIZE.Height * FIELD_BLIND_HEIGHT + PosField.Y, JEWEL_SIZE.Width, height));
#endif
			}
			public void AddFadeoutDiagramAtEnergy(){
				Parent.FadeoutDiagramList.Add(new FadeoutDiagram(new FCAMRectangle(false, PosEnergy.X, PosEnergy.Y, ENERGY_SIZE.Width, ENERGY_SIZE.Height), FCAMColor.White, DiagramKind.FillRectangle));
			}
			/// <summary>
			/// 
			/// </summary>
			public void Calc(){
				//if(State == PlayerState.CastMagic){
				if(IsMagicEffecting()){
					MagicFrame = (MagicFrame + 1) % DIAGRAM_ANIMATION_INTERVAL_AT_EFFECTING_MAGIC;
					//	魔法の効果を発揮中
					if(MagicFrame == 0){
						AddFadeoutDiagramAtEnergy();
					}
					Magic.DoCalc();
				}else if(State != PlayerState.ReceiveMagic){
					//	魔法を受けていなければ
					//	フィールド上のジュエル
					for(int row = FIELD_BLIND_HEIGHT; row < FIELD_TABLESIZE.Rows; row++){
						for(int column = 0; column < FIELD_TABLESIZE.Columns; column++){
							if(Field.Data[column, row].Kind == 0)
								Field.Data[column, row].Frame = 0;
							else
								Field.Data[column, row].Frame++;
							Field.Data[column, row].Frame %= MAX_JEWEL_FRAME;
						}
					}
				}
			}
			/// <summary>
			/// 前描画処理
			/// ShiftY はキャラクターが負けた時のシフト用
			/// </summary>
			public void DrawBefore(FCAMGraphics g, BattleState battleState, int losePortraitShiftY = int.MinValue, int backPortraitShiftY = int.MinValue){
				FCAMColor leftColor, rightColor;
				FCAMFlip portraitFlip = PlayerId == 0 ? FCAMFlip.H : FCAMFlip.None;
				if(!IsParticipant){
					return;
				}
#if SHOW_ALL_CONTROL_JEWELS
				//	次の操作ジュエル（フィールドの頭）
				if(Parent.State != BattleState.Pause && NUMBER_NEXT_JEWEL_SETS >= 1){
					for(int row = JewelSetList[1].Size.Rows - 1; row >= 0; row--){
						//	※JewelSetPosGap.Y は -JEWEL_SIZE.Height～0
						int gapY = FCAMMath.Min(0, (JewelSetPos.Row - JEWEL_SET_SIZE.Rows) * JEWEL_SIZE.Height + (JewelSetPosGap.Y + JEWEL_SIZE.Height));
						DrawJewel(g, 0, row, JewelSetList[1].Jewels[0, row], PosNextJewelSetOfTopOnField.X, PosNextJewelSetOfTopOnField.Y + gapY);
					}
				}
#endif
#if RANKING
				Common.DrawString(PosScore, FCAMStrings.Format(ScoreTextFormat, Score));
				Common.DrawString(PosJewels, FCAMStrings.Format(JewelsTextFormat, Jewels));
#endif
				//	LED
				{
					bool[] leds = new bool[NUMBER_LEDS];
					int i;
					switch(Kind){
						case PlayerKind.NonParticipant:
							i = 0;
							break;
						case PlayerKind.Human:
							i = 1;
							break;
						case PlayerKind.CPU:
							i = 4 + Game.MemoryData.Difficult;
							break;
						default:
							i = 2;
							break;
					}
#if DEBUG
					if(FCAMDebug.IsDebug){
						i |= 8;
					}
#endif
					leds[3] = (i & 1) > 0;
					leds[2] = (i & 2) > 0;
					leds[1] = (i & 4) > 0;
					leds[0] = (i & 8) > 0;
					for(i = 0; i < leds.Length; i++){
						FCAMColor blender = leds[i] ? FCAMColor.Red : FCAMColor.Black;
						g.FillRectangle(PosLEDArray[i].X, PosLEDArray[i].Y, LED_SIZE.Width, LED_SIZE.Height, blender);
					}
				}
				//	通常時のキャラクター画像
				if(backPortraitShiftY != int.MinValue && Parent.Mode != BattleMode.Free){
					if(State != PlayerState.Lose){
						//	負けてないキャラは、背景ポートレイトをシフトさせない。
						backPortraitShiftY = 0;
					}else{
						if(battleState == BattleState.SceneEnd){
							//	敗北キャラはバトル終了後に画像を描画しない（肥満画像が変わりに描画されるため）
							backPortraitShiftY = int.MinValue;
						}
					}
					g.Draw(Portraits[0], PosPortrait.X, PosPortrait.Y + backPortraitShiftY, PORTRAIT_SIZE.Width, PORTRAIT_SIZE.Height, (Portraits[0].Width - FIELD_SIZE.Width) / 2, 0, portraitFlip);
				}
				//	フィールド上のジュエル
				if(Parent.State != BattleState.Pause){
					for(int row = FIELD_BLIND_HEIGHT; row < FIELD_TABLESIZE.Rows; row++){
						for(int column = 0; column < FIELD_TABLESIZE.Columns; column++){
							JewelData data = Field.Data[column, row];
							int animation = (data.Frame - MAX_JEWEL_FRAME + NUMBER_JEWEL_ANIMATION_FRAMES) / NUMBER_JEWEL_ANIMATION_INTERVAL;
							DrawJewel(g, column, row, data, PosField.X + data.PosGap.X, PosField.Y + data.PosGap.Y, 0, 1);
						}
					}
				}
				//	次のジュエル
				if(Parent.State != BattleState.Pause){
					for(int i = 0; i < NUMBER_NEXT_JEWEL_SETS; i++){
						for(int row = JewelSetList[i + 1].Size.Rows - 1; row >= 0; row--){
							DrawJewel(g, 0, row, JewelSetList[i + 1].Jewels[0, row], PosNextJewelSets[i].X, PosNextJewelSets[i].Y, 0, i + 1);
						}
					}
				}
				//	エナジー
				int energyWidth;
				Common.DrawString(PosEnergyLevel, EnergyLevel.ToString());
				if(Energy < MAX_ENERGY){
					for(int i = 0; Energy >= i * ENERGY_LEVEL_INTERVAL; i++){
						if(i == 0){
							leftColor = FCAMColor.DarkBlue;
							rightColor = FCAMColor.Blue;
						}else{
							leftColor = FCAMColor.DarkGreen;
							rightColor = FCAMColor.Green;
						}
						if(IsMagicEffecting()){
							leftColor.Brighten(FCAMColor.MaxValue / 2);
							rightColor.Brighten(FCAMColor.MaxValue / 2);
						}
						energyWidth = ENERGY_SIZE.Width * (FCAMMath.Min(Energy - i * ENERGY_LEVEL_INTERVAL, ENERGY_LEVEL_INTERVAL)) / ENERGY_LEVEL_INTERVAL;
						g.FillHorizontalGradatedRectangle(PosEnergy.X, PosEnergy.Y, energyWidth, ENERGY_SIZE.Height, leftColor, rightColor);
					}
				}else{
					//int i = FCAMColor.HalfValue / 2;
					//leftColor = new FCAMColor(i, i, i, FCAMColor.MaxValue);
					//rightColor = new FCAMColor(i * 4 / 5, FCAMColor.MaxValue);
					leftColor = FCAMColor.Red;
					rightColor = FCAMColor.LightRed;
					if(IsMagicEffecting()){
						leftColor.Brighten(FCAMColor.MaxValue / 2);
						rightColor.Brighten(FCAMColor.MaxValue / 2);
					}
					g.FillHorizontalGradatedRectangle(PosEnergy.X, PosEnergy.Y, ENERGY_SIZE.Width, ENERGY_SIZE.Height, leftColor, rightColor);
				}
				//	現在のジュエル（ポーズでないかつ空でない時）
				if(Parent.State != BattleState.Pause && !NowJewelSet.IsEmpty()){
					for(int row = NowJewelSet.Size.Rows - 1; row >= 0; row--){
#if	!SHOW_ALL_CONTROL_JEWELS
	#if !SHOW_TRANSPARENT_CONTROL_JEWELS
						DrawJewel(g, JewelSetPos.Column, JewelSetPos.Row + row, NowJewelSet.Jewels[0, row], PosField.X + JewelSetPosGap.X, PosField.Y + JewelSetPosGap.Y, PosField.Y + FIELD_BLIND_HEIGHT * JEWEL_SIZE.Height);
	#else
						//	半透明のジュエル
						DrawJewel(g, JewelSetPos.Column, JewelSetPos.Row + row, NowJewelSet.Jewels[0, row], PosField.X + JewelSetPosGap.X, PosField.Y + JewelSetPosGap.Y, PosField.Y + FIELD_BLIND_HEIGHT * JEWEL_SIZE.Height, 1, true);
	#endif
#else
						//	以前はフィールドより上のジュエルは表示しないようにしていたが、
						//	フィールドより上の方にあるジュエルも表示するようにした。
						DrawJewel(g, JewelSetPos.Column, JewelSetPos.Row + row, NowJewelSet.Jewels[0, row], PosField.X + JewelSetPosGap.X, PosField.Y + JewelSetPosGap.Y);
#endif
					}
					//	現在のジュエルの周りの線
					#if false
						y = JEWEL_SIZE.Height * (JewelSetPos.Row - FIELD_BLIND_HEIGHT) + JewelSetPosGap.Y;
						height = JEWEL_SIZE.Height * NowJewelSet.Size.Rows;
						if(y < 0){
							height += y;
							y = 0;
						}
						if(height > 0){
							g.DrawRectangle(JewelSetPos.Column * JEWEL_SIZE.Width + PosField.X + JewelSetPosGap.X, y + JEWEL_SIZE.Height * FIELD_BLIND_HEIGHT + PosField.Y, JEWEL_SIZE.Width, height, FCAMColor.White);
						}
					#else
						FCAMRectangle posize = GetNowJewelSetPosize();
						g.DrawRectangle(posize.X, posize.Y, posize.Width, posize.Height, FCAMColor.White);
					#endif
				}
				//	
				//	##########	半透明で描画された操作ジュエルを考慮して、それと同じ位置になることがある名前をダメージをあとで描画
				//	
				//	名前
				if(Parent.Mode != BattleMode.Free){
					DrawName();
				}
				//	ダメージ
				leftColor = FCAMColor.LightRed;
				rightColor = FCAMColor.LightRed;
				g.FillHorizontalGradatedRectangle(PosDamage.X, PosDamage.Y, Damage * 2, DAMAGE_SIZE.Height, leftColor, rightColor);
				//	
				//	##########
				//	
				//	負けた時のキャラクター画像
				if(State == PlayerState.Lose && losePortraitShiftY != int.MinValue && Parent.Mode != BattleMode.Free){
					g.Draw(Portraits[1], PosPortrait.X, PosPortrait.Y + losePortraitShiftY, PORTRAIT_SIZE.Width, PORTRAIT_SIZE.Height, (Portraits[1].Width - FIELD_SIZE.Width) / 2, 0, portraitFlip);
				}
				//	デバッグ。CPUのスコアボード
#if DEBUG
				if(Parent.State != BattleState.Pause && FCAMDebug.IsDebug && IsCPU){
					FCAMFont beforeFont = g.Font;
					g.Font = ScoreboardFont;
					for(int row = FIELD_BLIND_HEIGHT; row < FIELD_TABLESIZE.Rows; row++){
						for(int column = 0; column < FIELD_TABLESIZE.Columns; column++){
							g.DrawString(PosField.X + column * JEWEL_SIZE.Width, PosField.Y + row * JEWEL_SIZE.Height, "" + this.CPU.ScoreBoard.Data[column, row]);
						}
					}
					g.Font = beforeFont;
				}
#endif
			}
			/// <summary>
			/// 後描画処理
			/// </summary>
			public void DrawAfter(FCAMGraphics g){
				//	魔法
				if(Magic != null && Magic.IsEffecting()){
					Magic.DoDraw(g);
				}
				//	デバッグ描画
#if DEBUG
				if(FCAMDebug.IsDebug){
					int x;
					int y;
					if(PlayerId == 0){
						x = PosField.X + FIELD_SIZE.Width;
						y = PosField.Y + FIELD_SIZE.Height - 128;
					}else{
						x = PosField.X - 160;
						y = PosField.Y + FIELD_SIZE.Height - 64;
					}
					g.DrawString(x, y, "DAMAGE:" + Damage + FCAMSystem.NewLine
									 + "LEVEL:" + Level + FCAMSystem.NewLine
									 + "JEWELS:" + Jewels);
				}
#endif
			}
			/// <summary>
			/// invertTransparent: 反転（はみ出た部分）透明で描画モード
			///		null: 通常モード（反転しない）
			///		no: borderY ではみ出た部分を透過せずに描画
			///		yes: borderY ではみ出た部分を透過で描画
			/// </summary>
			private void DrawJewel(	FCAMGraphics g,
									int column,
									int row,
									JewelData jewel,
									int shiftX,
									int shiftY,
									int borderY = 0,
									int scaledown = 1,
									bool? invertTransparent = null){
				int pattern = (jewel.Frame - MAX_JEWEL_FRAME + NUMBER_JEWEL_ANIMATION_FRAMES) / NUMBER_JEWEL_ANIMATION_INTERVAL;
				int x = column * JEWEL_SIZE.Width / scaledown + shiftX;
				int y = row * JEWEL_SIZE.Height / scaledown + shiftY;
				int width = JEWEL_SIZE.Width / scaledown;
				int height = JEWEL_SIZE.Height / scaledown;
				int srcX = jewel.Kind * JEWEL_SIZE.Width;
				int srcY = JewelFrame * JEWEL_SIZE.Height;
				int srcWidth = JEWEL_SIZE.Width;
				int srcHeight = JEWEL_SIZE.Height;
				if(pattern < 0){
					pattern = 0;
				}
				if(!invertTransparent.HasValue){
					//	はみ出た部分を描画しない
					if(y < borderY){
						srcY += borderY - y;
						height -= borderY - y;
						y = borderY;
						srcHeight = height;
					}
					if(height > 0){
						g.Draw(Parent.TextureJewel, x, y, width, height, srcX, srcY + JEWEL_SIZE.Height * pattern, srcWidth, srcHeight, jewel.Blender);
					}
				}else{
					//	はみ出た部分を描画する
					Func<bool, int> getSign = (i, border) => i >= border;
					bool sign;
					for(int i = 0; i < 2; i++){
						sign = getSign(y, borderY);
						FCAMColor blender = jewel.Blender;
						if(sign == false && invertTransparent == true){
							blender.A /= 2;
						}
						//	描画
						int realY = y;
						int realSrcY = srcY;
						int realHeight = height;
						if(!sign && y + JEWEL_SIZE.Height > borderY){
							//	負の符号（ボーダーより下）でかつ描画するとボーダーを超える場合
							realHeight -= y + JEWEL_SIZE.Height - borderY;
						}else if(i == 1){
							//	二回目の描画＝ボーダーを跨いだ場合
							realHeight = y - borderY;
							realY -= realHeight;
							realSrcY += JEWEL_SIZE.Height - realHeight;
						}
						int realSrcHeight = realHeight;
						g.Draw(Parent.TextureJewel, x, realY, width, realHeight, srcX, realSrcY + JEWEL_SIZE.Height * pattern, srcWidth, realSrcHeight, blender);
						y += JEWEL_SIZE.Height;
						//	符号が不変＝ボーダーの境界を跨がない場合は描画は1回でOK
						if(getSign(y, borderY) == sign){
							break;
						}
					}
				}
			}
			/*
			private void DrawJewel(FCAMGraphics g, int column, int row, int jewel, int gapX, int gapY, int borderY = 0, int animation = 0, int scaledown = 1){
				int x = column * JEWEL_SIZE.Width / scaledown + gapX;
				int y = row * JEWEL_SIZE.Height / scaledown + gapY;
				int width = JEWEL_SIZE.Width / scaledown;
				int height = JEWEL_SIZE.Height / scaledown;
				int srcX = jewel * JEWEL_SIZE.Width;
				int srcY = JewelFrame * JEWEL_SIZE.Height;
				int srcWidth = JEWEL_SIZE.Width;
				int srcHeight = JEWEL_SIZE.Height;
				if(y < borderY){
					srcY += borderY - y;
					height -= borderY - y;
					y = borderY;
					srcHeight = height;
				}
				if(height > 0){
					g.Draw(Parent.TextureJewel, x, y, width, height, srcX, srcY + JEWEL_SIZE.Height * animation, srcWidth, srcHeight, color ?? FCAMColor.White);
				}
			}
			 */
			/// <summary>
			/// 質問と応答を y を基準に描画する。
			/// </summary>
			public void DrawContinue(FCAMGraphics g, int y){
				int x;
				FCAMRectangle rectangle;
				FCAMRectangle[] rectangles = new FCAMRectangle[]{CONTINUE_YES_POSSIZE, CONTINUE_NO_POSSIZE};
				rectangle = CONTINUE_ASK_POSSIZE;
				x = (FIELD_SIZE.Width - rectangle.Width) / 2 + PosField.X;
				g.Draw(Parent.TextureText, x, y, rectangle.Width, rectangle.Height, rectangle.X, rectangle.Y);
				y += rectangle.Height;
				for(int i = 0; i < 2; i++){
					FCAMColor blender;
					rectangle = rectangles[i];
					x = (FIELD_SIZE.Width - rectangle.Width) / 2 + PosField.X;
					if(i == Select){
						blender.SetARGB(FCAMColor.MaxValue, FCAMColor.MaxValue, FCAMColor.MaxValue, FCAMColor.MinValue);
					}else{
						blender.SetAMono(FCAMColor.HalfValue, FCAMColor.MaxValue);
					}
					g.Draw(Parent.TextureText, x, y, rectangle.Width, rectangle.Height, rectangle.X, rectangle.Y, blender);
					y += rectangle.Height;
				}
			}
		}
	}
}
