﻿using FukuracamDX;
using System.Collections.Generic;

namespace FatopTheWorld.Scenes{
	/// <summary>
	/// このクラスをプログラムする時の注意。
	/// 1.描画処理は魔法を考えて二つに分ける
	///		∵魔法はターゲットに対して描画を行なうが、
	///		　主人公(Player 1)⇒相手(Player 2)への描画の際、
	///		　主人公による相手への描画の次に相手の描画によって上書きされてしまう
	///		　この自体を防ぐため、描画処理を前処理と後処理に分ける。
	/// </summary>
	partial class Battle: FCAMScene{
		//	ストーリーモード用のメソッド＋変数
		/// <summary>ストーリーモードかどうか。</summary>
		private bool IsStory{get{return(Mode == BattleMode.Story);}}
		private Player ProtagonistPlayer{get{return(Players[ProtagonistPlayerID]);}}
		private Player OpponentPlayer{get{return(Players[ProtagonistPlayerID == 0 ? 1 : 0]);}}
		public sbyte ProtagonistPlayerID = -1;
		public int ProtagonistBeginScore = 0;
		public int ProtagonistScore{get{return(ProtagonistPlayer.Score);}}
		public bool IsProtagonistWin{get{return(ProtagonistPlayer.State == PlayerState.Win);}}
		public bool IsProtagonistDraw{get{return(ProtagonistPlayer.State == PlayerState.Draw);}}
		public bool IsProtagonistLose{get{return(ProtagonistPlayer.State == PlayerState.Lose);}}
		/// <summary>
		/// </summary>
		private bool IsAutoProcessOnBattleEndInStory(sbyte playerID){
			return(IsStory && Players[playerID].State != PlayerState.Lose);
		}
		public void SetPlayer(FCAMContentsManager manager, sbyte playerID, sbyte characterID, PlayerKind kind){
			Players[playerID].SetCharacterID(manager, characterID);
			Players[playerID].Kind = kind;
		}
		//	
		//	
		//	
		/// <summary>ピンチになった時のBGMの速度ＵＰ率</summary>
		private const float BGM_SPEEDUP_RATE = 1.1f;
		/// <summary>連鎖時の効果音の速度がＵＰする率</summary>
		private const float CHAIN_SE_SPEEDUP_RATE = 1.05f;
		const string BGM_GAMEOVER_FILEPATH = @"Music\gameover.wav";
		const string FILEPATH_BACKGROUND_FORMAT = @"Battle\Background\{0:D2}.png";
#if RANKING
		const string FILEPATH_BACKGROUND_BLENDER = @"Battle\Background\battle.png";
#else
		const string FILEPATH_BACKGROUND_BLENDER = @"Battle\Background\battle(no-ranking).png";
#endif
		const string FILEPATH_MUSIC_FORMAT = @"Music\{0:D2}.mp3.wav";
		const string FILEPATH_SE_COUNT = @"SE\count.wav";
		const string FILEPATH_SE_START = @"SE\start.wav";
		const string FILEPATH_SE_BREAK = @"SE\break.wav";
		const string FILEPATH_SE_PUT = @"SE\put.wav";
		const string FILEPATH_SE_OJAMA = @"SE\ojama.wav";
		const string FILEPATH_SE_ROTATE = @"SE\rotate.wav";
		const string FILEPATH_SE_MAGIC = @"SE\magic.wav";
		const string FILEPATH_MAGIC = @"Battle\magic.png";
		const string FILEPATH_JEWEL = @"Battle\Jewel.png";
		const string FILEPATH_TEXT = @"Battle\text.png";
		const string LABEL_VS = "VS";
		const string LABEL_NEXT = "NEXT";
		const string LABEL_TIME = "TIME";
		const string LABEL_SCORE = "SCORE";
		const string LABEL_JEWELS = "JEWELS";
		const string TIME_TEXT_FORMAT = "{0:D2}:{1:D2}:{2:D2}";
		//
		static readonly FCAMTableSize FIELD_TABLESIZE = new FCAMTableSize(7, 15);
		static readonly FCAMSize JEWEL_SIZE = new FCAMSize(32);
		static readonly FCAMSize FIELD_SIZE = new FCAMSize(FIELD_TABLESIZE.Columns * JEWEL_SIZE.Width, FIELD_TABLESIZE.Rows * JEWEL_SIZE.Height);
		static readonly FCAMSize VISIBLE_FIELD_SIZE = new FCAMSize(FIELD_TABLESIZE.Columns * JEWEL_SIZE.Width, (FIELD_TABLESIZE.Rows - FIELD_BLIND_HEIGHT) * JEWEL_SIZE.Height);
		static readonly FCAMSize PORTRAIT_SIZE = VISIBLE_FIELD_SIZE;
		static readonly FCAMSize CHAIN_TEXT_CHARACTER_SIZE = new FCAMSize(32);
		static readonly short JEWEL_SET_START_POS_COLUMN = (short)(FIELD_TABLESIZE.Columns / 2);
		static readonly FCAMTableSize JEWEL_SET_SIZE = new FCAMTableSize(1, 3);
		/// <summary>ここにジュエルを置いたらゲームオーバーになる位置</summary>
		static readonly FCAMTablePoint JEWEL_DEAD_POINT = new FCAMTablePoint(JEWEL_SET_START_POS_COLUMN, JEWEL_SET_SIZE.Rows - 1);
		const int NUMBER_CHARACTER_FIGURES = 2;
		const int CHAIN_TEXT_GAP_Y = 80;
		const int NUMBER_CHAIN_TEXT_FRAMES = 20;
		const int MAX_CHAIN_TEXT_ANIMATION = 7;
		static readonly int[] CHAIN_TEXT_CHARS = new int[]{10, 11, 12};
		/// <summary>Energy</summary>
		static readonly FCAMSize ENERGY_SIZE = new FCAMSize(96, 16);
		const int ENERGY_LEVEL_INTERVAL = 50;
		/// <summary>ダメージ</summary>
		static readonly FCAMSize DAMAGE_SIZE = new FCAMSize(224, 8);
		///// <summary>ダメージ計算後の掛ける値。</summary>
		//const float DAMAGE_BASE_RATE = 0.3f;
		///// <summary>ダメージ計算時、連鎖に掛ける値。</summary>
		//const float DAMAGE_CHAIN_RATE = 3.1f;
		///// <summary>ダメージ計算後に、その値をn乗する。その乗数。</summary>
		//const float DAMAGE_RATE = 0.5f;
		///// <summary>ダメージ計算時、消去したジュエル数に掛ける値。</summary>
		//const float DAMAGE_DELETE_RATE = 3.2f;
		/// <summary>連鎖ごとのダメージのベース値。</summary>
		//static readonly int[] DAMAGE_BOARD;
		//static Battle(){
		//    DAMAGE_BOARD = new int[FIELD_TABLESIZE.Columns * FIELD_TABLESIZE.Rows];
		//    for(int i = 0; i < DAMAGE_BOARD.Length; i++){
		//        DAMAGE_BOARD[i] = (i - 1) <= 0 ? 0 : (i - 1) * FIELD_TABLESIZE.Columns;
		//    }
		//}
		/// <summary>連鎖テキストの垂直移動数</summary>
		const int NUMBER_CHAIN_TEXT_RISE_DOWN_STEPS = -1;
		const int NUMBER_NEXT_JEWEL_SETS = 2;
		const sbyte NUMBER_KIND_NORMAL_JEWELS = 6;
		const sbyte MAGIC_JEWEL = 7;
		const sbyte OBSTACLE_JEWEL = 8;
		const sbyte HARDOBSTACLE_JEWEL = 9;
		const short MAX_LEVEL = 99;
		const int LEVEL_UP_INTERVAL = 100;
		/// <summary>FRAME > ANIMATION。フレームは単数または複数のアニメーションの総合。</summary>
		const int MAX_JEWEL_FRAME = 40;
		const int NUMBER_JEWEL_ANIMATION = 8;
		const int NUMBER_JEWEL_ANIMATION_INTERVAL = 1;
		const int NUMBER_JEWEL_ANIMATION_FRAMES = NUMBER_JEWEL_ANIMATION * NUMBER_JEWEL_ANIMATION_INTERVAL;
		/// <summary>フィールド上でジュエルの見えない箇所の高さ ＝ フィールド上でジュエルが一番初めに見える Row</summary>
		const int FIELD_BLIND_HEIGHT = 2;
		/// <summary>ジュエルの水平方向移動速度</summary>
		const int JEWEL_HORZ_MOVE_SPEED = 16;
		/// <summary>ジュエルの垂直方向移動速度</summary>
		const int JEWEL_VERT_MOVE_SPEED = 16;
		/// <summary>宙に浮いたジュエルの落下速度</summary>
		const int NORMAL_JEWEL_FALL_SPEED = 4;
		const int OBSTACLE_JEWEL_FALL_SPEED = 32;
		/// <summary></summary>
		const int NUMBER_JEWEL_FRAGMENT_SHAPES = 4;
		const int NUMBER_JEWEL_FRAGMENT_ROTATION_SPEED_STEPS = 2;
		const int MIN_JEWEL_FRAGMENT_ROTATION_SPEED = 1;
		const int MAX_JEWEL_FRAGMENT_ROTATION_SPEED = 8;
		const int MIN_JEWEL_FRAGMENT_VELOCITY = 1;
		const int MAX_JEWEL_FRAGMENT_VELOCITY = 8;
		const int JEWEL_FRAGMENT_GRAVITY = 1;
		const int JEWEL_FRAGMENT_SRC_Y = 32 * 9;
		//	魔法の最大レベル
		const int MAX_MAGIC_LEVEL = 2;
		const int MAX_ENERGY = ENERGY_LEVEL_INTERVAL * MAX_MAGIC_LEVEL;
		/// <summary></summary>
		static readonly FCAMRectangle COUNT_POSSIZE = new FCAMRectangle(false, 416, 0, 64, 64);
		const int NUMBER_COUNTS = 3;
		const int NUMBER_MENU_ITEMS = 6;
		/// <summary></summary>
		static readonly FCAMRectangle RESULT_WIN_POSSIZE = new FCAMRectangle(false, 0, 256, 256, 128);
		static readonly FCAMRectangle RESULT_LOSE_POSSIZE = new FCAMRectangle(false, 256, 256, 256, 128);
		static readonly FCAMRectangle RESULT_DRAW_POSSIZE = new FCAMRectangle(false, 0, 384, 256, 128);
		static readonly FCAMRectangle CONTINUE_ASK_POSSIZE = new FCAMRectangle(false, 256, 384, 256, 64);
		static readonly FCAMRectangle CONTINUE_YES_POSSIZE = new FCAMRectangle(false, 416, 192, 64, 32);
		static readonly FCAMRectangle CONTINUE_NO_POSSIZE = new FCAMRectangle(false, 416, 224, 64, 32);
		static readonly FCAMRectangle PAUSE_POSSIZE = new FCAMRectangle(false, 256, 448, 256, 64);
		static readonly FCAMColor BackgroundBlenderColor = new FCAMColor(FCAMColor.MaxValue * 3 / 4, FCAMColor.MaxValue);
		//
		//	フレーム関係
		const int NUMBER_END_FRAMES = 30;
		const int NUMBER_END_SUBFRAMES = 60;
		const int NUMBER_SCENE_FRAMES = 30;
		/// <summary>定数（読み取り）定義しても良いが、４人プレイなどで画面構成が変わる可能性を想定し、画面配置は変数とする。</summary>
		/// 
		FCAMPoint PosVSLabel, PosInfo, PosTimeLabel, PosTime, PosEnergyJewelsLabel, PosScoreLabel, PosJewelsLabel, PosNextLabel;
		FCAMTexture TextureBackground, TextureBackgroundBlender, TextureJewel, TextureText, TextureMagic;
		FCAMDateTime DateTime;
		Player[] Players;
		long StartTime;
		int MaxPlayers{get; set;}
		FCAMRandom Random;
		IFCAMAudioCue CueCount, CueStart, CueBreak, CuePut, CueOjama, CueRotate, CueMagic;
		/// <summary>破片のリスト</summary>
		List<JewelFragment> JewelFragmentList;
		/// <summary></summary>
		List<ChainText> ChainTextList;
		/// <summary></summary>
		List<FadeoutDiagram> FadeoutDiagramList;
		/// <summary></summary>
		BattleState State;
		//	共通で使われる変数
		//
		/// <summary></summary>
		int Frame = 0;
		int SubFrame = 0;
		PlayerKeyState[] PlayerKeys;
		int[] Cursors;
		//	外部からアクセス可能な変数
		public BattleMode Mode;
		/// <summary>
		/// ジュエルに対する落下速度（お邪魔ジュエルと通常のジュエルの落下速度が異なるため）
		/// </summary>
		static int GetJewelFallSpeed(int jewel){
			return(jewel == OBSTACLE_JEWEL ? OBSTACLE_JEWEL_FALL_SPEED : NORMAL_JEWEL_FALL_SPEED);
		}
		static bool IsJewelNormal(sbyte kind){
			return(kind > 0 && kind <= NUMBER_KIND_NORMAL_JEWELS);
		}
		static bool IsJewelObstract(sbyte kind){
			return(kind >= OBSTACLE_JEWEL);
		}
		/// <summary>
		/// 指定された PlayerField 内で消去可能なジュエルを計算し、消去フィールドとして取得する。
		/// (但し画面外のジュエルは消去の対象としない＝コラムスの隠し技風味)
		/// また消去した際に獲得出来るエナジー数や与えられるダメージ数なども計算する。
		/// </summary>
		/// <param name="deleteLength">0 以上の値：消去に必要な最低の結合数。</param>
		/// <returns>消去可能なジュエルの位置を true で示したフィールド。</returns>
		static DeleteField GetDeleteField(PlayerField field, int deleteLength, int chain){
			List<sbyte> magicDeletesJewelList = new List<sbyte>();
			DeleteField deleteField = new DeleteField(field.Size);
			short numberJewels = 0;
			bool magicJewelPutBottom = false;
			deleteField.HVPairJewels = 0;
			deleteField.DiagonallyPairJewels = 0;
			deleteField.DeleteOrScrapeJewels = 0;
			deleteField.DamageJewels = 0;
			deleteField.HardDamageJewels = 0;
			for(int row = FIELD_BLIND_HEIGHT; row < field.Size.Rows; row++){
				for(int column = 0; column < field.Size.Columns; column++){
					int length;
					sbyte jewel = field.Data[column, row].Kind;
					if(jewel == 0)
						continue;
					//
					numberJewels++;
					if(jewel == OBSTACLE_JEWEL || jewel == HARDOBSTACLE_JEWEL)
						continue;
					//	マジックジュエルの時
					if(jewel == MAGIC_JEWEL){
						if(row + 1 >= field.Size.Rows){
							//	一番下にマジックジュエルがある時、攻撃を発生させる。
							magicJewelPutBottom = true;
						}else{
							jewel = field.Data[column, row + 1].Kind;
							if(IsJewelNormal(jewel)){
								magicDeletesJewelList.Add(jewel);
							}
						}
						//	マジックジュエル自身は消去される
						deleteField.Data[column, row] = true;
						continue;
					}
					// 水平方向
					length = 1;
					while(length < deleteLength && column + length < field.Size.Columns){
						if(jewel == field.Data[column + length, row].Kind)
							length++;
						else
							break;
					}
					if(length > 1)
						deleteField.HVPairJewels += length - 1;
					if(length == deleteLength){
						while(length > 0){
							length--;
							if(!deleteField.Data[column + length, row])
								deleteField.DeleteOrScrapeJewels++;
							deleteField.Data[column + length, row] = true;
						}
					}
					// 垂直方向
					length = 1;
					while(length < deleteLength && row + length < field.Size.Rows){
						if(jewel == field.Data[column, row + length].Kind)
							length++;
						else
							break;
					}
					if(length > 1)
						deleteField.HVPairJewels += length - 1;
					if(length == deleteLength){
						while(length > 0){
							length--;
							if(!deleteField.Data[column, row + length])
								deleteField.DeleteOrScrapeJewels++;
							deleteField.Data[column, row + length] = true;
						}
					}
					// 右下斜め
					length = 1;
					while(length < deleteLength && column + length < field.Size.Columns && row + length < field.Size.Rows){
						if(jewel == field.Data[column + length, row + length].Kind)
							length++;
						else
							break;
					}
					if(length > 1)
						deleteField.DiagonallyPairJewels += length - 1;
					if(length == deleteLength){
						while(length > 0){
							length--;
							if(!deleteField.Data[column + length, row + length])
								deleteField.DeleteOrScrapeJewels++;
							deleteField.Data[column + length, row + length] = true;
						}
					}
					// 左下斜め
					length = 1;
					while(length < deleteLength && column - length >= 0 && row + length < field.Size.Rows){
						if(jewel == field.Data[column - length, row + length].Kind)
							length++;
						else
							break;
					}
					if(length > 1)
						deleteField.DiagonallyPairJewels += length - 1;
					if(length == deleteLength){
						while(length > 0){
							length--;
							if(!deleteField.Data[column - length, row + length])
								deleteField.DeleteOrScrapeJewels++;
							deleteField.Data[column - length, row + length] = true;
						}
					}
				}
			}
			//	お邪魔ジュエルの消去
			//	リスト化してあとで行なわないと、削除したお邪魔ジュエルの周辺も削除してしまう。
			//	⇒但しそれは事前の for(int i = 0 前の処理で大丈夫なはず
			DeleteField deleteDamageField = new DeleteField(deleteField.Size);
			for(int row = FIELD_BLIND_HEIGHT; row < field.Size.Rows; row++){
				for(int column = 0; column < field.Size.Columns; column++){
					if(!deleteField.Data[column, row])
						continue;
					if(!IsJewelNormal(field.Data[column, row].Kind))
						continue;
					for(int i = 0; i < 4; i++){
						int targetColumn = column + (i == 0 ? -1 : (i == 2 ? 1 : 0));
						int targetRow = row + (i == 1 ? -1 : (i == 3 ? 1 : 0));
						if(field.GetJewel(targetColumn, targetRow) == OBSTACLE_JEWEL || field.GetJewel(targetColumn, targetRow) == HARDOBSTACLE_JEWEL){
							deleteDamageField.Data[targetColumn, targetRow] = true;
						}
					}
				}
			}
			for(int row = FIELD_BLIND_HEIGHT; row < field.Size.Rows; row++){
				for(int column = 0; column < field.Size.Columns; column++){
					if(!deleteDamageField.Data[column, row])
						continue;
					//	例えお邪魔ジュエルの周囲で消去を行なっても、
					//	それが強固ジュエルならお邪魔ジュエルに削れるだけなので、
					//	加算するお邪魔ジュエルを変える。
					if(field.GetJewel(column, row) == HARDOBSTACLE_JEWEL){
						deleteField.HardDamageJewels++;
					}else{
						deleteField.DamageJewels++;
					}
					deleteField.Data[column, row] = deleteDamageField.Data[column, row];
				}
			}
			//	実際に消去されるジュエル数
			deleteField.Energy = deleteField.DeleteOrScrapeJewels;
			//	ダメージの計算。
			if(deleteField.DeleteJewels >= deleteLength){
				deleteField.Damage = FCAMMath.Max(0, (deleteField.DeleteJewels - deleteLength) * chain / 2
													 + (chain <= 1 ? 0 : deleteField.Size.Columns / (chain - 1) + (chain - 2))
												 );
#if DEBUG
				if(deleteField.Damage > 0){
					FCAMDebug.NOP();
				}
#endif
			}
			//deleteField.Damage = FCAMMath.Max(0, DAMAGE_BOARD[chain] + (deleteField.DeleteJewels - deleteLength) * chain);
			//if(deleteField.DeleteJewels > deleteLength){
			//    deleteField.Damage = (int)(FCAMMath.Pow(
			//        FCAMMath.Pow(
			//            (deleteField.DeleteJewels - deleteLength * chain + (chain - 1) * DAMAGE_CHAIN_RATE), DAMAGE_DELETE_RATE
			//        ) * chain, DAMAGE_RATE
			//    ) * DAMAGE_BASE_RATE);
			//    //deleteField.Damage = FCAMMath.Max(0, (int)FCAMMath.Pow((FCAMMath.Pow(deleteField.DeleteJewels, DAMAGE_DELETE_RATE) * DAMAGE_BASE_RATE - (chain > 1 ? 0 : deleteLength * DAMAGE_BASE_RATE)) * (DAMAGE_CHAIN_RATE * chain), DAMAGE_RATE));
			//}
			//	マジックジュエルに関する処理。
			//	※必ずお邪魔より先に行なうこと。消去した周りのお邪魔ジュエルが消えないようにするため。
			//	※必ずエナジー＆ダメージ換算前に行なうこと。マジックジュエルによる消去でエナジーを加算させないためと
			//		ダメージを加算出来るようにするため
			if(magicJewelPutBottom){
				//	マジックジュエルが最下部に置かれていたら、
				deleteField.Damage += numberJewels / 3;
			}
			foreach(sbyte jewel in magicDeletesJewelList){
				for(int row = FIELD_BLIND_HEIGHT; row < field.Size.Rows; row++){
					for(int column = 0; column < field.Size.Columns; column++){
						if(jewel == field.Data[column, row].Kind){
							deleteField.Data[column, row] = true;
							deleteField.DeleteOrScrapeJewels++;
							//	魔法でお邪魔ジュエルは消せないため、もしそれがあったらバグとする。
							if(jewel == OBSTACLE_JEWEL){
								deleteField.DamageJewels++;
								Error.ID_00006.Throw();
							}
							if(jewel == HARDOBSTACLE_JEWEL){
								//	ハードジュエルも一発で消去出来るので、削り数には加えない
								Error.ID_00007.Throw();
							}
						}
					}
				}
			}
			return(deleteField);
		}
		/// <summary>
		/// 指定されたプレイヤーのフィールド内のジュエルを消去フィールドを元に消去する。
		/// </summary>
		/// <returns>消去されたジュエル数</returns>
		static int DeleteJewelsIn(PlayerField field, DeleteField deleteField){
			int numberDeletes = 0;
			for(int row = 0; row < field.Size.Rows; row++){
				for(int column = 0; column < field.Size.Columns; column++){
					if(!deleteField.Data[column, row])
						continue;
					if(field.Data[column, row].Delete())
						numberDeletes++;
				}
			}
			return(numberDeletes);
		}
		/// <summary>
		/// PlayerField 内のジュエルで、空中に浮いているものを落下させる。
		/// max = false なら1段階落下(1行分ではない)、true なら一番したまで落下させる。
		/// </summary>
		/// <returns>TRUE: 落下が行なわれた。</returns>
		static bool FallFloatingJewels(PlayerField field, bool max){
			bool droped = false, loop;
			//	下から順に落下処理を行なわないと渋滞する。
			do{
				loop = false;
				for(int row = field.Size.Rows - 1; row	> 0; row--){
					for(int column = 0; column < field.Size.Columns; column++){
						if(field.Data[column, row].PosGap.Y < 0){
							field.Data[column, row].PosGap.Y += field.Data[column, row].FallSpeed;
							droped = true;
							loop = true;
						}
						if(field.Data[column, row].Kind != 0 || field.Data[column, row - 1].Kind == 0)
							continue;
						//	落下させるジュエルが Gap による落下中なら待機
						if(field.Data[column, row - 1].PosGap.Y < 0)
							continue;
						//	落下した場合
						field.Data[column, row].Put(field.Data[column, row - 1].Kind, field.Data[column, row - 1].Blender, field.Data[column, row - 1].FallSpeed);
						field.Data[column, row - 1].Delete();
						field.Data[column, row].PosGap.Y = -JEWEL_SIZE.Height + field.Data[column, row].FallSpeed;
						droped = true;
						loop = true;
					}
				}
			}while(loop && max);
			return(droped);
		}
		/// <summary>
		/// 加算スコアを計算
		/// </summary>
		static int CalcScorePoint(int deletes, int chain, int level, int deleteLength, int damageJewels){
			return(((deletes - deleteLength + chain) * 30 + damageJewels * 10) * level);
		}
		/// <summary>
		/// 
		/// </summary>
		~Battle(){
			foreach(Player player in Players){
				player.Dispose();
			}
		}
		/// <summary>
		/// 
		/// </summary>
		public Battle(bool vs){
			MaxPlayers = vs ? 2 : 4;
			//	解放時に Load されていない時エラーするので、ここでは 0 配列を作成。
			Players = new Player[0];
			Cursors = new int[MaxPlayers];
			PlayerKeys = new PlayerKeyState[MaxPlayers];
			JewelFragmentList = new List<JewelFragment>();
			ChainTextList = new List<ChainText>();
			FadeoutDiagramList = new List<FadeoutDiagram>();
			Random = new FCAMRandom();
		}
		protected override void Load(FCAMGraphics g, FCAMInput input, FCAMContentsManager manager){
#if DEBUG
			if(FCAMDebug.IsDebug){
				Base.Text = "Battle";
			}
#endif
			PlayerKeys[0] = new PlayerKeyState(new FCAMInputState[]{
				Common.KeyLeft, Common.KeyUp, Common.KeyRight, Common.KeyDown, Common.KeyA, Common.KeyB, Common.KeyL, Common.KeyR
			});
			PlayerKeys[1] = new PlayerKeyState(new FCAMInputState[]{
				Common.KeyLeft2, Common.KeyUp2, Common.KeyRight2, Common.KeyDown2, Common.KeyA2, Common.KeyB2, Common.KeyL2, Common.KeyR2
			});
			//	TextureBackground = manager.MakeTexture(FCAMStrings.Format(FILEPATH_BACKGROUND_FORMAT, Common.Stage));
			string path = FCAMStrings.Format(FILEPATH_BACKGROUND_FORMAT, Memory.GetClearedCharactersAtStage(Common.StageOnFile));
			//	パスが存在しない、もしくはストーリーモードでない時は0番を使用。
			if(!manager.Exist(path) || !IsStory){
				path = FCAMStrings.Format(FILEPATH_BACKGROUND_FORMAT, 0);
			}
			TextureBackground = manager.MakeTexture(path);
			TextureBackgroundBlender = manager.MakeTexture(FILEPATH_BACKGROUND_BLENDER);
			TextureJewel = manager.MakeTexture(FILEPATH_JEWEL);
			TextureText = manager.MakeTexture(FILEPATH_TEXT);
			TextureMagic = manager.MakeTexture(FILEPATH_MAGIC);
			CueCount = manager.MakeCue(FILEPATH_SE_COUNT);
			CueStart = manager.MakeCue(FILEPATH_SE_START);
			CueBreak = manager.MakeCue(FILEPATH_SE_BREAK);
			CuePut = manager.MakeCue(FILEPATH_SE_PUT);
			CueOjama = manager.MakeCue(FILEPATH_SE_OJAMA);
			CueRotate = manager.MakeCue(FILEPATH_SE_ROTATE);
			CueMagic = manager.MakeCue(FILEPATH_SE_MAGIC);
			PosNextLabel.Set(288, 96);
			PosVSLabel.Set(304, 8);
			PosInfo.Set(256, 48);
			PosEnergyJewelsLabel.Set(272, 224);
			PosTimeLabel.Set(288, 296);
			PosTime.Set(256, 312);
			PosScoreLabel.Set(280, 368);
			PosJewelsLabel.Set(272, 432);
			Players = new Player[MaxPlayers];
			for(sbyte i = 0; i < MaxPlayers; i++){
				if(Players[i] != null)
					continue;
				Players[i] = new Player(manager, this, i, "", PlayerKeys[i]);
			}
			switch(Mode){
				case BattleMode.Free1P:{
					Players[0].Kind = PlayerKind.Human;
					Players[1].Kind = PlayerKind.NonParticipant;
					Mode = BattleMode.Free;
					break;
				}
				case BattleMode.Free2P:{
					Players[0].Kind = PlayerKind.NonParticipant;
					Players[1].Kind = PlayerKind.Human;
					Mode = BattleMode.Free;
					break;
				}
				case BattleMode.Test:{
					Players[1].Kind = PlayerKind.CPU;
					Players[1].SetCharacterID(manager, (sbyte)(2 + Game.MemoryData.CharaShift));
					break;
				}
				case BattleMode.CPUBattleTest:{
					Players[0].Kind = PlayerKind.CPU;
					Players[1].Kind = PlayerKind.CPU;
					Players[0].SetCharacterID(manager, (sbyte)(0 + Game.MemoryData.CharaShift));
					Players[1].SetCharacterID(manager, (sbyte)(1 + Game.MemoryData.CharaShift));
					break;
				}
				case BattleMode.CPUTest:{
					Players[0].Kind = PlayerKind.CPU;
					Players[1].Kind = PlayerKind.NonParticipant;
					Players[0].SetCharacterID(manager, (sbyte)(2 + Game.MemoryData.CharaShift));
					break;
				}
			}
			if(!IsStory){
				ChangeState(BattleState.Menu);
			}else{
				ChangeState(BattleState.BattleStart);
			}
		}
		protected override void Unload(){
			FCAMSystem.DisposeClass(out CueMagic);
			FCAMSystem.DisposeClass(out CueRotate);
			FCAMSystem.DisposeClass(out CueOjama);
			FCAMSystem.DisposeClass(out CuePut);
			FCAMSystem.DisposeClass(out CueBreak);
			FCAMSystem.DisposeClass(out CueStart);
			FCAMSystem.DisposeClass(out CueCount);
			FCAMSystem.DisposeClass(out TextureMagic);
			FCAMSystem.DisposeClass(out TextureText);
			FCAMSystem.DisposeClass(out TextureJewel);
			FCAMSystem.DisposeClass(out TextureBackgroundBlender);
			FCAMSystem.DisposeClass(out TextureBackground);
		}
		/// <summary>
		/// デバッグキーメモ（KeyDebug2 押しっぱ＋？？？）
		/// ・L or R: プレイヤー1 or 2のエナジー+1
		/// ・A or B: プレイヤー1 or 2にダメージ+1
		/// ・3 + L or R: プレイヤー1 or 2のエナジーレベル+1
		/// ・3 + A or B: プレイヤー1 or 2を倒す
		/// ・矢印キー: CPUだろうと強制的に操作。
		/// ・3 + ←→: プレイヤー1 or 2のジュエル+LEVEL_UP_INTERVAL
		/// </summary>
		protected override void Calc(FCAMInput input, FCAMContentsManager manager){
			//
			if(Common.KeyDebug2.Press){
				switch(State){
					case BattleState.BattleNow:{
						if(Common.KeyDebug3.Press){
							if(Common.KeyL){
								Common.KeyL.SetTime(1);
								Players[0].EnergyLevel++;
							}
							if(Common.KeyR){
								Common.KeyR.SetTime(1);
								Players[1].EnergyLevel++;
							}
							if(Common.KeyA){
								Common.KeyA.SetTime(1);
								Players[0].State = PlayerState.Lose;
								State = BattleState.BattleBeforeEnd;
							}
							if(Common.KeyB){
								Common.KeyB.SetTime(1);
								Players[1].State = PlayerState.Lose;
								State = BattleState.BattleBeforeEnd;
							}
							if(Common.KeyLeft){
								Common.KeyLeft.SetTime(1);
								Players[0].Jewels += LEVEL_UP_INTERVAL;
							}
							if(Common.KeyRight){
								Common.KeyRight.SetTime(1);
								Players[1].Jewels += LEVEL_UP_INTERVAL;
							}
						}else{
							if(Common.KeyL.Press){
								Common.KeyL.SetTime(1);
								Players[0].Energy++;
							}
							if(Common.KeyR.Press){
								Common.KeyR.SetTime(1);
								Players[1].Energy++;
							}
							if(Common.KeyA.Press){
								Common.KeyA.SetTime(1);
								Players[0].Damage++;
							}
							if(Common.KeyB.Press){
								Common.KeyB.SetTime(1);
								Players[1].Damage++;
							}
						}
						break;
					}
				}
			}
			//	初期化処理
			foreach(Player player in Players){
				player.ResetKeys();
				if(State == BattleState.Menu){
					if(player.Keys.Left.Accel)
						player.KeyLeft = true;
					if(player.Keys.Up)
						player.KeyUp = true;
					if(player.Keys.Right.Accel)
						player.KeyRight = true;
					if(player.Keys.Down)
						player.KeyDown = true;
					if(player.Keys.Rotate)
						player.KeyRotate = true;
					if(player.Keys.Pause)
						player.KeyPause = true;
				}else{
					//	CPUでも一時的に操作は認める（両方 CPU でポーズをかけ、その後終了させたい時など）
					//	
					if(player.Keys.Left.Accel)
						player.KeyLeft = true;
					if(player.Keys.Up)
						player.KeyUp = true;
					if(player.Keys.Right.Accel)
						player.KeyRight = true;
					if(player.Keys.Down.Press)
						player.KeyDown = true;
					if(player.Keys.Rotate)
						player.KeyRotate = true;
					if(player.Keys.Pause)
						player.KeyPause = true;
#if ATTACK_MAGIC
					if(player.Keys.Attack)
						player.KeyAttack = true;
#endif
					if(player.Keys.Support)
						player.KeySupport = true;
				}
			}
			//	ゲーム状態ごとの分岐処理
			switch(State){
				//	メニュー時
				case BattleState.Menu:{
					//	バーサスモード以外はメニュー無し
					int numberPlayers = 0;
					bool choosing = false;
					if(Mode != BattleMode.VS){
						NextBattle(manager);
						break;
					}
					foreach(Player player in Players){
						if(player.Kind != PlayerKind.NonParticipant || player.State != PlayerState.Wait){
							numberPlayers++;
						}
					}
					for(int i = 0; i < Players.Length; i++){
						Player source, target;
						sbyte characterID, kind, height, jewel;
						int targetCursor;
						bool changeCharacterID = false, changeKind = false;
						source = Players[i];
						//	自分の設定が完了したら、他の設定も可能とする
						//	（１人プレイヤーだけでもプレイヤー２の設定を変更出来るようにするため：その逆もまた然り）。
						if(i == 0 && Players[i].State == PlayerState.Wait){
							target = Players[targetCursor = 1];
						}else if(i == 1 && Players[i].State == PlayerState.Wait){
							target = Players[targetCursor = 0];
						}else{
							target = Players[targetCursor = i];
						}
						//	※targetCursor の仕様変更
						//	targetCursor はそのままでよし（1P, 2P それぞれのカーソルを描画しているため）
						targetCursor = i;
						//	
						Common.ReadyPlayCueSelect(Cursors[targetCursor]);
						if(source.KeyUp){
							Cursors[targetCursor]--;
						}
						if(source.KeyDown){
							Cursors[targetCursor]++;
						}
						//Cursors[targetCursor] = FCAMMath.Within(0, Cursors[targetCursor], NUMBER_MENU_ITEMS - 1);
						Cursors[targetCursor] = (Cursors[targetCursor] + NUMBER_MENU_ITEMS) % NUMBER_MENU_ITEMS;
						Common.PlayCueSelect(Cursors[targetCursor]);
						characterID = sbyte.MinValue;
						kind = (sbyte)target.Kind;
						height = sbyte.MinValue;
						jewel = sbyte.MinValue;
						if(source.KeyLeft){
							switch(Cursors[targetCursor]){
								case 0:{
									characterID = (sbyte)(target.CharacterID - 1);
									changeCharacterID = true;
									break;
								}
								case 1:{
									kind--;
									changeKind = true;
									break;
								}
								case 2:{
									height = (sbyte)(target.StartHeight - 1);
									Common.CueSelect.Play();
									break;
								}
								case 3:{
									jewel = (sbyte)(target.StartJewel - 1);
									Common.CueSelect.Play();
									break;
								}
							}
						}
						if(source.KeyRight){
							switch(Cursors[targetCursor]){
								case 0:{
									characterID = (sbyte)(target.CharacterID + 1);
									changeCharacterID = true;
									break;
								}
								case 1:{
									kind++;
									changeKind = true;
									break;
								}
								case 2:{
									height = (sbyte)(target.StartHeight + 1);
									Common.CueSelect.Play();
									break;
								}
								case 3:{
									jewel = (sbyte)(target.StartJewel + 1);
									Common.CueSelect.Play();
									break;
								}
							}
						}
						characterID = (sbyte)FCAMMath.Within(Game.MemoryData.CharaShift, characterID, Game.MemoryData.ClearedCharacters);
						if(changeCharacterID){
							Common.CueSelect.Play();
							target.SetCharacterID(manager, characterID);
						}
						if(changeKind){
							Common.CueSelect.Play();
							//kind = (sbyte)FCAMMath.Within(0, kind, (numberPlayers > 1) ? 2 : 1);
							kind = (sbyte)FCAMMath.Within((sbyte)PlayerKind.MIN, kind, (sbyte)PlayerKind.MAX);
							target.Kind = (PlayerKind)kind;
#if TRACE
							FCAMDebug.Inform(".Scenes.Battle.Calc()-CHANGE KIND: Player[{0}].Kind={1}", target.PlayerId, target.Kind);
#endif
						}
						//	プレイヤー数が1人（一方が不参加）の場合、（他方が）必ず参加者とするようにする。
						if(numberPlayers <= 1 && target.State != PlayerState.Wait && target.Kind == PlayerKind.NonParticipant){
							target.Kind = PlayerKind.CPU;
#if TRACE
							FCAMDebug.Inform(".Scenes.Battle.Calc()-MODIFY KIND: Player[{0}].Kind={1}", target.PlayerId, target.Kind);
#endif
						}
						if(height != sbyte.MinValue){
							height = (sbyte)FCAMMath.Within(0, height, FIELD_TABLESIZE.Rows - FIELD_BLIND_HEIGHT - target.DeleteLength);
							target.StartHeight = height;
						}
						if(jewel != sbyte.MinValue){
							jewel = (sbyte)FCAMMath.Within(0, jewel, 2);
							target.StartJewel = (JewelKind)jewel;
						}
						//
						if(source.KeyPause){
							if(source.State == PlayerState.Wait){
								source.State = PlayerState.Normal;
								Common.CueCancel.Play();
							}else{
								Cursors[targetCursor] = NUMBER_MENU_ITEMS - 1;
								Common.CueCancel.Play();
							}
						}
						if(source.KeyRotate){
							switch(Cursors[targetCursor]){
								case 4:{
									target.State = PlayerState.Wait;
									Common.CueDecide.Play();
									break;
								}
								case 5:{
									ChangeState(BattleState.SceneEnd);
									Common.CueDecide.Play();
									break;
								}
							}
						}
						//
						if(source.State != PlayerState.Wait){
							choosing = true;
						}
					}
					if(!choosing){
						NextBattle(manager);
					}
					break;
				}
				//
				case BattleState.BattleStart:
					if(Common.KeyDebug2){
						Frame = Base.FPS * NUMBER_COUNTS;
					}
					if(Frame % Base.FPS == 0){
						if(Frame / Base.FPS >= NUMBER_COUNTS){
							CueStart.Play();
						}else{
							CueCount.Play();
						}
					}
					if(Frame >= Base.FPS * NUMBER_COUNTS){
						ChangeState(BattleState.BattleNow);
						foreach(Player player in Players){
							//	ゲーム開始時に既にジュエルは全てセットされているはずなので、
							//	フィールドの上でもジュエルを見せている場合は、このセット処理は不要
							//	※逆に見せない場合は、ゲーム開始時に"NEXT"が始めのジュエルである必要があるため、
							//	　SetNextJewelSet() を呼ぶ必要がある。
							//		⇒必ず呼ぶようにする（フィールドの上で見えても、そこでは全体が見えないｔめ）
							//			⇒その代わり、ゲーム開始前のジュエルセットのリストは、一番初めが空のジュエルセットにして対応
//#if !SHOW_ALL_CONTROL_JEWELS && !SHOW_TRANSPARENT_CONTROL_JEWELS
							player.SetNextJewelSet();
//#endif
							//	ジュエルをセットしてからCPUを計算する
							player.ResetMoveRange();
							player.RecalcCPUAim(this);
						}
						StartTime = FCAMSystem.Ticks;
					}
					//	カウント・スタート音を鳴らすために、フレームの加算は最後にする。
					Frame++;
					return;
				//
				case BattleState.SceneEnd:
					if(Frame < NUMBER_SCENE_FRAMES){
						Frame++;
						break;
					}
					PopScene();
					break;
				//	バトル終了前（勝利か敗北か引き分けかを判断）
				case BattleState.BattleBeforeEnd:{
					int numberLoser = 0;
					int numberPlayers = 0;
					//	プレイヤーごとの描画
					foreach(Player player in Players){
						if(player.IsParticipant){
							numberPlayers++;
						}
						if(player.State == PlayerState.Lose){
							numberLoser++;
						}
					}
					if(numberPlayers == 1){
						foreach(Player player in Players){
							player.State = PlayerState.Normal;
						}
					}else if(numberLoser == numberPlayers){
						foreach(Player player in Players){
							player.State = PlayerState.Draw;
						}
					}else{
						foreach(Player player in Players){
							if(player.State != PlayerState.Lose)
								player.State = PlayerState.Win;
						}
					}
					ChangeState(BattleState.BattleEnd);
					break;
				}
				//	バトル終了
				case BattleState.BattleEnd:
					if(Frame < NUMBER_END_FRAMES){
						Frame++;
						break;
					}
					if(SubFrame < NUMBER_END_SUBFRAMES){
						SubFrame++;
						break;
					}
					foreach(Player player in Players){
						if(IsStory && player.PlayerId != ProtagonistPlayerID)
							continue;
						if(IsAutoProcessOnBattleEndInStory(player.PlayerId)){
							if(player.KeyRotate){
								if(player.State == PlayerState.Win){
									ChangeState(BattleState.SceneEnd);
								}else if(player.State == PlayerState.Draw){
									NextBattle(manager);
								}
							}
						}else{
							if(player.KeyUp){
								player.Select--;
							}
							if(player.KeyDown){
								player.Select++;
							}
							player.Select = FCAMMath.Within<sbyte>(0, player.Select, 1);
							if(player.KeyRotate){
								switch(player.Select){
									case 0:{
										NextBattle(manager);
										break;
									}
									case 1:{
										if(Mode == BattleMode.VS){
											ChangeState(BattleState.Menu);
										}else{
											ChangeState(BattleState.SceneEnd);
										}
										break;
									}
								}
							}
						}
					}
					break;
				//
				//
				case BattleState.Pause:
					foreach(Player player in Players){
						if(player.KeyPause){
							ChangeState(BattleState.BattleNow);
							continue;
						}
						if(Mode != BattleMode.VS){
							continue;
						}
						//	バーサスモードでは、ポーズ時に継続するか否かの選択が可能
						if(player.KeyUp){
							player.Select--;
						}
						if(player.KeyDown){
							player.Select++;
						}
						player.Select = FCAMMath.Within<sbyte>(0, player.Select, 1);
						if(player.KeyRotate){
							switch(player.Select){
								case 0:{
									ChangeState(BattleState.BattleNow);
									break;
								}
								case 1:{
									ChangeState(BattleState.Menu);
									break;
								}
							}
						}
					}
					break;
				case BattleState.BattleNow:{
					DateTime.Ticks = FCAMSystem.Ticks - StartTime;
					//	連鎖テキストの処理
					//	下記ジュエルフラグメントと同様の理由で処理を行なう。
					for(int i = ChainTextList.Count - 1; i >= 0; i--){
						ChainText text = ChainTextList[i];
						text.Frame++;
						text.CenterPos.Y += NUMBER_CHAIN_TEXT_RISE_DOWN_STEPS;
						if(text.Frame >= NUMBER_CHAIN_TEXT_FRAMES){
							ChainTextList.RemoveAt(i);
						}
					}
					//	ジュエルフラグメントの処理
					//	最大値 - 1 から順繰りにやることで、
					//	ジュエルフラグメントを削除した際のインデックスのずれ問題を発生させない。
					for(int i = JewelFragmentList.Count - 1; i >= 0; i--){
						JewelFragment fragment = JewelFragmentList[i];
						fragment.CenterPos.X += fragment.Velocity.X;
						fragment.CenterPos.Y += fragment.Velocity.Y;
						fragment.Velocity.Y += JEWEL_FRAGMENT_GRAVITY;
						fragment.Rotation += fragment.RotationSpeed;
						//	下方向で画面外になったフラグメントはリストから削除する。
						if(fragment.CenterPos.Y >= Base.GraphicsHeight + JEWEL_SIZE.Height){
							JewelFragmentList.RemoveAt(i);
						}
					}
					//	フェードアウトダイアログの処理
					//	上記ジュエルフラグメントと同様の理由で処理を行なう。
					for(int i = FadeoutDiagramList.Count - 1; i >= 0; i--){
						FadeoutDiagram diagram = FadeoutDiagramList[i];
						if(diagram.Calc()){
							FadeoutDiagramList.RemoveAt(i);
						}
					}
					//	プレイヤーの処理
					bool isPinch = false;
					foreach(Player player in Players){
						//	非プレイヤー（誰もいない）なら、処理はしない。
						if(!player.IsParticipant){
							continue;
						}
						//	ピンチなら、あとでBGMを早くする
						if(player.Field.IsPinch(player.NowJewelSet)){
							isPinch = true;
						}
						//	ポーズ
						if(player.KeyPause){
							ChangeState(BattleState.Pause);
							continue;
						}
						//	計算
						player.Calc();
						//	CPUが操作を行なうかどうか。デバッグキーが押されている場合は無視される。
						bool cpuOperation = player.IsCPU && !Common.KeyDebug2.Press;
						if(cpuOperation){
							//	CPU操作なら入力キーを全てキャンセルする。
							player.ResetKeys();
						}
						switch(player.State){
							case PlayerState.Normal:
								//	操作中
								//	もしマジックジュエルがあるなら連鎖処理（消去処理）にする。
								for(int row = 0; row < player.Field.Size.Rows; row++){
									for(int column = 0; column < player.Field.Size.Columns; column++){
										if(player.Field.Data[column, row].Kind == MAGIC_JEWEL){
											player.State = PlayerState.Chain;
										}
									}
								}
								if(player.State == PlayerState.Chain)
									break;
								//	プレイヤーの操作処理は事前に行なってるので、CPU の操作処理のみ
								//	但し CPU が操作可能状態である時。
								if(cpuOperation && player.CPU.NextOperate()){
									//	CPU の操作待ち時間をリセット
									player.CPU.ResetOperateWait();
									switch(player.CPUAim.UseMagic){
										case MagicKind.Nothing:{
											break;
										}
#if ATTACK_MAGIC
										case MagicKind.Attack:{
											player.KeyAttack = true;
											player.CPUAim.UseMagic = MagicKind.Used;
											break;
										}
#endif
										case MagicKind.Support:{
											player.KeySupport = true;
											player.CPUAim.UseMagic = MagicKind.Used;
											break;
										}
										case MagicKind.Used:{
											player.ResetMoveRange();
											player.RecalcCPUAim(this);
											break;
										}
										default:{
											Error.ID_00002.Throw();
											break;
										}
									}
									switch(FCAMMath.GetSign(player.CPUAim.Point.Column - player.JewelSetPos.Column)){
										case -1:{
											player.KeyLeft = true;
											break;
										}
										case 1:{
											player.KeyRight = true;
											break;
										}
										case 0:{
											//	回転を終えてなければ回転
											if(player.CPUAim.Rotations > 0){
												player.KeyRotate = true;
												player.CPUAim.Rotations--;
											}else{
												//	落下処理
												//	(CPU計算で割り出した結果でのピンチ度とプレイヤーの性質に合わせて落下方法を決める)
												switch(player.CPU.DropType){
													case CPUDropType.Instant:{
														switch(player.CPUAim.PinchLevel){
															case CPUPinchLevel.Absolute:{
																break;
															}
															case CPUPinchLevel.Probability:{
																player.KeyDown = true;
																break;
															}
															default:{
																player.KeyUp = true;
																break;
															}
														}
														break;
													}
													case CPUDropType.Normal:{
														switch(player.CPUAim.PinchLevel){
															case CPUPinchLevel.Absolute:{
																break;
															}
															default:{
																player.KeyDown = true;
																break;
															}
														}
														break;
													}
												}
											}
											break;
										}
									}
								}
								//	
								//	操作
								int addPutWait = 0;
								int addGapY = 0;
								int addScore = 0;
								//	左移動
								if(player.KeyLeft){
									if(player.CanMove(-1, 0)){
										player.State = PlayerState.MoveLeft;
									}else if(player.IsCPU){
										//	移動不可の場合は、コンピュータは目的地を再計算する。
										player.MoveRange.StartColumn = player.JewelSetPos.Column;
										player.MoveRange.Row = player.JewelSetPos.Row;
										player.RecalcCPUAim(this);
									}
								}
								//	右移動
								if(player.KeyRight){
									if(player.CanMove(1, 0)){
										player.State = PlayerState.MoveRight;
									}else if(player.IsCPU){
										//	移動不可の場合は、コンピュータは目的地を再計算する。
										player.MoveRange.EndColumn = player.JewelSetPos.Column;
										player.MoveRange.Row = player.JewelSetPos.Row;
										player.RecalcCPUAim(this);
									}
								}
								//	回転
								if(player.KeyRotate){
									player.NowJewelSet.Rotate();
									CueRotate.Play();
								}
								//	落下
								if(player.KeyDown){
									addGapY = JEWEL_VERT_MOVE_SPEED;
									addPutWait = 2;
									addScore = 1;
								}else{
									if(player.KeyUp){
										addGapY = JEWEL_VERT_MOVE_SPEED;
										addPutWait = 1;
										addScore = 2;
									}else{
										addGapY = GetFallSpeed(player);
										addPutWait = 1;
									}
								}
								//	落下の処理。↑キーが押されてる時は即席落下
								do{
									player.JewelSetPosGap.Y += addGapY;
									if(player.JewelSetPosGap.Y >= 0){
										if(player.MoveJewelSetRow(1)){
											player.JewelSetPosGap.Y -= JEWEL_SIZE.Height;
											player.Score += addScore;
											player.ResetPutWait();
										}else{
											player.JewelSetPosGap.Y = 0;
											player.PutWait += addPutWait;
										}
									}else{
										player.Score += addScore;
										player.ResetPutWait();
									}
									//	設置
									if(player.PutWait > GetPutTime(player)){
										player.PutJewelSet();
										player.SetNowJewelSetEmpty();
										player.State = PlayerState.Chain;
										player.Chain = 0;
										player.ResetPutWait();
										CuePut.Play();
									}
								}while(player.KeyUp && player.State == PlayerState.Normal);
								//	魔法を唱える
								if((player.KeyAttack || player.KeySupport) && !player.IsMagicEffecting()){
									if(player.EnergyLevel > 0){
										List<Player> targetList = new List<Player>();
										foreach(Player target in Players){
											if(target != player){
												targetList.Add(target);
											}
										}
#if ATTACK_MAGIC
										if(player.KeyAttack)
										{
											player.CastAttackMagic(player, targetList.ToArray());
										}
										else
#endif
										if(player.KeySupport){
											player.CastSupportMagic(player, targetList.ToArray());
										}
										player.EnergyLevel--;
										player.AddFadeoutDiagramAtEnergy();
										CueMagic.Play();
									}else{
										//	魔法を唱えるためのエナジーが足りない時
									}
								}
								break;
							case PlayerState.Chain:
								//	チェック中
								int numberDeletes;
								DeleteField deleteField;
								if(FallFloatingJewels(player.Field, false))
									break;
								deleteField = GetDeleteField(player.Field, player.DeleteLength, player.Chain + 1);
								numberDeletes = DeleteJewelsAndCalcIn(player, deleteField);
								if(numberDeletes == 0){
									player.State = PlayerState.Check;
								}else{
									CueBreak.Frequency = (int)(FCAMMath.Pow(CHAIN_SE_SPEEDUP_RATE, player.Chain - 1) * CueBreak.DefaultFrequency);
									CueBreak.Play();
								}
								//	フリーモードではエナジーはない。
								if(Mode != BattleMode.Free){
									player.Energy += deleteField.Energy;
								}
								//	まずは自分のダメージを相殺。
								int recover = FCAMMath.Min(player.Damage, deleteField.Damage);
								player.Damage -= recover;
								deleteField.Damage -= recover;
								//	対戦相手にダメージを与える。
								foreach(Player target in Players){
									if(player == target)
										continue;
									target.Damage += deleteField.Damage;
								}
								break;
							case PlayerState.Check:
								bool falled = FallFloatingJewels(player.Field, false);
								if(falled || player.Damage > 0){
									List<sbyte> columnList = new List<sbyte>();
									for(sbyte column = 0; column < FIELD_TABLESIZE.Columns; column++){
										columnList.Add(column);
									}
									//	ダメージがある時
									while(player.Damage > 0 && columnList.Count > 0){
										int i = player.DamageRandom.NextInt(columnList.Count);
										int column = columnList[i];
										columnList.RemoveAt(i);
										if(player.Field.Data[column, 0].Kind == 0){
											//	ダメージを受ける（お邪魔ジュエルを落とす。）
											player.Field.Data[column, 0].Put(OBSTACLE_JEWEL);
											player.Damage--;
											CueOjama.Play();
										}
									}
								}else{
									//	チェック処理完了
									//	お邪魔ジュエルの落下速度を通常に戻すことで、以降の落下で通常ジュエルとの差異を無くす。
									player.DoToAllJewelsInField(
										delegate(ref JewelData jewel){
											jewel.FallSpeed = NORMAL_JEWEL_FALL_SPEED;
										}
									);
									//	次のジュエルへ行く
									player.SetNextJewelSet();
									//	CPUなら思考
									if(player.IsCPU){
										player.ResetMoveRange();
										player.RecalcCPUAim(this);
									}
									player.State = PlayerState.Normal;
								}
								if((!falled && player.Damage > 0) || player.State == PlayerState.Normal){
									//	落下してないがダメージが蓄積されている場合 or プレイヤー状態が変わった時、出口が塞がっていたら負けとする。
									if(player.Field.Data[JEWEL_SET_START_POS_COLUMN, 2].Kind != 0){
										player.State = PlayerState.Lose;
										State = BattleState.BattleBeforeEnd;
									}
								}
								break;
						}
						//	水平方向の移動がスムーズになるよう、プレイヤーの状態分岐のあとに水平方向の処理を行なう。
						switch(player.State){
							case PlayerState.MoveLeft:
								player.JewelSetPosGap.X -= JEWEL_HORZ_MOVE_SPEED;
								if(player.JewelSetPosGap.X <= -JEWEL_SIZE.Width){
									player.MoveJewelSetColumn(-1);
									player.State = PlayerState.Normal;
								}
								break;
							case PlayerState.MoveRight:
								player.JewelSetPosGap.X += JEWEL_HORZ_MOVE_SPEED;
								if(player.JewelSetPosGap.X >= JEWEL_SIZE.Width){
									player.MoveJewelSetColumn(1);
									player.State = PlayerState.Normal;
								}
								break;
						}
					}
					//	もしプレイヤーのいずれかがピンチなら、BGMを早くする。
					if(FCAMAudioCue.IsCueUsable(Common.BGMCue)){
						if(isPinch){
							if(Common.BGMCue.Frequency == Common.BGMCue.DefaultFrequency){
								Common.BGMCue.Frequency = (int)(BGM_SPEEDUP_RATE * Common.BGMCue.Frequency);
							}
						}else{
							if(Common.BGMCue.Frequency != Common.BGMCue.DefaultFrequency){
								Common.BGMCue.Frequency = Common.BGMCue.DefaultFrequency;
							}
						}
					}
					break;
				}
			}
			//
		}
		protected override void Draw(FCAMGraphics g){
			//	共通画像描画
			//	
			//	背景が透過されることがあるため、一旦初期化。
			g.Clear(FCAMColor.Black);
			//	画像サイズが異なる場合に備えて自動拡大縮小
			g.DrawTexture(TextureBackground, 0, 0, g.Width, g.Height);
			g.DrawTexture(TextureBackgroundBlender, BackgroundBlenderColor);
			Common.DrawString(PosVSLabel, LABEL_VS);
			Common.DrawString(PosNextLabel, LABEL_NEXT);
			if(Mode == BattleMode.Free){
				Common.DrawString(PosInfo, Common.GetModeText(2));
			}else if(Mode == BattleMode.VS){
				Common.DrawString(PosInfo, Common.GetModeText(1));
			}else{
				Common.DrawString(PosInfo, Common.GetModeText());
			}
			Common.DrawString(PosTimeLabel, LABEL_TIME);
			Common.DrawString(PosTime, FCAMStrings.Format(TIME_TEXT_FORMAT, DateTime.Hour, DateTime.Minute, DateTime.Second));
#if RANKING
			Common.DrawString(PosScoreLabel, LABEL_SCORE);
			Common.DrawString(PosJewelsLabel, LABEL_JEWELS);
#endif
			//	プレイヤー描画（前処理）
			{
				int y2 = (int)FCAMMath.Pow(Frame);
				int y = State == BattleState.BattleEnd ? -1 * (int)FCAMMath.Pow((NUMBER_END_FRAMES - Frame) * 2) : 0;
				for(int i = 0; i < Players.Length; i++){
					Players[i].DrawBefore(g, State, y, y2);
				}
			}
			//	ジュエルフラグメント描画
			//	
			for(int i = JewelFragmentList.Count - 1; i >= 0; i--){
				DrawJewelFragment(g, JewelFragmentList[i]);
			}
			//	連鎖
			foreach(ChainText chainText in ChainTextList){
				string sz = chainText.Chain.ToString();
				List<int> list = new List<int>();
				FCAMPoint point;
				FCAMColor blender;
				int animation;
				foreach(char c in sz.ToCharArray()){
					list.Add(c - '0');
				}
				list.AddRange(CHAIN_TEXT_CHARS);
				point = chainText.CenterPos;
				animation = FCAMMath.Min(chainText.Chain - 2, MAX_CHAIN_TEXT_ANIMATION);
				blender = new FCAMColor(FCAMColor.MaxValue / NUMBER_CHAIN_TEXT_FRAMES * (NUMBER_CHAIN_TEXT_FRAMES - chainText.Frame), FCAMColor.MaxValue);
				for(int i = 0; i < list.Count; i++){
					g.Draw(TextureText, point.X, point.Y, CHAIN_TEXT_CHARACTER_SIZE.Width, CHAIN_TEXT_CHARACTER_SIZE.Height, list[i] * CHAIN_TEXT_CHARACTER_SIZE.Width, animation * CHAIN_TEXT_CHARACTER_SIZE.Height, blender);
					point.X += CHAIN_TEXT_CHARACTER_SIZE.Width;
				}
			}
			//	フェードアウトダイアグラム描画
			//	
			foreach(FadeoutDiagram diagram in FadeoutDiagramList){
				diagram.Draw(g);
			}
			//	プレイヤー描画（後処理）
			for(int i = 0; i < Players.Length; i++){
				Players[i].DrawAfter(g);
			}
			//
			//	描画
			//	
			switch(State){
				//	ゲーム開始時
				case BattleState.BattleStart:{
					//	カウント
					FCAMColor blender;
					int x, y, width, height, scale;
					//	カウント・スタート音を鳴らすためにフレームを最後に加算しているため、
					//	フレームが1つ越えた状態で描画がなされるのを防ぐために Min を使用。
					int frame = FCAMMath.Min(Frame, Base.FPS * NUMBER_COUNTS - 1);
					scale = frame % Base.FPS;
					blender = new FCAMColor(FCAMColor.MaxValue - FCAMColor.MaxValue * scale / Base.FPS, FCAMColor.MaxValue);
					width = COUNT_POSSIZE.Width + scale;
					height = COUNT_POSSIZE.Height + scale;
					foreach(Player player in Players){
						x = player.PosField.X + (FIELD_SIZE.Width - width) / 2;
						y = player.PosField.Y + (FIELD_SIZE.Height - height) / 2;
						g.Draw(TextureText, x, y, width, height, COUNT_POSSIZE.X, (frame / Base.FPS) * COUNT_POSSIZE.Height + COUNT_POSSIZE.Y, COUNT_POSSIZE.Width, COUNT_POSSIZE.Height, blender);
					}
					break;
				}
				//	シーン終了時
				case BattleState.SceneEnd:{
					g.FillRectangle(new FCAMColor(FCAMColor.MaxValue * Frame / NUMBER_SCENE_FRAMES, FCAMColor.MinValue));
					break;
				}
				//	ゲーム終了時
				case BattleState.BattleEnd:{
					FCAMColor blender;
					foreach(Player player in Players){
						FCAMRectangle rectangle;
						int x, y;
						y = FIELD_SIZE.Height / 2 + player.PosField.Y;
						//	勝者以外のプレイヤーのフィールドは暗くする
						if(player.State != PlayerState.Win){
							g.FillRectangle(
								player.PosVisibleField.X, player.PosVisibleField.Y, VISIBLE_FIELD_SIZE.Width, VISIBLE_FIELD_SIZE.Height,
								new FCAMColor(FCAMColor.HalfValue / 2 * Frame / NUMBER_END_FRAMES, FCAMColor.MinValue)
							);
						}
						//	勝敗の描画
						switch(player.State){
							case PlayerState.Lose:{
								rectangle = RESULT_LOSE_POSSIZE;
								break;
							}
							case PlayerState.Draw:{
								rectangle = RESULT_DRAW_POSSIZE;
								break;
							}
							case PlayerState.Win:{
								rectangle = RESULT_WIN_POSSIZE;
								break;
							}
							default:
								rectangle.Height = -64;
								break;
						}
						if(rectangle.Height > 0){
							blender.SetAMono(224, FCAMColor.MaxValue);
							y = (FIELD_SIZE.Height - rectangle.Height) / 2 + player.PosField.Y - FCAMMath.Within(0, SubFrame - 30, 30) * 2 ;
							x = (FIELD_SIZE.Width - rectangle.Width) / 2 + player.PosField.X;
							if(player.PlayerId == 0){
								x -= FCAMMath.Pow(2 * (NUMBER_END_FRAMES - Frame)) / 2;
							}else{
								x += FCAMMath.Pow(2 * (NUMBER_END_FRAMES - Frame)) / 2;
							}
							g.Draw(TextureText, x, y, rectangle.Width, rectangle.Height, rectangle.X, rectangle.Y, blender);
						}
						//	質問の描画用にY位置をずらす
						y += rectangle.Height + 32;
						//	質問と応答の描画
						if(SubFrame < NUMBER_END_SUBFRAMES/* || player.IsCPU*/)
							continue;
						if(IsStory){
							if(IsAutoProcessOnBattleEndInStory(player.PlayerId))
								continue;
							if(player.PlayerId != ProtagonistPlayerID)
								continue;
						}
						player.DrawContinue(g, y);
					}
					break;
				}
				//	ポーズ時
				case BattleState.Pause:{
					int x, y;
					foreach(Player player in Players){
						x = (FIELD_SIZE.Width - PAUSE_POSSIZE.Width) / 2 + player.PosField.X;
						y = (FIELD_SIZE.Height - PAUSE_POSSIZE.Height) / 2 + player.PosField.Y;
						g.Draw(TextureText, x, y, PAUSE_POSSIZE.Width, PAUSE_POSSIZE.Height, PAUSE_POSSIZE.X, PAUSE_POSSIZE.Y);
						y += PAUSE_POSSIZE.Height + 32;
						//	質問と応答の描画
						if(Mode == BattleMode.VS){
							player.DrawContinue(g, y);
						}
					}
					break;
				}
				//	メニュー時
				case BattleState.Menu:{
					/*
					bool wait = false;
					if(Players[0].State == PlayerState.Wait || Players[1].State == PlayerState.Wait){
						wait = true;
					}
					 */
					g.FillRectangle(new FCAMColor(FCAMColor.HalfValue, FCAMColor.MinValue));
					//	[PlayerId, AreaId(左 or 右)] = -1 なら非選択
					int[, ] selectedItem = new int[2, 2];
					for(int i = 0; i < Players.Length; i++){
						Player player = Players[i];
						int index = -1;
						switch(Cursors[i]){
							case 0:{index = 4; break;}
							case 1:{index = 7; break;}
							case 2:{index = 10; break;}
							case 3:{index = 13; break;}
							case 4:{index = 16; break;}
							case 5:{index = 19; break;}
						}
						if(player.State == PlayerState.Wait){
							selectedItem[i, i] = -1;
							selectedItem[i, i == 0 ? 1 : 0] = index;
						}else{
							selectedItem[i, i] = index;
							selectedItem[i, i == 0 ? 1 : 0] = -1;
						}
					}
					for(int i = 0; i < Players.Length; i++){
						Player player = Players[i];
						FCAMPoint pos = player.PosVisibleField;
						//int index = -1;
						string[] szArray;
						//	二重処理になるが、名前を再度描画させ暗くない名前を表示させる。
						player.DrawName();
						if(player.State == PlayerState.Wait){
							szArray = new string[]{
								"",
								"",
								"",
								"",
								"",
								"",
								"",
								"",
								"",
								"",
								"",
								"WAITING...",
								"",
								"",
								"",
								"",
								"",
								"",
								"",
								"",
								"",
								"",
								"",
								"",
							};
						}else{
							string name = player.Name.Trim();
							szArray = new string[]{
								"",
								"",
								"",
								"  CHARACTER: ",
								"   " + (name.Length <= 10 ? name : name.Substring(0, 10)),
								"   " + (name.Length <= 10 ? "" : name.Substring(10, name.Length - 10)),
								"  PLAYER: ",
								"   " + (player.Kind == PlayerKind.Human ? "HUMAN" : (player.Kind == PlayerKind.CPU ? "CPU" : "NONE")),
								"",
								"  <HANDICAP>",
								"   HEIGHT: " + player.StartHeight,
								"",
								"   OBSTACLE: ",
								"    "+ (player.StartJewel == JewelKind.Normal ? "NO" : (player.StartJewel == JewelKind.Damage ? "YES" : "HARD")),
								"",
								"",
								"  START",
								"",
								"",
								"  END",
								"",
								"",
								"",
								"",
							};
							//switch(Cursors[i]){
							//    case 0:{index = 4; break;}
							//    case 1:{index = 7; break;}
							//    case 2:{index = 10; break;}
							//    case 3:{index = 13; break;}
							//    case 4:{index = 16; break;}
							//    case 5:{index = 19; break;}
							//}
						}
						for(int j = 0; j < szArray.Length; j++){
							FCAMColor blender;
							var ac = szArray[j].ToCharArray();
							bool selected = false;
							for(int k = 0; k < 2; k++){
								int index = selectedItem[k, i];
								if(j == index || j == index + 1){
									selected = true;
									blender = FCAMColor.Yellow;
									if(ac.Length > 0){
										if(k == 0){
											//	{ = 1P>
											ac[0] = '{';
										}else if(k == 1){
											//	| = 2P>
											ac[1] = '|';
											//	} = <2P
											//	ac[0] = '}';
										}
									}
								}
							}
							blender = selected ? FCAMColor.Yellow : FCAMColor.HalfWhite;
							Common.DrawString(pos, (new string(ac))+ FCAMSystem.NewLine, blender);
							pos.Y += Common.FONT_SIZE.Height;
						};
					}
					break;
				}
			}
			//	デバッグ描画
			//	
#if DEBUG
			if(!FCAMDebug.IsDebug || !Game.DrawDebug)
				return;
			g.DrawString(FCAMDrawStringFormat.Right, 0, 0, "Players[0].PutWait = " + Players[0].PutWait);
#endif
		}
		/// <summary>
		/// 
		/// </summary>
		void ChangeState(BattleState state){
			Frame = 0;
			SubFrame = 0;
			State = state;
			switch(state){
				case BattleState.Menu:{
					InitBattle();
					//	NextBattle() メソッドで State が変換されるので元に戻す。
					//	State = BattleState.Menu;
					//	今は戻されない？からコメントアウト
					Common.ChangeBGM(BGM.VSMenu);
					break;
				}
				case BattleState.BattleStart:{
					InitBattle();
					if(IsStory){
						Common.ChangeBGM(Audio.BGMPath[Memory.GetClearedCharactersAtStage(Common.StageOnFile)]);
					}else{
						//	ストーリーモードでない時のBGMは、選択キャラからランダムで選択。
						var list = new List<int>();
						foreach(Player elem in Players){
							if(elem.IsParticipant){
								list.Add(elem.CharacterID);
							}
						}
						Common.ChangeBGM(Audio.BGMPath[list[FCAMRandom.Common.NextInt(list.Count)]]);
					}
					break;
				}
			}
		}
		/// <summary>
		/// 
		/// </summary>
		void InitBattle(){
			Player.ResetNextJewelRandomSeed();
			JewelFragmentList.Clear();
			ChainTextList.Clear();
			FadeoutDiagramList.Clear();
			foreach(Player player in Players){
				player.Reset();
				player.DeleteLength = FIELD_TABLESIZE.Columns / 2;
				player.ClearField();
				for(int i = 0; i < NUMBER_NEXT_JEWEL_SETS; i++){
					player.SetNextJewelSet();
				}
#if SHOW_ALL_CONTROL_JEWELS || SHOW_TRANSPARENT_CONTROL_JEWELS
			//	もしフィールドの上に操作ジュエルを見せる場合
			//	開始時にジュエルセットリストをシフトするため
			//	（見えると言っても全体が見えるわけでないし、見えない場合だと"NEXT"が初めのジュエルセットにならないといけないため）
			//	一番初めのジュエルは空にする。
				player.NowJewelSet.Clear();
#endif
				//	CPUの計算は、バトル開始（カウント終了）時に行なうこと（そこでジュエルセットリストをシフトしてゲームが始まるため）
				//	player.ResetMoveRange();
				//	player.RecalcCPUAim(this);
			}
			//	ストーリーモードなら、主人公のスコアを継続させるため、前回のスコアを登録する。
			if(IsStory){
				ProtagonistPlayer.Score = ProtagonistBeginScore;
			}
		}
		/// <summary>
		/// 
		/// </summary>
		void NextBattle(FCAMContentsManager manager){
			//	InitBattle();
			//	※InitBatle() は ChangeState() 側で行なうようにした。
			ChangeState(BattleState.BattleStart);
			//	string path = FCAMStrings.Format(FILEPATH_MUSIC_FORMAT, Common.Stage);
			//	※BattleStart時の曲変更は ChangeState() 側で行なうようにした。
			StartTime = 0;
		}
		/// <summary>
		/// ジュエルフラグメントの描画
		/// </summary>
		void DrawJewelFragment(FCAMGraphics g, JewelFragment fragment){
			int centerX = fragment.CenterPos.X;
			int centerY = fragment.CenterPos.Y;
			int width = JEWEL_SIZE.Width;
			int height = JEWEL_SIZE.Height;
			int srcX = fragment.Jewel * JEWEL_SIZE.Width;
			int srcY = fragment.Shape * JEWEL_SIZE.Height + JEWEL_FRAGMENT_SRC_Y;
			g.DrawRotatedAtOrigin(new FCAMVector(0, 0, fragment.Rotation), null, TextureJewel, centerX, centerY, width, height, srcX, srcY);
		}
		/// <summary>
		/// 指定されたプレイヤーのフィールド内のジュエルを消去フィールドを元に消去する。
		/// 同時に連鎖を初期化したりなども行なう。
		/// </summary>
		/// <returns>消去されたジュエル数</returns>
		int DeleteJewelsAndCalcIn(Player player, DeleteField deleteField){
			int numberDeletes;
			FCAMPoint center = new FCAMPoint(0);
			for(int row = 0; row < player.Field.Size.Rows; row++){
				for(int column = 0; column < player.Field.Size.Columns; column++){
					if(!deleteField.Data[column, row])
						continue;
					center.X += column * JEWEL_SIZE.Width;
					center.Y += row * JEWEL_SIZE.Height;
					for(sbyte shape = 0; shape < NUMBER_JEWEL_FRAGMENT_SHAPES; shape++){
						FCAMPoint pos;
						FCAMPoint velocity;
						pos = player.GetJewelPos(column, row);
						pos.X += JEWEL_SIZE.Width / 2;
						pos.Y += JEWEL_SIZE.Height / 2;
						velocity.X = Random.NextSign() * Random.NextInt(MIN_JEWEL_FRAGMENT_VELOCITY, MAX_JEWEL_FRAGMENT_VELOCITY);
						velocity.Y = -1 * Random.NextInt(MIN_JEWEL_FRAGMENT_VELOCITY, MAX_JEWEL_FRAGMENT_VELOCITY);
						JewelFragmentList.Add(new JewelFragment(player.Field.Data[column, row].Kind, shape, 0, (short)(Random.NextSign() * Random.NextInt(MIN_JEWEL_FRAGMENT_ROTATION_SPEED, MAX_JEWEL_FRAGMENT_ROTATION_SPEED) * NUMBER_JEWEL_FRAGMENT_ROTATION_SPEED_STEPS), pos, velocity));
					}
				}
			}
			numberDeletes = DeleteJewelsIn(player.Field, deleteField);
			if(numberDeletes == 0)
				return(0);
			player.Chain++;
			player.Jewels += numberDeletes;
			player.Score += CalcScorePoint(numberDeletes, player.Chain, player.Level, player.DeleteLength, deleteField.DamageJewels);
			if(player.Chain > 1){
				center.X = center.X / numberDeletes + JEWEL_SIZE.Width / 2 - (CHAIN_TEXT_CHARACTER_SIZE.Width * (player.Chain.ToString().Length + CHAIN_TEXT_CHARS.Length)) / 2 + player.PosField.X;
				center.Y = center.Y / numberDeletes + JEWEL_SIZE.Height / 2 - CHAIN_TEXT_CHARACTER_SIZE.Height / 2 + player.PosField.Y;
				ChainTextList.Add(new ChainText(player.Chain, center));
			}
			return(numberDeletes);
		}
		/// <summary>
		/// ストーリーモードにおける、難易度＆キャラごとの設置時間。
		/// なお第二インデックスが 0（主人公）や、存在しないステージは 0 としておく（バグと分かり易くするため）。
		/// </summary>
		static readonly int[,] PUT_TIME_EVERY_CHARACTERS = {
																{ 0, 30, 29, 28, 27, 26,  0,  0,  0,  0,  0,  0},
																{ 0, 25, 24, 23, 22, 21, 20, 19, 18,  0,  0,  0},
																{ 0, 17, 16, 15, 14, 13, 12, 11, 10,  9,  8,  7},
																{ 0,  0,  0,  0,  0,  0,  0,  0,  0,  6,  5,  4},
															};
		/// <summary>
		/// 設置までの時間を取得する。
		/// </summary>
		int GetPutTime(Player player){
			if(IsStory){
				//	ストーリーモードなら、プレイヤー問わず敵キャラクターIDごとなどで固定
				return PUT_TIME_EVERY_CHARACTERS[Game.MemoryData.Difficult, OpponentPlayer.CharacterID];
			}else{
				//	ストーリーモード以外は、各々のプレイヤーのレベルに応じた設置時間とする。
				return FCAMMath.Within(5, 31 - player.Level, 30);
			}
		}
		/// <summary>
		/// ストーリーモードにおける、難易度＆キャラごとの落下速度。
		/// なお第二インデックスが 0（主人公）や、存在しないステージは 999 としておく（バグと分かり易くするため）。
		/// </summary>
		static readonly int[,] FALL_SPEED_EVERY_CHARACTERS = {
																{ 999,   1,   1,   1,   1,   2, 999, 999, 999, 999, 999, 999},
																{ 999,   2,   2,   3,   3,   3,   4,   4,   5, 999, 999, 999},
																{ 999,   5,   6,   6,   7,   7,   8,   8,   9,  10,  11,  12},
																{ 999, 999, 999, 999, 999, 999, 999, 999, 999,  12,  15,  18},
															};
		/// <summary>
		/// 落下速度を取得する。
		/// </summary>
		int GetFallSpeed(Player player){
			if(IsStory){
				//	ストーリーモードなら、プレイヤー問わず敵キャラクターIDごとなどで固定
				return FALL_SPEED_EVERY_CHARACTERS[Game.MemoryData.Difficult, OpponentPlayer.CharacterID];
			}else{
				//	ストーリーモード以外は、各々のプレイヤーのレベルに応じた落下速度とする。
				return FCAMMath.Within(1, player.Level, 15);
			}
		}
	}
}
