﻿//#define	AUTO_BGM

using FukuracamDX;
using System.Collections.Generic;

namespace FatopTheWorld.Scenes{
	enum Commands: sbyte{
		Unknown = -1,
		None = 0,
		CommentOut,
		LanguageBranchStarts,
		LanguageBranchEnds,
		Opening,
		KeyWait,
		Background,
		PlayMusic,
		PlaySound,
		Versus,
		TurnPage,
		PortraitLeft,
		PortraitCenter,
		PortraitRight,
		PortraitNothing,
		ActiveLeft,
		ActiveCenter,
		ActiveRight,
		ActiveNarration,
		ActiveAll,
		LeftTalks,
		CenterTalks,
		RightTalks,
		NarrationTalks,
		AllTalks,
		NextStory,
		Ending,
		Quit,
	}
	struct StoryChar{
		static readonly string[] COMMAND_NAMES = new string[]{
			"[",	//	文字の表示を意味するもの。"[" にしてるのは、これをコマンドとしてのちに文字として追加するため
			"co",
			"lbs",
			"lbe",
			"op",
			"kw",
			"bg",
			"pm",
			"ps",
			"vs",
			"tp",
			"pl",
			"pc",
			"pr",
			"pn",
			"al",
			"ac",
			"ar",
			"an",
			"aa",
			"lt",
			"ct",
			"rt",
			"nt",
			"at",
			"ns",
			"ed",
			"qt",
			"",
		};
		public Commands Command;
		public object[] Parameters;
		public StoryChar(object[] commandAndParameters = null){
			if(commandAndParameters == null){
				commandAndParameters = new object[0];
			}
			Command = Commands.Unknown;
			Parameters = new object[commandAndParameters.Length - 1];
			for(int i = 1; i < commandAndParameters.Length; i++){
				Parameters[i - 1] = commandAndParameters[i];
			}
			for(sbyte i = 0; i < COMMAND_NAMES.Length; i++){
				if((string)commandAndParameters[0] == COMMAND_NAMES[i]){
					Command = (Commands)i;
				}
			}
			if(Command == Commands.Unknown){
				Error.ID_00008.Throw(commandAndParameters[0]);
			}
		}
		public StoryChar(Commands command, object[] parameters = null){
			if(parameters == null){
				parameters = new object[0];
			}
			Command = command;
			Parameters = parameters;
		}
		public StoryChar(char character): this(Commands.None, new object[]{character}){}
	}
	class StoryTeller{
		const char COMMAND_PARAMETERS_SPLITER_CHAR = ',';
		const string COMMAND_DEFINITION_CHARS = "[]";
		public int Pos;
		public List<StoryChar> CharList{get; private set;}
		private System.Text.StringBuilder __Message;
		public string Message{get{return(__Message.ToString());}}
		public StoryTeller(){
			__Message = new System.Text.StringBuilder();
			CharList = new List<StoryChar>();
			CharList.Add(new StoryChar(Commands.NextStory));
		}
		public void LoadStory(FCAMContentsManager manager, string path){
			bool command = false;
			List<StoryChar> list = new List<StoryChar>();
			System.Text.StringBuilder sb = new System.Text.StringBuilder();
			foreach(char c in manager.MakeString(path).ToCharArray()){
				if(command){
					if(c == COMMAND_DEFINITION_CHARS[1]){
						try{
							StoryChar sc = new StoryChar(sb.ToString().Split(COMMAND_PARAMETERS_SPLITER_CHAR));
							//	作ったコマンドが省略形([lt] = [kw][tp][al] 等)ならそれを展開
							//	または自動展開([vs,xx] = [xs,xx][tp])
							switch(sc.Command){
								case Commands.Versus:{
									list.Add(sc);
									list.Add(new StoryChar(Commands.TurnPage));
									break;
								}
								case Commands.PortraitLeft:{
									list.Add(sc);
									list.Add(new StoryChar(Commands.ActiveLeft));
									break;
								}
								case Commands.PortraitCenter:{
									list.Add(sc);
									list.Add(new StoryChar(Commands.ActiveCenter));
									break;
								}
								case Commands.PortraitRight:{
									list.Add(sc);
									list.Add(new StoryChar(Commands.ActiveRight));
									break;
								}
								case Commands.PortraitNothing:{
									list.Add(new StoryChar(Commands.PortraitLeft));
									list.Add(new StoryChar(Commands.PortraitCenter));
									list.Add(new StoryChar(Commands.PortraitRight));
									break;
								}
								case Commands.LeftTalks:{
									list.Add(new StoryChar(Commands.KeyWait));
									list.Add(new StoryChar(Commands.TurnPage));
									if(sc.Parameters.Length != 0){
										list.Add(new StoryChar(Commands.PortraitLeft, sc.Parameters));
									}
									list.Add(new StoryChar(Commands.ActiveLeft));
									break;
								}
								case Commands.CenterTalks:{
									list.Add(new StoryChar(Commands.KeyWait));
									list.Add(new StoryChar(Commands.TurnPage));
									if(sc.Parameters.Length != 0){
										list.Add(new StoryChar(Commands.PortraitCenter, sc.Parameters));
									}
									list.Add(new StoryChar(Commands.ActiveCenter));
									break;
								}
								case Commands.RightTalks:{
									list.Add(new StoryChar(Commands.KeyWait));
									list.Add(new StoryChar(Commands.TurnPage));
									if(sc.Parameters.Length != 0){
										list.Add(new StoryChar(Commands.PortraitRight, sc.Parameters));
									}
									list.Add(new StoryChar(Commands.ActiveRight));
									break;
								}
								case Commands.NarrationTalks:{
									list.Add(new StoryChar(Commands.KeyWait));
									list.Add(new StoryChar(Commands.TurnPage));
									list.Add(new StoryChar(Commands.ActiveNarration));
									break;
								}
								case Commands.AllTalks:{
									list.Add(new StoryChar(Commands.KeyWait));
									list.Add(new StoryChar(Commands.TurnPage));
									list.Add(new StoryChar(Commands.ActiveAll));
									break;
								}
								default:{
									list.Add(sc);
									break;
								}
							}
						}catch(FCAMException e){
							Error.ID_00009.Throw(path, e.ToString());
						}
						sb.Length = 0;
						command = false;
					}else{
						sb.Append(c);
					}
				}else{
					if(c == COMMAND_DEFINITION_CHARS[0]){
						command = true;
					}else{
						list.Add(new StoryChar(c));
					}
				}
			}
			CharList = list;
			Pos = 0;
			__Message.Length = 0;
		}
		/// <summary>
		/// 
		/// </summary>
		public StoryChar Next(bool decide, bool skip){
			StoryChar sc = new StoryChar(Commands.NextStory);
			while(Pos < CharList.Count){
				sc = CharList[Pos];
				switch(sc.Command){
					case Commands.None:{
						char c = (char)sc.Parameters[0];
						bool b = c == '\r' || c == '\n';
						if(__Message.Length == 0 && b){
						}else{
							__Message.Append(c);
						}
						Pos++;
						break;
					}
					case Commands.KeyWait:{
						//	直前が改行でなくなるまで削除を繰り返す。
						bool removeNewline = false;
						while(__Message.Length > 0){
							char c = __Message[__Message.Length - 1];
							if(c == '\r' || c == '\n'){
								__Message.Remove(__Message.Length - 1, 1);
								removeNewline = true;
							}else{
								break;
							}
						}
						if(removeNewline){
							CharList.Insert(Pos + 1, new StoryChar('\n'));
						}
						if(decide){
							Pos++;
						}
						goto default;
					}
					case Commands.TurnPage:{
						__Message.Length = 0;
						Pos++;
						skip = false;
						break;
					}
					default:{
						skip = false;
						break;
					}
				}
				if(!skip)
					break;
			}
			return(sc);
		}
	}
	[System.Flags] enum ActivePortrait: sbyte{
		Narration = 0, Left = 1, Center = 2, Right = 4, All = Left | Center | Right
	}
	class Story: FCAMScene{
		enum StoryState: sbyte{
			Telling, MovingToBattle
		}
		const int LAST_BGM_STAGE = 8;
		const string SE_GAMEOVER_FILEPATH = @"SE\gameover.wav";
		const string SE_BATTLE_START_FILEPATH = @"SE\battle_start.wav";
		const string FILEPATH_MESSAGE_WINDOW = @"Story\images\MessageWindow.png";
		const string FILEPATH_FORMAT_BACKGROUND = @"Story\images\backgrounds\{0}.png";
		const string FILEPATH_FORMAT_CHARACTER = @"Character\{0:2}-{1:1}.png";
		const string FILEPATH_FORMAT_STORY_EVERY_DIFFICULT = @"Story\{0}\story{1:D2}.txt";
#if TRIAL
		const string FILEPATH_FORMAT_STORY = @"Story\Trial\story{0:D2}.txt";
#else
		const string FILEPATH_FORMAT_STORY = @"Story\story{0:D2}.txt";
#endif
		const int FRAME_MENU = 32;
		const int BLINKTIME_INTERVAL = 10;
		const int FRAME_TOBATTLE = 64;
		const float TOBATTLE_SCALE_STEPS = 1.5f;
		const float TOBATTLE_ROTATE_STEPS = 4.0f;
		/// <summary>
		/// lbs～lbe間でバトルの前後の状態。lbs 呼び出しごとに null で初期化。
		/// null:バトル前共通バトル開始コマンド前
		/// false:バトル前共通バトル開始コマンド中 or バトル後共通コマンドバトル終了コマンド中
		/// true:バトル後共通コマンドバトル終了コマンド前
		/// </summary>
		private bool? LanguageBranchBattleState;
		/// <summary>lbs～lbe コマンド内か。LoadStory ごとに false で初期化。</summary>
		private bool LanguageBranch;
		/// <summary>言語による分岐が有効か（lbs～lbe コマンド内でかつ言語が NonJpn の時）。</summary>
		private bool LanguageBranchEnabled{
			get{
				return Game.Instance.Language == Languages.NonJpn && LanguageBranch;
			}
		}
		private string LanguageBranchCharaID;
		int Frame = 0;
		int Animation = 0;
		int BlinkTime = 0;
		StoryState State;
		FCAMTexture TextureBackground, TextureMessageWindow, TextureLeft, TextureCenter, TextureRight;
		StoryTeller Teller;
		ActivePortrait ActiveCharacter;
		bool ShowKeyWait = false;
		/// <summary>バトルシーンから復帰したあと、勝ち負けによってストーリーを展開させるか判定させるため用</summary>
		bool Battle;
		/// <summary>バトルシーンから復帰したあと、BGMを元に戻すため用</summary>
		string BeforeBGM = "";
		long BeforeBGMPosition = 0;
		static readonly FCAMRectangle MESSAGE_WINDOW_POSSIZE = new FCAMRectangle(false, 0, 360, 640, 120);
		static readonly FCAMRectangle MESSAGE_POSSIZE = new FCAMRectangle(false, 16, 376, 624, 104);
		//
		public sbyte PlayerID = -1;
		public sbyte CharacterID = 0;
		/// <summary>
		/// ストーリーファイル読み込み。
		/// 読み込めなかったら（ファイルがなかったら）そのファイルパスを返す。ある場合は null。
		/// </summary>
		private string LoadStory(FCAMContentsManager manager, sbyte number){
			LanguageBranch = false;
			string path = FCAMStrings.Format(FILEPATH_FORMAT_STORY_EVERY_DIFFICULT, Game.MemoryData.Difficult, number);
			if(!manager.Exist(path)){
				//	難易度ごとのストーリーファイルが存在しない場合は、共通のストーリーファイルを読み込む。
				path = FCAMStrings.Format(FILEPATH_FORMAT_STORY, number);
			}
			Frame = 0;
			Animation = 0;
			if(manager.Exist(path)){
				Teller.LoadStory(manager, path);
				return(null);
			}
			State = StoryState.Telling;
			return(path);
		}
		private void QuitStory(bool ending){
			ExchangeScene<Scenes.Gameover>();
			if(ending){
				PushScene<Scenes.Ending>();
			}
		}
		private FCAMTexture LoadPortrait(FCAMContentsManager manager, string characterID, string expressionID){
			return(manager.MakeTexture(FCAMStrings.Format(FILEPATH_FORMAT_CHARACTER, characterID, expressionID)));
		}
		private void StartBattle(){
			Animation = 0;
			State = StoryState.MovingToBattle;
			BeforeBGM = Common.BGMFilePath;
			if(Common.BGMCue != null){
				BeforeBGMPosition = Common.BGMCue.Position;
			}
			Common.ChangeBGM(BGM.Nothing);
			Common.PlaySE(SE_BATTLE_START_FILEPATH);
		}
		protected override void Load(FCAMGraphics g, FCAMInput input, FCAMContentsManager manager){
#if DEBUG
			if(FCAMDebug.IsDebug){
				Base.Text = "Story";
			}
#endif
			TextureMessageWindow = manager.MakeTexture(FILEPATH_MESSAGE_WINDOW);
			TextureBackground = null;
			TextureLeft = null;
			TextureCenter = null;
			TextureRight = null;
			Teller = new StoryTeller();
			ActiveCharacter = ActivePortrait.Narration;
			Common.Battle.ProtagonistBeginScore = 0;
			Battle = false;
			string notFoundPath;
			if((notFoundPath = LoadStory(manager, Common.StageOnFile)) != null){
				//	QuitStory();
				//	ストーリーが読み込めなかったら例外を投げる
				Error.ID_00010.Throw(notFoundPath);
			}else{
				ChangeBGMToStory();
			}
		}
		protected override void Unload(){
			FCAMSystem.DisposeClass(out TextureRight);
			FCAMSystem.DisposeClass(out TextureCenter);
			FCAMSystem.DisposeClass(out TextureLeft);
			FCAMSystem.DisposeClass(out TextureBackground);
			FCAMSystem.DisposeClass(out TextureMessageWindow);
		}
		protected override void Active(FCAMScene popedScene, FCAMContentsManager manager){
			ChangeBGMToStory();
			Frame = 0;
		}
		protected override void Calc(FCAMInput input, FCAMContentsManager manager){
			switch(State){
				case StoryState.Telling:{
					ShowKeyWait = false;
#if DEBUG
					if(FCAMDebug.IsDebug){
						if(Common.KeyDebug2){
							Frame = FRAME_MENU;
						}
					}
#endif
					BlinkTime = (BlinkTime + 1) % (BLINKTIME_INTERVAL * 2);
					if(Frame < FRAME_MENU){
						if(Frame == 0){
							//	シーンの始めに間を作る
							FCAMSystem.Sleep(1000);
						}
						Frame++;
					}
					if(!Battle){
						StoryChar sc;
						if(LanguageBranchEnabled){
							//	言語による分岐モードなら、[vs] [lbe]以外のコマンドを無視する。
							//	但し言語モード独自の対戦前のメッセージを表示してる最中なら、対戦までメッセージを表示
							if(LanguageBranchBattleState == null || LanguageBranchBattleState == true){
								do{
									sc = Teller.CharList[Teller.Pos];
									if(sc.Command == Commands.Versus || sc.Command == Commands.LanguageBranchEnds){
										break;
									}
									Teller.Pos++;
								}while(true);
								if(sc.Command == Commands.Versus){
									//	戦闘前に独自メッセージを表示してなければ、それを表示
									if(LanguageBranchBattleState == null){
										int i = Teller.Pos;
										foreach(char c in Title.STAGETITLE_NAME[Game.MemoryData.Difficult, Common.Stage] + FCAMSystem.NewLine + " "){
											Teller.CharList.Insert(i++, new StoryChar(c));
										}
										Teller.CharList.Insert(i, new StoryChar(Commands.KeyWait));
										FCAMSystem.DisposeClass(out TextureCenter);
										if(LanguageBranchCharaID != null){
											TextureCenter = LoadPortrait(manager, LanguageBranchCharaID, "0");
											ActiveCharacter = ActivePortrait.All;
										}
										Teller.Next(Common.CommonKeyA(), Common.CommonKeyB().Press);
										sc = new StoryChar(Commands.None);
										LanguageBranchBattleState = false;
									}
								}else if(sc.Command == Commands.LanguageBranchEnds){
									//	言語による分岐終了後、開始時にキャラクターＩＤが設定されている時はメッセージ表示
									if(LanguageBranchCharaID != null){
										int i = Teller.Pos;
										Teller.CharList.Insert(i++, new StoryChar(Commands.TurnPage));
										foreach(char c in "YOU WIN!" + FCAMSystem.NewLine + " "){
											Teller.CharList.Insert(i++, new StoryChar(c));
										}
										Teller.CharList.Insert(i, new StoryChar(Commands.KeyWait));
										Teller.Next(Common.CommonKeyA(), Common.CommonKeyB().Press);
										sc = new StoryChar(Commands.None);
										LanguageBranchBattleState = false;
									}
								}
							}else{
								//	言語による分岐モードでの共通メッセージ表示中なら
								sc = Teller.Next(Common.CommonKeyA(), Common.CommonKeyB().Press);
							}
						}else{
							//	通常なら
							sc = Teller.Next(Common.CommonKeyA(), Common.CommonKeyB().Press);
						}
						switch(sc.Command){
							case Commands.CommentOut:{
								Teller.Pos++;
								break;
							}
							case Commands.LanguageBranchStarts:{
								Teller.Pos++;
								LanguageBranch = true;
								LanguageBranchCharaID = sc.Parameters.Length == 0 ? null : sc.Parameters[0].ToString();
								LanguageBranchBattleState = null;
								break;
							}
							case Commands.LanguageBranchEnds:{
								Teller.Pos++;
								LanguageBranch = false;
								break;
							}
							case Commands.Opening:{
								Teller.Pos++;
								PushScene<Scenes.Opening>();
								break;
							}
							case Commands.KeyWait:{
								ShowKeyWait = true;
								break;
							}
							case Commands.Background:{
								FCAMSystem.DisposeClass(out TextureBackground);
								if(sc.Parameters.Length == 1){
									TextureBackground = manager.MakeTexture(string.Format(FILEPATH_FORMAT_BACKGROUND, sc.Parameters[0].ToString()));
								}
								Teller.Pos++;
								break;
							}
							case Commands.PlayMusic:{
								switch(sc.Parameters.Length){
									case 0:{
										Common.ChangeBGM(BGM.Nothing);
										break;
									}
									case 1:{
										//Common.ChangeBGM(Audio.BGMDir + sc.Parameters[0].ToString());
										Common.ChangeBGM(FCAMSystem.ParseEnum<BGM>(sc.Parameters[0].ToString()));
										break;
									}
									case 2:{
										//Common.ChangeBGM(Audio.BGMDir + sc.Parameters[0].ToString(), true, int.Parse(sc.Parameters[1].ToString()));
										Common.ChangeBGM(FCAMSystem.ParseEnum<BGM>(sc.Parameters[0].ToString()), true, int.Parse(sc.Parameters[1].ToString()));
										break;
									}
									default:{
										Error.ID_00004.Throw();
										break;
									}
								}
								Teller.Pos++;
								break;
							}
							case Commands.Versus:{
								StartBattle();
								Teller.Pos++;
								break;
							}
							case Commands.PortraitLeft:{
								FCAMSystem.DisposeClass(out TextureLeft);
								if(sc.Parameters.Length != 0){
									TextureLeft = LoadPortrait(manager, sc.Parameters[0].ToString(), sc.Parameters[1].ToString());
								}
								Teller.Pos++;
								break;
							}
							case Commands.PortraitCenter:{
								FCAMSystem.DisposeClass(out TextureCenter);
								if(sc.Parameters.Length != 0){
									TextureCenter = LoadPortrait(manager, sc.Parameters[0].ToString(), sc.Parameters[1].ToString());
								}
								Teller.Pos++;
								break;
							}
							case Commands.PortraitRight:{
								FCAMSystem.DisposeClass(out TextureRight);
								if(sc.Parameters.Length != 0){
									TextureRight = LoadPortrait(manager, sc.Parameters[0].ToString(), sc.Parameters[1].ToString());
								}
								Teller.Pos++;
								break;
							}
							case Commands.ActiveLeft:{
								ActiveCharacter = ActivePortrait.Left;
								Teller.Pos++;
								break;
							}
							case Commands.ActiveCenter:{
								ActiveCharacter = ActivePortrait.Center;
								Teller.Pos++;
								break;
							}
							case Commands.ActiveRight:{
								ActiveCharacter = ActivePortrait.Right;
								Teller.Pos++;
								break;
							}
							case Commands.ActiveNarration:{
								ActiveCharacter = ActivePortrait.Narration;
								Teller.Pos++;
								break;
							}
							case Commands.ActiveAll:{
								ActiveCharacter = ActivePortrait.All;
								Teller.Pos++;
								break;
							}
							case Commands.NextStory:{
#if AUTO_BGM
								Common.ChangeBGM(BGM.Nothing);
#endif
								Common.Stage++;
								if(LoadStory(manager, Common.StageOnFile) != null){
									//	ストーリーファイルが読み込めなかった場合。
									//	エンディングへ行く
									//GotoEnding();
									//	エンディングやストーリー終了は確りとストーリーファイルに記述するため、
									//	もし読み込みに失敗したら例外を投げる。
									FCAMSystem.Throw(new FCAMException("Program Error: Not found the story file!"));
								}else{
									//	次のストーリーに行く時にデータを保存する。
									FCAMFile.Write(Memory.FILEPATH_MEMORY, Game.MemoryData);
									ChangeBGMToStory();
								}
								break;
							}
							case Commands.Ending:{
								QuitStory(true);
								Teller.Pos++;
								//	エンディング後、更に先のステージがある場合はステージを増加させ、そして保存する。
								if(Common.Stage < Common.MaxStage){
									Common.Stage++;
								}
								//	次のストーリーに行く時にデータを保存する。
								FCAMFile.Write(Memory.FILEPATH_MEMORY, Game.MemoryData);
								break;
							}
							case Commands.Quit:{
								this.PopScene();
								break;
							}
						}
					}else if(Battle){
						//	バトル終了後
						if(Common.Battle.IsProtagonistLose){
							//	ゲームオーバー
							Common.ChangeBGM(BGM.Nothing);
							Common.PlaySE(SE_GAMEOVER_FILEPATH);
							PopScene();
						}else{
							Common.ChangeBGM(BeforeBGM);
							if(Common.BGMCue != null){
								Common.BGMCue.Position = BeforeBGMPosition;
							}
							Common.Battle.ProtagonistBeginScore += Common.Battle.ProtagonistScore;
							//	言語による分岐
							if(LanguageBranchEnabled){
								FCAMSystem.DisposeClass(out TextureCenter);
								if(LanguageBranchCharaID != null){
									TextureCenter = LoadPortrait(manager, LanguageBranchCharaID, "1");
								}
								LanguageBranchBattleState = true;
							}
						}
						Battle = false;
					}
					break;
				}
				case StoryState.MovingToBattle:{
					if(Animation >= FRAME_TOBATTLE){
						Common.Battle.Mode = Scenes.Battle.BattleMode.Story;
						Common.Battle.ProtagonistPlayerID = PlayerID;
						PushScene(Common.Battle);
						if(PlayerID == 0){
							Common.Battle.SetPlayer(manager, 0, CharacterID, Scenes.Battle.PlayerKind.Human);
							Common.Battle.SetPlayer(manager, 1, Memory.GetClearedCharactersAtStage(Common.StageOnFile), Scenes.Battle.PlayerKind.CPU);
						}else{
							Common.Battle.SetPlayer(manager, 0, Memory.GetClearedCharactersAtStage(Common.StageOnFile), Scenes.Battle.PlayerKind.CPU);
							Common.Battle.SetPlayer(manager, 1, CharacterID, Scenes.Battle.PlayerKind.Human);
						}
						Battle = true;
						State = StoryState.Telling;
					}else{
						Animation++;
					}
					break;
				}
			}
		}
		protected override void Draw(FCAMGraphics g){
			switch(State){
				case StoryState.Telling:{
					if(Battle){
						//	バトルから復帰後は、処理が進むまで描画しない。
						break;
					}
					g.Font = Common.FontStory;
					var blender = new {
						Normal = new FCAMColor(FCAMColor.MaxValue * 3 / 4, FCAMColor.MaxValue),
						Active = new FCAMColor(FCAMColor.MaxValue, FCAMColor.MaxValue),
						Passive = new FCAMColor(FCAMColor.MaxValue / 2, FCAMColor.MaxValue)
					};
					//FCAMColor blender = new FCAMColor(FCAMColor.MaxValue * 3 / 4, FCAMColor.MaxValue);
					//FCAMColor activeBlender = new FCAMColor(FCAMColor.MaxValue, FCAMColor.MaxValue);
					//FCAMColor passiveBlender = new FCAMColor(FCAMColor.MaxValue / 2, FCAMColor.MaxValue);
					string message = Teller.Message + ((BlinkTime) / BLINKTIME_INTERVAL == 0 && ShowKeyWait ? "▼" : "");
					if(TextureBackground == null){
						g.Clear(FCAMColor.Black);
					}else{
						g.DrawTexture(TextureBackground);
					}
					if(TextureLeft != null){
						g.DrawTexture(TextureLeft, (g.Width / 2) - TextureLeft.Width, TextureLeft.Height - g.Height, FCAMFlip.H, (ActiveCharacter & ActivePortrait.Left) != 0 ? blender.Active : blender.Passive);
					}
					if(TextureCenter != null){
						g.DrawTexture(TextureCenter, (g.Width - TextureCenter.Width) / 2, TextureCenter.Height - g.Height, (ActiveCharacter & ActivePortrait.Center) != 0 ? blender.Active : blender.Passive);
					}
					if(TextureRight != null){
						g.DrawTexture(TextureRight, g.Width / 2, TextureRight.Height - g.Height, (ActiveCharacter & ActivePortrait.Right) != 0 ? blender.Active : blender.Passive);
					}
					g.Draw(TextureMessageWindow, MESSAGE_WINDOW_POSSIZE.X, MESSAGE_WINDOW_POSSIZE.Y, MESSAGE_WINDOW_POSSIZE.Width, MESSAGE_WINDOW_POSSIZE.Height, 0, 0, blender.Normal);
					g.DrawString(MESSAGE_POSSIZE, message);
					break;
				}
				case StoryState.MovingToBattle:{
					g.FillRotatedRectangleAtOrigin(new FCAMVector(0.0f, 0.0f, Animation * TOBATTLE_ROTATE_STEPS), null, g.Width / 2, g.Height / 2, FCAMMath.Pow(TOBATTLE_SCALE_STEPS, Animation / 2), FCAMMath.Pow(TOBATTLE_SCALE_STEPS, Animation / 2), FCAMColor.HalfWhite);
					break;
				}
			}
			g.FillRectangle(new FCAMColor(FCAMMath.Max(FCAMColor.MaxValue - Frame * 8, FCAMColor.MinValue), FCAMColor.MinValue));
		}
		private void ChangeBGMToStory(){
#if AUTO_BGM
			if(Common.Stage == LAST_BGM_STAGE){
				Common.ChangeBGM(BGM.StoryLast, true);
			}else{
				Common.ChangeBGM(BGM.StoryNormal, true);
			}
#endif
		}
	}
}
