﻿using System.Collections.Generic;
using FukuracamDX;

namespace FatopTheWorld.Scenes.Telop{
	class TelopBase: FCAMScene{
		public string TelopFileNameFormat{get; protected set;}
		const string HELP_MESSAGE_FORMAT = "A Button: to next telop   B Button: to {0}";
		const int FRAMES_BASE = 300;
		const int NEXTTELOP_FADE = -128;
		int FRAME_NEXTTELOP;
		int FRAME_FADEOUT;
		const int FRAMES_PER_LINE = 50;
		const int TELOP_MOVE_SPEED = 8;
		const int TELOP_FADE_SPEED = 8;
		int TELOPNUMBER_NEXTSCENE;
		int frame, telopFileNumber, telopFrame;
		int alpha;
		string telopData;
		List<string> telopLines;
		void NextTelop(FCAMContentsManager manager){
			telopFileNumber++;
			if(!string.IsNullOrEmpty(TelopFileNameFormat)){
				telopData = manager.MakeString(FCAMStrings.Format(TelopFileNameFormat, telopFileNumber));
			}else{
				telopData = "";
			}
			telopLines.Clear();
			foreach(string value in FCAMStrings.SplitLine(telopData)){
				telopLines.Add(value);
			}
			telopFrame = 0;
			FRAME_NEXTTELOP = FRAMES_BASE + telopLines.Count * FRAMES_PER_LINE;
			FRAME_FADEOUT = FRAME_NEXTTELOP / 2;
		}
		protected override void Load(FCAMGraphics g, FCAMInput input, FCAMContentsManager manager){
#if DEBUG
			if(FCAMDebug.IsDebug){
				Base.Text = this.GetType().Name;
			}
#endif
			Common.ChangeBGM(BGM.Nothing);
			telopLines = new List<string>();
			frame = telopFileNumber = 0;
			for(TELOPNUMBER_NEXTSCENE = 1; TELOPNUMBER_NEXTSCENE < 100; TELOPNUMBER_NEXTSCENE++){
				if(!manager.Exist(FCAMStrings.Format(TelopFileNameFormat, TELOPNUMBER_NEXTSCENE + 1))){
					break;
				}
			}
			NextTelop(manager);
			g.Font = Base.GetDefaultFont();
		}
		protected override void Unload(){
		}
		protected override void Calc(FCAMInput input, FCAMContentsManager manager){
#if DEBUG
			if(FCAMDebug.IsDebug){
				if(Common.KeyDebug2){
					telopFrame = FRAME_NEXTTELOP;
					telopFileNumber = TELOPNUMBER_NEXTSCENE;
				}
			}
#endif
			frame++;
			telopFrame++;
			if(Common.CommonKeyA()){
				telopFrame = FRAME_NEXTTELOP;
			}
			if(Common.CommonKeyB()){
				telopFrame = FRAME_NEXTTELOP;
				telopFileNumber = TELOPNUMBER_NEXTSCENE;
			}
			if(telopFrame > FRAME_NEXTTELOP){
				if(telopFileNumber < TELOPNUMBER_NEXTSCENE){
					NextTelop(manager);
				}else{
					PopScene();
				}
			}
			alpha = telopFrame < FRAME_FADEOUT ? telopFrame * TELOP_FADE_SPEED : FCAMColor.MaxValue - (telopFrame - FRAME_FADEOUT) * TELOP_FADE_SPEED;
			if(alpha < NEXTTELOP_FADE){
				telopFrame = FRAME_NEXTTELOP;
			}
		}
		protected override void Draw(FCAMGraphics g){
			FCAMRectangle rectangle = g.Font.Measure(telopData);
			int baseY;
			//g.FillGradatedRectangle(0, 0, g.Width, g.Height, FCAMColor.Red, FCAMColor.Blue, FCAMColor.Green, FCAMColor.Gray);
			g.FillRectangle(0, 0, g.Width, g.Height, FCAMColor.Black);
			rectangle.VerticalCenterInto(g.Height);
			baseY = rectangle.Y;
			for(int i = 0; i < telopLines.Count; i++){
				string sz = telopLines[i];
				int y;
				rectangle = g.Font.Measure(sz);
				rectangle.HorizontalCenterInto(g.Width);
				y = baseY + i * g.Font.Height;
				if(telopFrame < FRAME_NEXTTELOP / 2){
				    rectangle.Offset(0, FCAMMath.Min(telopFrame * TELOP_MOVE_SPEED - g.Font.Height, y));
				}else{
				    rectangle.Offset(0, FCAMMath.Max((telopFrame - FRAME_NEXTTELOP / 2) * TELOP_MOVE_SPEED + baseY, y));
				}
				g.DrawString(new FCAMColor(FCAMMath.Within(FCAMColor.MinValue, alpha, FCAMColor.MaxValue), FCAMColor.MaxValue),
							 rectangle,
							 sz);
			}
			Common.DrawHelp(FCAMStrings.Format(HELP_MESSAGE_FORMAT, PeekScene().Previous is Scenes.Story ? "Story" : "Title" ));
		}
	}
}
